<?php

/*
  $Id$

  osCommerce, Open Source E-Commerce Solutions
  http://www.oscommerce.com

  Copyright (c) 2023 osCommerce

  Released under the GNU General Public License
 */

require (__DIR__ . "/paypal_rest/paypal.php");

use \PayPal;

if (!class_exists('paypal_rest')) {
class paypal_rest extends PaypalMiddleWare {

    var $code, $title, $description, $enabled;
    
    const SANDBOX_PARTNER_ATTRIBUTION_ID = 'FLAVORsb-mstny8200048_MP';
    const PARTNER_MERCHANT_SANDBOX_ID = '88LZN7N3UWLF8';
    const PARTNER_ATTRIBUTION_ID = 'HOLBIGROUPLTD_SP_OSCOMMERCE';
    const PARTNER_MERCHANT_ID = '9TZHDRH8BP3US';

    const PARTNER_APP_CLIENT_ID = 'AVC9I0Kbo0YkVYIx0LHY91HJKBGosc8-t5PMW-Pzve369xVbcf68_4MUw9cu3BC3zvz9UYqGRnk_u903';
    const PARTNER_APP_CLIENT_SECRET = '';
    const PARTNER_APP_SANDBOX_CLIENT_ID = 'Aax41DjEyTorujP5KPKaGr2J_hnez5hmh7W46Wvs81LWr2un124l2THRnJ7XlZPUiPZgLXzImbc0Zcg7';
    const PARTNER_APP_SANDBOX_CLIENT_SECRET =  '';

    const PAYMENT_PAGE = 1;
    const CONFIRMATION_PAGE = 2;
    const PROCESS_PAGE = 3;

    CONST OPYS_PENDING = 0;
    CONST OPYS_PROCESSING = 10;
    CONST OPYS_SUCCESSFUL = 20;
    CONST OPYS_REFUSED = 30;
    CONST OPYS_REFUNDED = 40;
    CONST OPYS_CANCELLED = 50;
    CONST OPYS_DISCOUNTED = 100;



    public $paid_status;
    private static $warningsFile = DIR_FS_CATALOG . 'ext' . DIRECTORY_SEPARATOR . 'modules' . DIRECTORY_SEPARATOR . 'payment' . DIRECTORY_SEPARATOR . 'runtime' . DIRECTORY_SEPARATOR . 'logs' . DIRECTORY_SEPARATOR . 'warnings_pp.log';
    private static $errorFile = DIR_FS_CATALOG . 'ext' . DIRECTORY_SEPARATOR . 'modules' . DIRECTORY_SEPARATOR . 'payment' . DIRECTORY_SEPARATOR . 'runtime' . DIRECTORY_SEPARATOR . 'logs' . DIRECTORY_SEPARATOR . 'errors_pp.log';
    private $jsIncluded = false;
    private $html_buffer = '';
    private static $debug = false;
    private $sendExVat = true; //true to send items See comment in patch webhook processing
    private $requireItemVat = false; //DE pay upon invoice /purchase_units/0/items/0/tax is required for provided payment source. Shipping is sent as product
    private $pp_commit = 'true'; //string! false - order total could be changed after confirmation. Incomaptible with APMs (is Eligable == false)
    private $jsLibAdded = false;
    
    use PaypalManagerTrait;
    use PaypalPartnerTrait;
    use PaypalPartnerTraitAPM;
    use PaypalPartnerTraitFraudNet;


    protected $webHooks = [
      'CHECKOUT.ORDER.APPROVED',
      'CHECKOUT.PAYMENT-APPROVAL.REVERSED',
      'PAYMENT.AUTHORIZATION.CREATED',
      'PAYMENT.AUTHORIZATION.VOIDED',
      'PAYMENT.CAPTURE.COMPLETED',
      'PAYMENT.CAPTURE.DENIED',
      'PAYMENT.CAPTURE.REFUNDED',
      'PAYMENT.CAPTURE.REVERSED',
      'PAYMENT.REFERENCED-PAYOUT-ITEM.COMPLETED',
      'PAYMENT.REFERENCED-PAYOUT-ITEM.FAILED',
    ];

    public static $threeDSDefaults = [
        'status' => 1,
        'contingencies' => 'SCA_WHEN_REQUIRED',
        '3dsa_y_y_1' => 1,
        '3dsa_y_n_2' => 0,
        '3dsa_y_r_2' => 0,
        '3dsa_y_a_1' => 1,
        '3dsa_y_u_3' => 0,
        '3dsa_y_u_2' => 0,
        '3dsa_y_c_3' => 0,
        '3dsa_y__2' => 0,
        '3dsa_n__2' => 1,
        '3dsa_u__2' => 1,
        '3dsa_u__3' => 0,
        '3dsa_b__2' => 1,
        '3dsa___3' => 0
    ];

    protected static $possibleFundings = [
      'paypal' => ['default_translation' => 'PayPal', 'apm' => 0, 'countries' => [], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_PAYPAL_TITLE'],
      'card' => ['default_translation' => 'Credit or debit cards', 'apm' => 0, 'countries' => [], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_CARD_TITLE'],
      'credit' => ['default_translation' => 'PayPal Credit', 'apm' => 0, 'countries' => [], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_CREDIT_TITLE'],
      'paylater' => ['default_translation' => 'Pay Later', 'apm' => 0, 'countries' => ['au', 'de', 'fr', 'it', 'sp', 'gb', 'us'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_PAYLATER_TITLE'],
      /**/
      //'venmo' => ['default_translation' => 'Venmo', 'apm' => 1, 'countries' => ['US'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_VENMO_TITLE'],
      //'pui' => ['default_translation' => 'Pay upon invoice', 'apm' => 1, 'countries' => ['de'], 'translation_key' => 'TEXT_PAY_UPON_INVOICE'],
      'bancontact' => ['default_translation' => 'Bancontact', 'apm' => 1, 'countries' => ['be'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_BANCONTACT_TITLE'],
      'blik' => ['default_translation' => 'BLIK', 'apm' => 1, 'countries' => ['pl'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_BLIK_TITLE'],
      'eps' => ['default_translation' => 'EPS', 'apm' => 1, 'countries' => ['at'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_EPS_TITLE'],
      'giropay' => ['default_translation' => 'GIROPAY', 'apm' => 1, 'countries' => ['de'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_GIROPAY_TITLE'],
      'ideal' => ['default_translation' => 'iDEAL', 'apm' => 1, 'countries' => ['nl'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_IDEAL_TITLE'],
      //'mercadopago' => ['default_translation' => 'Mercado Pago', 'apm' => 1, 'countries' => ['br', 'mx'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_MERCADOPAGO_TITLE'],
      'mybank' => ['default_translation' => 'MyBank', 'apm' => 1, 'countries' => ['it'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_MYBANK_TITLE'],
      'p24' => ['default_translation' => 'Przelewy24', 'apm' => 1, 'countries' => ['pl'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_P24_TITLE'],
      'sepa' => ['default_translation' => 'SEPA-Lastschrift', 'apm' => 0, 'countries' => [], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_SEPA_TITLE'],
      'sofort' => ['default_translation' => 'Sofort', 'apm' => 1, 'countries' => ['at', 'be', 'de', 'sp', 'it', 'nl', 'gb'], 'translation_key' => 'MODULE_PAYMENT_PAYPAL_REST_SOFORT_TITLE'],

        //'applepay' => 'ApplePay',
        /*'itau' => 'Itau',
        'zimpler' => 'Zimpler',
        'wechatpay' => 'WeChatPay',
        'payu' => 'PayU',
        'verkkopankki' => 'Verkkopankki',
        'trustly' => 'Trustly',
        'oxxo' => 'Oxxo',
        'boleto' => 'Boleto',
        'boletobancario' => 'BoletoBancario',
        'maxima' => 'Maxima',
        'multibanco' => 'Multibanco',
        'satispay' => 'Satispay',
      /**/
    ];



    function paypal_rest() {

        global $HTTP_GET_VARS, $PHP_SELF, $order, $payment;

        //$this->signature = 'paypal|paypal_rest|0.9|2.3';
        $this->api_version = '2';

        $this->code = 'paypal_rest';
        $this->title = MODULE_PAYMENT_PAYPAL_REST_TEXT_TITLE;
        $this->public_title = MODULE_PAYMENT_PAYPAL_REST_TEXT_PUBLIC_TITLE;
        $this->description = MODULE_PAYMENT_PAYPAL_REST_TEXT_DESCRIPTION;
        $this->sort_order = defined('MODULE_PAYMENT_PAYPAL_REST_SORT_ORDER') ? MODULE_PAYMENT_PAYPAL_REST_SORT_ORDER : 0;
        $this->enabled = defined('MODULE_PAYMENT_PAYPAL_REST_STATUS') && (MODULE_PAYMENT_PAYPAL_REST_STATUS == 'True') ? true : false;
        $this->order_status = defined('MODULE_PAYMENT_PAYPAL_REST_ORDER_STATUS_ID') && ((int) MODULE_PAYMENT_PAYPAL_REST_ORDER_STATUS_ID > 0) ? (int) MODULE_PAYMENT_PAYPAL_REST_ORDER_STATUS_ID : 0;
        $this->paid_status = defined('MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID') && ((int) MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID> 0) ? (int) MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID : 0;

        if (defined('MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER')) {
            if (MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER == 'Sandbox') {
                $this->title .= ' [Sandbox]';
                $this->public_title .= ' (' . $this->code . '; Sandbox)';
            }
        }

        if (!function_exists('curl_init')) {
            $this->description = '<div class="secWarning">' . MODULE_PAYMENT_PAYPAL_REST_ERROR_ADMIN_CURL . '</div>' . $this->description;

            $this->enabled = false;
        }

        if ($this->enabled === true) {
            if (!$this->_isReady()) {
                $this->description = '<div class="secWarning">' . MODULE_PAYMENT_PAYPAL_REST_ERROR_ADMIN_CONFIGURATION . '</div>' . $this->description;

                $this->enabled = false;
            }
        }

        if ($this->enabled === true) {
            if (isset($order) && is_object($order)) {
                $this->update_status();
            }
        }

        $configs = [
          'log.LogEnabled' => true,
          'log.FileName' => DIR_FS_CATALOG . 'ext' . DIRECTORY_SEPARATOR . 'modules' . DIRECTORY_SEPARATOR . 'payment' . DIRECTORY_SEPARATOR . 'runtime' . DIRECTORY_SEPARATOR . 'logs' . DIRECTORY_SEPARATOR . 'pp.log',
          'cache.FileName' => DIR_FS_CATALOG . 'ext' . DIRECTORY_SEPARATOR . 'modules' . DIRECTORY_SEPARATOR . 'payment' . DIRECTORY_SEPARATOR . 'runtime' . DIRECTORY_SEPARATOR . 'auth.cache' . DIRECTORY_SEPARATOR . 'paypal.json',
          'log.LogLevel' => 'DEBUG',
          'cache.enabled' => 1,
          'mode' => defined('MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER')? strtolower(MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER):'test'
          ];
        if ($tmp = self::getAttributionId()) {
            $configs['http.headers.PayPal-Partner-Attribution-Id'] = $tmp;
        }
        $conf = PayPal\Core\PayPalConfigManager::getInstance()->addConfigs($configs);
        if ($this->enabled === true
                        //&& \Yii::$app->id == 'app-frontend'
//2do
            ) {
          //moved to checkout_initialization_method $this->checkout_initialization_method_js();
        }

        global $HTTP_POST_VARS;
        if (isset($HTTP_POST_VARS['payment']) && substr($HTTP_POST_VARS['payment'], 0, strlen($this->code)) == $this->code) {
            $payment = $this->code;
        }
/** @var shoppingCart $cart */
        global $cart;

        if ($this->enabled && defined('MODULE_PAYMENT_PAYPAL_REST_VIRTUAL_GUEST') && MODULE_PAYMENT_PAYPAL_REST_VIRTUAL_GUEST == 'True'
            && is_object($cart) && $cart->get_content_type() == 'virtual'
            //&& ($this->isInstance()) && !$this->isShippingNeeded() && !$this->getCustomerAssigned()
            ) {
            $this->enabled = false;
        }

        // Stock Check
        if (STOCK_CHECK == 'true' && is_object($cart) && method_exists($cart, 'get_products')) {
          $any_out_of_stock = false;

          $products = $cart->get_products();
          for ($i=0, $n=sizeof($products); $i<$n; $i++) {
            $stock_check = tep_check_stock($products[$i]['id'], $products[$i]['quantity']);
            if (tep_not_null($stock_check)) {
              $any_out_of_stock = 1;
            }
          }
          // Out of Stock
          if ( (STOCK_ALLOW_CHECKOUT != 'true') && $any_out_of_stock)  {
            $this->enabled = false;
          }
        }

    }

    function update_status() {
        global $order;

        if (($this->enabled == true) && ((int) MODULE_PAYMENT_PAYPAL_REST_ZONE > 0)) {
            $check_flag = false;
            $check_query = tep_db_query("select zone_id from " . TABLE_ZONES_TO_GEO_ZONES . " where geo_zone_id = '" . MODULE_PAYMENT_PAYPAL_REST_ZONE . "' and zone_country_id = '" . $order->delivery['country']['id'] . "' order by zone_id");
            while ($check = tep_db_fetch_array($check_query)) {
                if ($check['zone_id'] < 1) {
                    $check_flag = true;
                    break;
                } elseif ($check['zone_id'] == $order->delivery['zone_id']) {
                    $check_flag = true;
                    break;
                }
            }

            if ($check_flag == false) {
                $this->enabled = false;
            }
        }
    }


    function checkout_initialization_method_js() {
        if (empty($this->jsIncluded)) {
            $this->jsIncluded = true;
            $clid = $this->_getClientId();
            $seller = $this->getSeller();
            if ($seller->isOnBoarded()) {
                global $currency;
                $locale = $this->getLocale(false);
                
                if ($this->hasOwnKeys()) {
                    $mid = '';
                } else {
                    $mid = '&merchant-id=' . $seller->payer_id;
                }

                $df = $ef = '';
                $tmp = $this->getFundings();

                if (!empty($tmp)) {
                    if (!empty($tmp['disabled'])) {
                        $_disabled = preg_replace(['/^paypal,/', '/,paypal,/', '/^pui,/', '/,pui,/', '/,pui$/'], ['', ',','', ',',''], $tmp['disabled']);
                        $df = '&disable-funding=' . $_disabled;
                    }
                    if (!empty($tmp['enabled'])) {
                        // paylater option is not shown in test more if paypal in allowed fundings :(
                        $_enabled = preg_replace(['/^paypal,/', '/,paypal,/', '/^pui,/', '/,pui,/', '/,pui$/'], ['', ',','', ',',''], $tmp['enabled']);
                        $ef = '&enable-funding=' . $_enabled;
                    }
                }
                $_comp = []; $components = '';
                if ($this->checkMessageOnProduct()) {
                    $_comp[] = 'messages';
                }
                if ($this->checkButtonOnProduct(true)) {
                    $_comp[] = 'buttons';
                }

                $clientToken = [];
                if ($this->checkCheckoutFields() && $this->currentPage() == 'checkout') {
                    $_comp = array_merge($_comp, ['buttons', 'payment-fields', 'marks', 'funding-eligibility']);
                    if ($cToken = $this->generateClientToken()) {
                        $clientToken = ['data-client-token' => $cToken];
                        $_comp[] = 'hosted-fields';
                    }
                    $this->addCss('.payment-method {width:100%} .payment-method td:nth-child(2) {text-align: right; vertical-align: top; padding-top: 0.5em;} .payment-method td:nth-child(1) .method {font-weight:bold;}'
                        . '.preloader { text-align:center} .preloader:before { /*content:"\f110";*/ content:">" -moz-animation:load8 1.5s infinite linear; -webkit-animation:load8 1.5s infinite linear; animation:load8 1.5s infinite linear; font-size:40px; line-height:40px; display:inline-block; overflow:hidden; width:40px; height:40px; /*font-family:FontAwesome*/}'
                        . '.payment-method .inline-block { display: inline-block;vertical-align: middle;padding: 0 5px 0 0;} .payment-method .paypal-mark span {overflow: visible;display: inline-block;padding-left: 5px;} .paypal-button-paypal_rest-agreement {text-align: justify; display: inline-block;} .payment-method .payment_class_paypal_rest .item-radio label {display: inline-block; }');
                }

                if (!empty($_comp)) {
                    $_comp = array_unique($_comp);
                    $components = '&components=' . implode(',', $_comp);
                }
                $commit = '&commit=' . $this->pp_commit;
                $buyerCountry = '';
                if ($this->getMode() !== 'Live') {
                    if (!$this->getCustomerAssigned() && defined('STORE_COUNTRY') && intval(STORE_COUNTRY)>0) {
                        $tmp = self::get_country_info_by_id(STORE_COUNTRY);
                        $buyerCountry = '&buyer-country=' . $tmp['countries_iso_code_2'];

                    } elseif(!empty($this->getBillingAddress())) {

                        $tmp = $this->getBillingAddress();
                        if (isset($tmp['country']['iso_code_2'])) {
                            $buyerCountry = '&buyer-country=' . $tmp['country']['iso_code_2'];
                        }
                    }
                }

                $this->addJsFile("https://www.paypal.com/sdk/js?client-id={$clid}{$mid}{$ef}{$components}{$commit}{$buyerCountry}&intent=" . ($this->_getIntent() == 'authorize' ? "authorize" : "capture") . '&currency=' . $currency . $df .'&locale=' . $locale . (0 && $this->debug ? '&debug=true' : ''), ['data-partner-attribution-id' => $this->getAttributionId() ] + $clientToken);

                $this->addJs("
                    function alertMessage(msg){
                        $('<div>' + msg + '</div>').dialog({modal: true});
                    }
                ");
                $this->addJs($this->getJavascript());
//                $this->addJsFile(\frontend\design\Info::themeFile('/js/bootstrap-datepicker.js'), ['position' => \common\components\View::POS_END]);
//                $this->addCssFile(\frontend\design\Info::themeFile('/css/bootstrap-datepicker.css'));

                $this->addJs($this->getJS());
            }
            return $this->html_buffer;
        }
    }

    /**
     * returns required HTML containers for buttons/messages.
     * @staticvar int $idx on product listing there could be several pay buttons
     * @return string  - HTML - payment buttons container, Css, JS
     */
    function checkout_initialization_method() {
        $ret = '';
        static $idx = 0;
        $js = $this->checkout_initialization_method_js();
        if ($this->checkButtonOnProduct(true)) {
            $idx++;
            $ret = '<div style="float:right; max-width: 200px;" class="paypal-button-container" id="paypal-button-container-' . $idx . '"></div>';
        }
        return $ret . $js;
    }

    /**
     * oscok
     * @return bool
     */
    public function checkMessageOnProduct() {
        return defined('MODULE_PAYMENT_PAYPAL_REST_PAY_LATER') && MODULE_PAYMENT_PAYPAL_REST_PAY_LATER == 'True';
    }

    /**
     * oscok
     * @return bool
     */
    public function checkButtonOnProduct($only=false) {
        global $PHP_SELF;
        return (
                basename($PHP_SELF) != FILENAME_PRODUCT_INFO ||
                (defined('MODULE_PAYMENT_PAYPAL_REST_BUY_IMMEDIATELLY') && MODULE_PAYMENT_PAYPAL_REST_BUY_IMMEDIATELLY != 'False')
            ) || (!$only && $this->checkMessageOnProduct());
    }


    function javascript_validation() {
        return false;
    }

    function selection() {
        global $payment;

        if (defined('MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT') && MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT !== 'False')
        {
            if ($this->checkCheckoutFields()) {

//2do                $this->registerCallback("paypalFieldsCallback");// common callback for all APMs, it'll call callback of selected method (if any)

                $methods = $apms = [];
                $apms = $this->getAPMTemplates();
                $js = '';

                foreach ($apms as $method => $apmTemplate) {
                    $extra = '';
                    $_title = self::getFundingTitle($method);
                    if (!empty($apms)) {
                        $js .= $this->addJs($this->getAPMJSCallback($method));
                        $extra .= '<div class="paypal-fields-container paypal-fields-' . $method . ' ' . $this->code . '_' . $method . '" style="display:none">' . $apmTemplate . '</div>';
                        $tmp = $this->getAPMJS($method);
                        if (!empty($tmp)) {
                            $extra .= '<script>' . $tmp . '</script>';
                        }
                    }
                    if ($method=='pui' && empty($apmTemplate)) {
                        continue;
                        $_title = defined('TEXT_PAY_UPON_INVOICE')?TEXT_PAY_UPON_INVOICE:'Kauf auf Rechnung';
                    }

                    $methods[] = [
                        'field' => tep_draw_radio_field('payment', $this->code . '_' . $method, ($this->code . '_' . $method == $payment), 'class="' . $this->code . '"'),
                        'title' => '<span class="inline-block"><span class="inline-block paypal-marks-icon paypal-marks-icon-' . $method . '"></span><span class="inline-block method">' .  $_title . '</span></span>' . $extra,
                        'hide' => (!in_array($method, ['paypal', 'pui'])), // show by JS as a lot could be not available for selected country
                    ];
                }
                if (count($methods)) {
//osc23 workaround to avoid checkout_payment.php update
                    $dr = <<<EOD
    $(document).ready(function(){
        window.initPPButtons();
        $('tr.moduleRow, tr.moduleRowSelected').on('click', function() {
            window.toggleSubFields_{$this->code}();
        });
    });
EOD;

                    $methods[count($methods)-1]['title'] .= $this->checkout_initialization_method_js() . '<script>hideFieldsTable();' . $dr . '</script>';
                    return array(
                        'id' => $this->code,
                        'module' => $this->public_title .  '<span id="pppTitleAfter"></span>',
                        'fields' => $methods,
                    );
                }

            } else {

                    $dr = <<<EOD
    $(document).ready(function(){
        window.initPPButtons();
        $('tr.moduleRow, tr.moduleRowSelected').on('click', function() {
            window.toggleSubFields_{$this->code}();
        });
    });
EOD;

                return array('id' => $this->code,
                             'module' => $this->public_title . $this->checkout_initialization_method_js() . '<script>' . $dr . '</script>');
            }
        }
    }

    function pre_confirmation_check() {
        global $messageStack;

        if (!$this->hasSession('partner_order_id')) {
            if (!defined('MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT') || strtolower(MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT) == 'false') {
                tep_redirect(tep_href_link(FILENAME_SHOPPING_CART, '', 'SSL'));
            }
            //usual checkout + APMs


        } else {

            if (tep_session_is_registered('ppartner_total_check')) {
                $messageStack->add('checkout_confirmation', '<span id="PayPalNotice">' . MODULE_PAYMENT_PAYPAL_REST_NOTICE_CHECKOUT_CONFIRMATION . '</span><script>$("#PayPalNotice").parent().css({backgroundColor: "#fcf8e3", border: "1px #faedd0 solid", color: "#a67d57", padding: "5px" });</script>', 'paypal');
            }

            $order->info['payment_method'] = '<img src="https://www.paypalobjects.com/webstatic/mktg/Logo/pp-logo-100px.png" border="0" alt="PayPal Logo" style="padding: 3px;" />';
        }
    }

    function confirmation() {
        global $comments;

        if (!isset($comments)) {
            $comments = null;
        }

        $confirmation = false;
/*2do
        if (empty($comments)) {
            $confirmation = array('fields' => array(array('title' => MODULE_PAYMENT_PAYPAL_REST_TEXT_COMMENTS,
                  'field' => tep_draw_textarea_field('ppecomments', 'soft', '60', '5', $comments))));
        }*/

        return $confirmation;
    }

    function process_button() {
        return false;
    }

    function before_process() {

        if (!$this->hasSession('partner_order_id')) {
            tep_redirect(tep_href_link(FILENAME_SHOPPING_CART, 'oid', 'SSL'));
        }

        /** @var order $order*/
        global $order;
        
        if (!is_object($order)) {
            if ($this->isInstance() ) {
                $order = $this->getOrderInstance();
            } else{
                $order = $this->createOrderInstance(true);
            }
        }

        $response = $this->getOrder($this->getSession('partner_order_id'));

        $direct = false;//direct payment from site - save not paid order with "redirected to " status
        if ($this->hasSession('partner_order_pending')) {
            $tmp = $this->getSession('partner_order_pending');
        }
        if (!empty($response->result->status) && in_array(strtoupper($response->result->status), ['PENDING_APPROVAL','COMPLETED', 'CREATED'])
            && $tmp == 'pay_upon_invoice') {
            $direct = true;
        }

        if (!empty($response->result->status) && strtoupper($response->result->status) == 'APPROVED') {
            if ($this->formatRaw($order->info['total']) != $response->result->purchase_units[0]->amount->value && !$this->hasSession('ppartner_total_check')) {
self::warning("pppbeforeprocess total changed ## " .print_r($this->formatRaw($order->info['total']) . '!= '. $response->result->purchase_units[0]->amount->value, true), 'TLDEBUG');
                $this->setSession('ppartner_total_check', true);

                tep_redirect(tep_href_link(FILENAME_CHECKOUT_CONFIRMATION, '', 'SSL'));
            }

            if ($this->hasSession('ppartner_total_check')) {
                try {
                    $response = $this->updateOrder($this->getSession('partner_order_id'));
                    $this->resetSession('ppartner_total_check');
                    if (!$response) {
                        //order wasn't updates on PP - restart (else incorrect amount will be authorized/captured
                        tep_redirect($this->getCheckoutUrl(['error_message' => MODULE_PAYMENT_PAYPAL_REST_RESTART], self::PAYMENT_PAGE));
                    }
                } catch (\Exception $ex) {
                    $this->sendDebugEmail($ex);
                    tep_redirect($this->getCheckoutUrl(['error_message' => MODULE_PAYMENT_PAYPAL_REST_RESTART], self::PAYMENT_PAGE));
                }
            }
        } elseif (!$direct) {
            tep_redirect($this->getCheckoutUrl(['error_message' => MODULE_PAYMENT_PAYPAL_REST_RESTART], self::PAYMENT_PAGE));
        }
        $_paid = false;
        //save order to get ID
        if ($direct || $this->saveOrderBefore() == 'Order') {
            /*2do PUI
            $invoiceId = $orderId = $this->saveOrderBySettings($direct);
             * 
             */
        } else {
            $orderId = $this->estimateOrderId();
            $invoiceId = $orderId . '-e-' . date('ymdHis');
        }

        //updateOrderInvoiceId on PayPal
        if (!empty($orderId) && ($orderId != intval($response->result->purchase_units[0]->invoice_id) || $invoiceId == $orderId) ) {
            try {
                $this->updateOrderInvoiceId($this->getSession('partner_order_id'), $invoiceId);
            } catch (\Exception $e) {
                //if not $direct not critical at all - estimated invoice id on PP Direct - no approval WH
                self::warning(" #### " .print_r($e->getMessage() . ' ' . $e->getTraceAsString(), true), 'exception_' . $this->code);
            }
        }

        if ($direct && $orderId) {
            $this->no_process($order);
            $order->notify_customer($order->getProductsHtmlForEmail(),[]);
            $this->no_process_after($order);
            tep_redirect(tep_href_link(FILENAME_CHECKOUT_SUCCESS, 'order_id=' . $orderId, 'SSL'));
        }

        try {

            if ($this->_getIntent() == 'authorize') {
                $presponse = $this->authorizeOrder($this->getSession('partner_order_id'));
                $rv = $presponse->result->purchase_units[0]->payments->authorizations[0]->id;
            } else {
                $presponse = $this->captureOrder($this->getSession('partner_order_id'));
                $rv = $presponse->result->purchase_units[0]->payments->captures[0]->id;
                $_paid = true;
            }
        } catch (\Exception $e) {
            self::error(" #### " .print_r($e->getMessage(), true), 'exception_' . $this->code);
        }

        if (!empty($presponse->result->status) && strtoupper($presponse->result->status) == "COMPLETED") {
            $this->setSession('partner_transaction_id', $rv);
            if ($_paid) { //captured - set paid order status w/o delay.
                $this->order_status = $this->paidOrderStatus();
                self::warning(" order_status #### " .print_r($this->order_status, true));
                if (!empty($this->order_status)) {
                    $order->info['order_status'] = $this->order_status;
                }
            }
/*2do order before payment
            if (!empty($orderId) && $this->saveOrderBefore() == 'Order') {
                $order = $this->getOrderInstanceWithId($orderId);
                if ($order) {
                    if ($this->order_status) {
                        \common\helpers\Order::setStatus($orderId, $this->order_status);
                        $order->info['order_status'] = $this->order_status;
                    }

                    $order->save_details();
                    $order->info['comments'] = '';

                    $order->notify_customer($order->getProductsHtmlForEmail(),[]);

                    $this->no_process($order);
                    $this->no_process_after($order);
                    tep_redirect(tep_href_link(FILENAME_CHECKOUT_SUCCESS, 'order_id=' . $order->info['orders_id'], 'SSL'));

                }
            }
            */
        } else {
            tep_redirect($this->getCheckoutUrl(['error_message' => (($this->_getIntent() == 'authorize')?MODULE_PAYMENT_PAYPAL_REST_RESTART_AUTHORIZE:MODULE_PAYMENT_PAYPAL_REST_RESTART_CAPTURE)], self::PAYMENT_PAGE));
        }
    }

    function after_process() {
        global $currencies;

//save auth/capture details in status history...
        $transactionID = $this->getSession('partner_transaction_id');
        $deferred = 0;
        /** @var PayPalHttp\HttpResponse $transaction */
        if ($this->_getIntent() == 'authorize') {
            $transaction = $this->getAuthorization($transactionID);
            $deferred = 1;
        } else {
            $transaction = $this->getCapture($transactionID);
        }

        if ($transaction) {

            global $insert_id;
            $order = $this->getOrderInstanceWithId($insert_id);

            $pp_result = [
              'Internal Order ID: ' . tep_output_string_protected($this->getSession('partner_order_id')),
              'Transaction ID: ' . tep_output_string_protected($transactionID),
              'Transactin Amount: ' . $currencies->format($transaction->result->amount->value, false, $order->info['currency'], $order->info['currency_value']),
              'Payment Status: ' . tep_output_string_protected($transaction->result->status),
              'Seller Protection: ' . tep_output_string_protected($transaction->result->seller_protection->status??'') .
              (is_array($transaction->result->seller_protection->dispute_categories)? ' - ' . implode(', ', $transaction->result->seller_protection->dispute_categories) :''),
            ];


            $sk = $this->getStatusCode($transaction);
            if (!empty($transaction->result->status_details->reason)) {
                $comment = $transaction->result->status_details->reason;
            } else {
                $comment = 'Customer\'s payment ' . $transaction->result->amount->value . $transaction->result->amount->currency_code;
            }

            $ret = $this->updatePaymentTransaction($transactionID,
                [
                  'fulljson' => json_encode($transaction),
                  'status_code' => $sk,
                  'status' => $transaction->result->status,
                  'amount' => $transaction->result->amount->value,
                  'comments' => $comment,
                  'date' => date('Y-m-d H:i:s', strtotime($transaction->result->update_time)),
                  'suborder_id' => 0,
                  'orders_id' => $order->info['orders_id'],
                  'deferred' => $deferred,
                // parent_transaction_id orders_id
            ]);

            $sql_data_array = array('orders_id' => $order->info['orders_id'],
              'orders_status_id' => $order->info['order_status'],
              'date_added' => 'now()',
              'customer_notified' => '0',
              'comments' => implode("\n", $pp_result));

            tep_db_perform(TABLE_ORDERS_STATUS_HISTORY, $sql_data_array);
        }
        $this->resetSession('partner_transaction_id');
        $this->resetSession('partner_order_id');
        //$this->resetSession('order');
        // array, probably with incorrect keys instead of ID
        $this->resetSession('sendto');
        $this->resetSession('billto');
    }


    function get_error() {
        return false;
    }

    function check() {
        if (!isset($this->_check)) {
            $check_query = tep_db_query("select configuration_value from " . TABLE_CONFIGURATION . " where configuration_key = 'MODULE_PAYMENT_PAYPAL_REST_STATUS'");
            $this->_check = tep_db_num_rows($check_query);
        }
        return $this->_check;
    }

    function install($parameter = null) {
        $params = $this->getParams();

        if (isset($parameter)) {
            if (isset($params[$parameter])) {
                $params = array($parameter => $params[$parameter]);
            } else {
                $params = array();
            }
        }

        foreach ($params as $key => $data) {
            $sql_data_array = array('configuration_title' => $data['title'],
              'configuration_key' => $key,
              'configuration_value' => (isset($data['value']) ? $data['value'] : ''),
              'configuration_description' => $data['desc'],
              'configuration_group_id' => '6',
              'sort_order' => '0',
              'date_added' => 'now()');

            if (isset($data['set_func'])) {
                $sql_data_array['set_function'] = $data['set_func'];
            }

            if (isset($data['use_func'])) {
                $sql_data_array['use_function'] = $data['use_func'];
            }

            tep_db_perform(TABLE_CONFIGURATION, $sql_data_array);
        }
    }

    function remove() {
        tep_db_query("delete from " . TABLE_CONFIGURATION . " where configuration_key in ('" . implode("', '", $this->keys()) . "')");
    }

    function keys() {
        $keys = array_keys($this->getParams());

        if ($this->check()) {
            foreach ($keys as $key) {
                if (!defined($key)) {
                    $this->install($key);
                }
            }
        }

        return $keys;
    }

    function getParams() {
        if (!defined('MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID')) {
            $check_query = tep_db_query("select orders_status_id from " . TABLE_ORDERS_STATUS . " where orders_status_name = 'PayPal [Transactions]' limit 1");

            if (tep_db_num_rows($check_query) < 1) {
                $status_query = tep_db_query("select max(orders_status_id) as status_id from " . TABLE_ORDERS_STATUS);
                $status = tep_db_fetch_array($status_query);

                $status_id = $status['status_id'] + 1;

                $languages = tep_get_languages();

                foreach ($languages as $lang) {
                    tep_db_query("insert into " . TABLE_ORDERS_STATUS . " (orders_status_id, language_id, orders_status_name) values ('" . $status_id . "', '" . $lang['id'] . "', 'PayPal [Transactions]')");
                }

                $flags_query = tep_db_query("describe " . TABLE_ORDERS_STATUS . " public_flag");
                if (tep_db_num_rows($flags_query) == 1) {
                    tep_db_query("update " . TABLE_ORDERS_STATUS . " set public_flag = 0 and downloads_flag = 0 where orders_status_id = '" . $status_id . "'");
                }
            } else {
                $check = tep_db_fetch_array($check_query);

                $status_id = $check['orders_status_id'];
            }
        } else {
            $status_id = MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID;
        }

        if (!defined('MODULE_PAYMENT_PAYPAL_REST_ORDER_STATUS_ID')) {
            $check_query = tep_db_query("select orders_status_id from " . TABLE_ORDERS_STATUS . " where orders_status_name = 'Processing' limit 1");

            if (tep_db_num_rows($check_query) < 1) {
                $paid_status_id = $status_id;
            } else {
                $check = tep_db_fetch_array($check_query);
                $paid_status_id = $check['orders_status_id'];
            }
        } else {
            $paid_status_id = MODULE_PAYMENT_PAYPAL_REST_ORDER_STATUS_ID;
        }

        $params = array(
          'MODULE_PAYMENT_PAYPAL_REST_EXTRA_CONFIG' => array('title' => 'API configuration',
            'desc' => '<a href="#" class="ppp_extraconfig_popup">Click here to configure</a>',
            'value' => '',
            //'use_func' => paypal_rest::extraConfigLink(),
            'set_func' => 'paypal_rest::extraConfigDescription('),

          'MODULE_PAYMENT_PAYPAL_REST_STATUS' => array('title' => 'Enable PayPal Rest module',
            'desc' => 'Do you want to accept PayPal Rest payments?',
            'value' => 'True',
            'set_func' => 'tep_cfg_select_option(array(\'True\', \'False\'), '),

          'MODULE_PAYMENT_PAYPAL_REST_VIRTUAL_GUEST' => array('title' => 'Disable PayPal for guest if only virtual products in cart',
            'desc' => 'PayPal does NOT provide billing address. Guest (express) order won\'t have any address',
            'value' => 'True',
            'set_func' => 'tep_cfg_select_option(array(\'True\', \'False\'), '),
/*
          'MODULE_PAYMENT_PAYPAL_REST_BUY_IMMEDIATELLY' => array('title' => 'Show PayPal buttons on product',
            'desc' => 'Allow to make PayPal purchase from product page',
            'value' => 'False',
            'set_func' => 'tep_cfg_select_option(array(\'Horizontal\', \'Vertical\', \'False\'), '),*/
          'MODULE_PAYMENT_PAYPAL_REST_SHOPPING_CART' => array('title' => 'Show PayPal button(s) on Shopping Cart',
            'desc' => 'Show PayPal buttons on Shopping Cart Page',
            'value' => 'Vertical',
            'set_func' => 'tep_cfg_select_option(array(\'Horizontal\', \'Vertical\', \'False\'), '),

          'MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT_LOGIN' => array('title' => 'Show PayPal button on Checkout Login',
            'desc' => 'Show PayPal buttons on Checkout Login Page',
            'value' => 'Horizontal',
            'set_func' => 'tep_cfg_select_option(array(\'Horizontal\', \'Vertical\', \'False\'), '),

          'MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT' => array('title' => 'Show PayPal button on Checkout',
            'desc' => 'Show PayPal buttons on Checkout Page',
            'value' => 'Vertical',
            'set_func' => 'tep_cfg_select_option(array(\'Horizontal\', \'Vertical\', \'Fields\', \'False\'), '),

/*
          'MODULE_PAYMENT_PAYPAL_REST_SELLER_ACCOUNT' => array('title' => 'Seller Account',
            'desc' => 'The email address of the seller account if no API credentials has been setup.',
            'value' => STORE_OWNER_EMAIL_ADDRESS),
          'MODULE_PAYMENT_PAYPAL_REST_API_USERNAME' => array('title' => 'API Username',
            'desc' => 'The username to use for the PayPal API service.'),
          'MODULE_PAYMENT_PAYPAL_REST_API_PASSWORD' => array('title' => 'API Password',
            'desc' => 'The password to use for the PayPal API service.'),
          'MODULE_PAYMENT_PAYPAL_REST_API_SIGNATURE' => array('title' => 'API Signature',
            'desc' => 'The signature to use for the PayPal API service.'),
          'MODULE_PAYMENT_PAYPAL_REST_ACCOUNT_OPTIONAL' => array('title' => 'PayPal Account Optional',
            'desc' => 'This must also be enabled in your PayPal account, in Profile > Website Payment Preferences.',
            'value' => 'False',
            'set_func' => 'tep_cfg_select_option(array(\'True\', \'False\'), '),
          'MODULE_PAYMENT_PAYPAL_REST_CHECKOUT_IMAGE' => array('title' => 'PayPal Buttons ',
            'desc' => 'Use static or dynamic Express Checkout image buttons. Dynamic images are used with PayPal campaigns.',
            'value' => 'Static',
            'set_func' => 'tep_cfg_select_option(array(\'Static\', \'Dynamic\'), '),
 */


          'MODULE_PAYMENT_PAYPAL_REST_PAY_LATER' => array('title' => 'Show PayPal Pay later info',
            'desc' => 'Displays Pay Later messaging for available offers. Restrictions apply. See terms and learn more on PayPal <a target=\'blank\' href=\'https://developer.paypal.com/docs/commerce-platforms/admin-panel/\'>PayPal</a> Add apropriate box.',
            'value' => 'False',
            'set_func' => 'tep_cfg_select_option(array(\'True\', \'False\'), '),
          'MODULE_PAYMENT_PAYPAL_REST_ORDER_BEFORE_PAYMENT' => array('title' => 'Save order before payment',
            'desc' => 'Save order before payment (slower checkout, exact invoice ID on PayPal, could be required if you have several orders a minute)',
            'value' => 'False',
            'set_func' => 'tep_cfg_select_option(array(\'True\', \'False\'), '),
          'MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_METHOD' => array('title' => 'Transaction Method',
            'desc' => 'The processing method to use for each transaction.',
            'value' => 'sale',
            'set_func' => 'tep_cfg_select_option(array(\'authorize\', \'sale\'), '),
          'MODULE_PAYMENT_PAYPAL_REST_ORDER_STATUS_ID' => array('title' => 'Set Pending Order Status',
            'desc' => 'Set the status of pending orders made with this payment module to this value',
            'value' => $status_id,
            'use_func' => 'tep_get_order_status_name',
            'set_func' => 'tep_cfg_pull_down_order_statuses('),
          'MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID' => array('title' => 'Set Paid Order Status',
            'desc' => 'Set the paid status of orders made with this payment module to this value',
            'value' => $paid_status_id,
            'use_func' => 'tep_get_order_status_name',
            'set_func' => 'tep_cfg_pull_down_order_statuses('),

          /*'MODULE_PAYMENT_PAYPAL_REST_CANCEL_ORDER_STATUS_ID' => array('title' => 'Set Cancelled Order Status',
            'desc' => 'Set the cancelled status of orders made with this payment module to this value',
            'value' => $status_id_ch,
            'use_func' => '\\common\\helpers\\Order::get_order_status_name',
            'set_func' => 'tep_cfg_pull_down_order_statuses('),*/
          'MODULE_PAYMENT_PAYPAL_REST_ZONE' => array('title' => 'Payment Zone',
            'desc' => 'If a zone is selected, only enable this payment method for that zone.',
            'value' => '0',
            'use_func' => 'tep_get_zone_class_title',
            'set_func' => 'tep_cfg_pull_down_zone_classes('),
          'MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER' => array('title' => 'Transaction Server',
            'desc' => 'Use the live or testing (sandbox) gateway server to process transactions?',
            'value' => 'Live',
            'set_func' => 'tep_cfg_select_option(array(\'Live\', \'Sandbox\'), '),
          'MODULE_PAYMENT_PAYPAL_REST_BUTTON_COLOR' => array('title' => 'Dynamic Button Color',
            'desc' => 'Color for Dynamic Button',
            'value' => 'gold',
            'set_func' => 'tep_cfg_select_option(array(\'gold\', \'blue\', \'silver\', \'white\', \'black\'), '),
          'MODULE_PAYMENT_PAYPAL_REST_BUTTON_SHAPE' => array('title' => 'Dynamic Button Shape',
            'desc' => 'Shape for Dynamic Button',
            'value' => 'pill',
            'set_func' => 'tep_cfg_select_option(array(\'pill\', \'rect\'), '),
          'MODULE_PAYMENT_PAYPAL_REST_BUTTON_SIZE' => array('title' => 'Dynamic Button Size',
            'desc' => 'Size for Dynamic Button',
            'value' => 'small',
            'set_func' => 'tep_cfg_select_option(array(\'small\', \'medium\', \'large\', \'responsive\'), '),
          'MODULE_PAYMENT_PAYPAL_REST_BUTTON_LABEL' => array('title' => 'Dynamic Button Label',
            'desc' => 'Label for Dynamic Button',
            'value' => 'checkout',
            'set_func' => 'tep_cfg_select_option(array(\'checkout\', \'pay\', \'buynow\', \'paypal\'), '),

          'MODULE_PAYMENT_PAYPAL_REST_BUTTON_FUNDING' => array('title' => 'Enable payment methods',
            'desc' => 'Enable funding payment methods',
            'value' => '',
            'set_func' => 'paypal_rest::multiOptionCheckbox(paypal_rest::possibleFundingArray(), '),


          'MODULE_PAYMENT_PAYPAL_REST_PROXY' => array('title' => 'Proxy Server',
            'desc' => 'Send API requests through this proxy server. (host:port, eg: 123.45.67.89:8080 or proxy.example.com:8080 or http://usr:passwd@123.45.67.89:8080)'),
          'MODULE_PAYMENT_PAYPAL_REST_DEBUG_EMAIL' => array('title' => 'Debug E-Mail Address',
            'desc' => 'All parameters of an invalid transaction will be sent to this email address.'),
          'MODULE_PAYMENT_PAYPAL_REST_SORT_ORDER' => array('title' => 'Sort order of display',
            'desc' => 'Sort order of display. Lowest is displayed first.',
            'value' => '0'));

        return $params;
    }

    function sendTransactionToGateway($url, $parameters) {
        self::sendRequest($url, $parameters);
    }
    
// format prices without currency formatting
    function format_raw($number, $currency_code = '', $currency_value = '') {
        global $currencies, $currency;

        if (empty($currency_code) || !$currencies->is_set($currency_code)) {
            $currency_code = $currency;
        }

        if (empty($currency_value) || !is_numeric($currency_value)) {
            $currency_value = $currencies->currencies[$currency_code]['value'];
        }

        return number_format(tep_round($number * $currency_value, $currencies->currencies[$currency_code]['decimal_places']), $currencies->currencies[$currency_code]['decimal_places'], '.', '');
    }

    function getPalDetails() {
        if (MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER == 'Live') {
            $api_url = 'https://api-3t.paypal.com/nvp';
        } else {
            $api_url = 'https://api-3t.sandbox.paypal.com/nvp';
        }

        $params = array('VERSION' => $this->api_version,
          'METHOD' => 'GetPalDetails',
          'USER' => MODULE_PAYMENT_PAYPAL_REST_API_USERNAME,
          'PWD' => MODULE_PAYMENT_PAYPAL_REST_API_PASSWORD,
          'SIGNATURE' => MODULE_PAYMENT_PAYPAL_REST_API_SIGNATURE);

        $post_string = '';

        foreach ($params as $key => $value) {
            $post_string .= $key . '=' . urlencode(utf8_encode(trim($value))) . '&';
        }

        $post_string = substr($post_string, 0, -1);

        $response = $this->sendTransactionToGateway($api_url, $post_string);
        $response_array = array();
        parse_str($response, $response_array);

        if (!isset($response_array['PAL'])) {
            $this->sendDebugEmail($response_array);
        }

        return $response_array;
    }

    function setExpressCheckout($parameters) {
        if (MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER == 'Live') {
            $api_url = 'https://api-3t.paypal.com/nvp';
        } else {
            $api_url = 'https://api-3t.sandbox.paypal.com/nvp';
        }

        $params = array('VERSION' => $this->api_version,
          'METHOD' => 'SetExpressCheckout',
          'PAYMENTREQUEST_0_PAYMENTACTION' => ((MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_METHOD == 'Sale') || (!tep_not_null(MODULE_PAYMENT_PAYPAL_REST_API_USERNAME)) ? 'Sale' : 'Authorization'),
          'RETURNURL' => tep_href_link('ext/modules/payment/paypal_rest/index.php', 'osC_Action=retrieve', 'SSL', true, false),
          'CANCELURL' => tep_href_link('ext/modules/payment/paypal_rest/index.php', 'osC_Action=cancel', 'SSL', true, false),
          'BRANDNAME' => STORE_NAME,
          'SOLUTIONTYPE' => (MODULE_PAYMENT_PAYPAL_REST_ACCOUNT_OPTIONAL == 'True') ? 'Sole' : 'Mark');

        if (tep_not_null(MODULE_PAYMENT_PAYPAL_REST_API_USERNAME)) {
            $params['USER'] = MODULE_PAYMENT_PAYPAL_REST_API_USERNAME;
            $params['PWD'] = MODULE_PAYMENT_PAYPAL_REST_API_PASSWORD;
            $params['SIGNATURE'] = MODULE_PAYMENT_PAYPAL_REST_API_SIGNATURE;
        } else {
            $params['SUBJECT'] = MODULE_PAYMENT_PAYPAL_REST_SELLER_ACCOUNT;
        }

        if (is_array($parameters) && !empty($parameters)) {
            $params = array_merge($params, $parameters);
        }

        $post_string = '';

        foreach ($params as $key => $value) {
            $post_string .= $key . '=' . urlencode(utf8_encode(trim($value))) . '&';
        }

        $post_string = substr($post_string, 0, -1);

        $response = $this->sendTransactionToGateway($api_url, $post_string);
        $response_array = array();
        parse_str($response, $response_array);

        if (($response_array['ACK'] != 'Success') && ($response_array['ACK'] != 'SuccessWithWarning')) {
            $this->sendDebugEmail($response_array);
        }

        return $response_array;
    }

    function getExpressCheckoutDetails($token) {
        if (MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER == 'Live') {
            $api_url = 'https://api-3t.paypal.com/nvp';
        } else {
            $api_url = 'https://api-3t.sandbox.paypal.com/nvp';
        }

        $params = array('VERSION' => $this->api_version,
          'METHOD' => 'GetExpressCheckoutDetails',
          'TOKEN' => $token);

        if (tep_not_null(MODULE_PAYMENT_PAYPAL_REST_API_USERNAME)) {
            $params['USER'] = MODULE_PAYMENT_PAYPAL_REST_API_USERNAME;
            $params['PWD'] = MODULE_PAYMENT_PAYPAL_REST_API_PASSWORD;
            $params['SIGNATURE'] = MODULE_PAYMENT_PAYPAL_REST_API_SIGNATURE;
        } else {
            $params['SUBJECT'] = MODULE_PAYMENT_PAYPAL_REST_SELLER_ACCOUNT;
        }

        $post_string = '';

        foreach ($params as $key => $value) {
            $post_string .= $key . '=' . urlencode(utf8_encode(trim($value))) . '&';
        }

        $post_string = substr($post_string, 0, -1);

        $response = $this->sendTransactionToGateway($api_url, $post_string);
        $response_array = array();
        parse_str($response, $response_array);

        if (($response_array['ACK'] != 'Success') && ($response_array['ACK'] != 'SuccessWithWarning')) {
            $this->sendDebugEmail($response_array);
        }

        return $response_array;
    }

    function doExpressCheckoutPayment($parameters) {
        if (MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER == 'Live') {
            $api_url = 'https://api-3t.paypal.com/nvp';
        } else {
            $api_url = 'https://api-3t.sandbox.paypal.com/nvp';
        }

        $params = array('VERSION' => $this->api_version,
          'METHOD' => 'DoExpressCheckoutPayment',
          'PAYMENTREQUEST_0_PAYMENTACTION' => ((MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_METHOD == 'Sale') || (!tep_not_null(MODULE_PAYMENT_PAYPAL_REST_API_USERNAME)) ? 'Sale' : 'Authorization'),
          'BUTTONSOURCE' => 'OSCOM23_EC');

        if (tep_not_null(MODULE_PAYMENT_PAYPAL_REST_API_USERNAME)) {
            $params['USER'] = MODULE_PAYMENT_PAYPAL_REST_API_USERNAME;
            $params['PWD'] = MODULE_PAYMENT_PAYPAL_REST_API_PASSWORD;
            $params['SIGNATURE'] = MODULE_PAYMENT_PAYPAL_REST_API_SIGNATURE;
        } else {
            $params['SUBJECT'] = MODULE_PAYMENT_PAYPAL_REST_SELLER_ACCOUNT;
        }

        if (is_array($parameters) && !empty($parameters)) {
            $params = array_merge($params, $parameters);
        }

        $post_string = '';

        foreach ($params as $key => $value) {
            $post_string .= $key . '=' . urlencode(utf8_encode(trim($value))) . '&';
        }

        $post_string = substr($post_string, 0, -1);

        $response = $this->sendTransactionToGateway($api_url, $post_string);
        $response_array = array();
        parse_str($response, $response_array);

        if (($response_array['ACK'] != 'Success') && ($response_array['ACK'] != 'SuccessWithWarning')) {
            $this->sendDebugEmail($response_array);
        }

        return $response_array;
    }

    function getProductType($id, $attributes) {
        foreach ($attributes as $a) {
            $virtual_check_query = tep_db_query("select pad.products_attributes_id from " . TABLE_PRODUCTS_ATTRIBUTES . " pa, " . TABLE_PRODUCTS_ATTRIBUTES_DOWNLOAD . " pad where pa.products_id = '" . (int) $id . "' and pa.options_values_id = '" . (int) $a['value_id'] . "' and pa.products_attributes_id = pad.products_attributes_id limit 1");

            if (tep_db_num_rows($virtual_check_query) == 1) {
                return 'Digital';
            }
        }

        return 'Physical';
    }

    function sendDebugEmail($response = array()) {
        global $HTTP_POST_VARS, $HTTP_GET_VARS;

        if (tep_not_null(MODULE_PAYMENT_PAYPAL_REST_DEBUG_EMAIL)) {
            $email_body = '';

            if (!empty($response)) {
                $email_body .= 'RESPONSE:' . "\n\n" . print_r($response, true) . "\n\n";
            }

            if (!empty($HTTP_POST_VARS)) {
                $email_body .= '$HTTP_POST_VARS:' . "\n\n" . print_r($HTTP_POST_VARS, true) . "\n\n";
            }

            if (!empty($HTTP_GET_VARS)) {
                $email_body .= '$HTTP_GET_VARS:' . "\n\n" . print_r($HTTP_GET_VARS, true) . "\n\n";
            }

            if (!empty($email_body)) {
                tep_mail('', MODULE_PAYMENT_PAYPAL_REST_DEBUG_EMAIL, 'PayPal Express Checkout Debug E-Mail', trim($email_body), STORE_OWNER, STORE_OWNER_EMAIL_ADDRESS);
            }
        }
    }


    /**
     * seller is boarded with current partner OR boarding process
     * @return bool
     */
    protected function _isReady() {
        if (!$this->getInstaller()->check()) {
            return false;
        }
        if (in_array(($_GET['action']??false), ['processOnBoard', 'checkOnBoarded', 'sellerDetails', 'deleteSeller'])) {
            return true;
        }
        $platformId = $this->getPlatformId();
        $seller = $this->getSeller($platformId);
        return ($seller->isOnBoarded() && $this->getPartnerId() == $seller->partner_id);
    }

    public static function getMode() {
        $ret = '';
        if (defined('MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER') ) {
            $ret = MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER;
        }
        return $ret;
    }

    public function boardViaLink() {
        if ($this->getMode() == 'Live') {
            $ret = self::PARTNER_APP_CLIENT_SECRET;
        } else {
            $ret = self::PARTNER_APP_SANDBOX_CLIENT_SECRET;
        }
        return empty($ret);
    }

    public static function sendRequest($url, $data, $method='POST', $headers = []) {
        $server = parse_url($url);

        if (!isset($server['port'])) {
            $server['port'] = ($server['scheme'] == 'https') ? 443 : 80;
        }

        if (!isset($server['path'])) {
            $server['path'] = '/';
        }

        $curl = curl_init($server['scheme'] . '://' . $server['host'] . $server['path'] . (isset($server['query']) ? '?' . $server['query'] : ''));
        curl_setopt($curl, CURLOPT_PORT, $server['port']);
        
        if (empty($headers) || !is_array($headers)) {
            curl_setopt($curl, CURLOPT_HEADER, false);
        } else {
            $tmp = [];
            foreach($headers as $k => $v) {
                $tmp[] = $k . ': ' . $v;
            }
            curl_setopt($curl, CURLOPT_HTTPHEADER, $tmp);
        }
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_FORBID_REUSE, true);
        curl_setopt($curl, CURLOPT_FRESH_CONNECT, true);
        if ($method=='POST') {
            curl_setopt($curl, CURLOPT_POST, true);

            if (is_array($data)) {
                curl_setopt($curl, CURLOPT_POSTFIELDS, http_build_query($data));
            } else {
                curl_setopt($curl, CURLOPT_POSTFIELDS, $data);
            }

        } else {
            curl_setopt( $curl, CURLOPT_CUSTOMREQUEST, 'GET' );
            if (!empty($data)) {
                curl_setopt($curl, CURLOPT_POSTFIELDS, $data);
            }
        }

        if (MODULE_PAYMENT_PAYPAL_REST_VERIFY_SSL == 'True') {
            curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, true);
            curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, 2);

            if (file_exists(DIR_FS_CATALOG . 'ext/modules/payment/paypal/paypal.com.crt')) {
                curl_setopt($curl, CURLOPT_CAINFO, DIR_FS_CATALOG . 'ext/modules/payment/paypal/paypal.com.crt');
            } elseif (file_exists(DIR_FS_CATALOG . 'includes/cacert.pem')) {
                curl_setopt($curl, CURLOPT_CAINFO, DIR_FS_CATALOG . 'includes/cacert.pem');
            }
        } else {
            curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
        }

        if (defined('MODULE_PAYMENT_PAYPAL_REST_PROXY') && tep_not_null(MODULE_PAYMENT_PAYPAL_REST_PROXY)) {

            $urlParts = parse_url(MODULE_PAYMENT_PAYPAL_REST_PROXY);
            if (!empty($urlParts) && array_key_exists("host", $urlParts)) {
                curl_setopt($curl, CURLOPT_PROXY, $urlParts["host"]);

                if (isset($urlParts["port"])) {
                    curl_setopt($curl, CURLOPT_PROXYPORT, $urlParts["port"]);
                }
                if (isset($urlParts["user"])) {
                    curl_setopt($curl, CURLOPT_PROXYUSERPWD, $urlParts["user"] . ":" . $urlParts["pass"]);
                }
            }
        }
//curl_setopt($curl, CURLOPT_VERBOSE, true); $streamVerboseHandle = fopen('php://temp', 'w+'); curl_setopt($curl, CURLOPT_STDERR, $streamVerboseHandle);
        $result = curl_exec($curl);

        curl_close($curl);
//rewind($streamVerboseHandle); $verboseLog = stream_get_contents($streamVerboseHandle); echo " #### <PRE>"  . __FILE__ .':' . __LINE__ . ' ' . print_r($verboseLog, true) ."</PRE>";

        return $result;
    }

/**
 * 
 * @return bool
 */
    public function checkCheckoutFields() {
        return (
                in_array($this->currentPage(), array('cart', 'checkout')) &&
                defined('MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT') && MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT == 'Fields'
            );
    }



/**
 * JS to render PP buttons and messages.
 * @return string JS
 */
    public function getJavascript() {
        global $currency;
        $locale = $this->getLocale(false);

        if (tep_session_is_registered('ppartner_total_check') && $this->currentPage() == 'confirmation' ) return;// returned after paypal to confirm final amount


        if ($this->currentPage() != 'catalog' || $this->checkButtonOnProduct(true)) {
            
            $jsContainerAndButtonsInit = <<< EOD
window.initPPButtons = function() {
        var mainSubmits = $('input[name=payment]:first').parents('form').find('submit, button[type=submit]');
        $(mainSubmits).after('<div class="paypal-button-container"></div>');
        window.pppRenderButtons();
        if ($('input[name=payment].{$this->code}:checked, input[name=payment][value={$this->code}]:checked').length) {
            $(mainSubmits).hide();
        } else {
            $('.paypal-button-container').hide();
        }
};

EOD;



            if ($this->checkCheckoutFields() && $this->isCheckout()) {
                //apm + CC
                $size = defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_SIZE') ? MODULE_PAYMENT_PAYPAL_REST_BUTTON_SIZE : 'small';
                $color = defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_COLOR') ? MODULE_PAYMENT_PAYPAL_REST_BUTTON_COLOR : 'gold';
                $shape = defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_SHAPE') ? MODULE_PAYMENT_PAYPAL_REST_BUTTON_SHAPE : 'pill';
                $label = defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_LABEL') ? MODULE_PAYMENT_PAYPAL_REST_BUTTON_LABEL : 'checkout';
                $layout = strtolower($this->currentPageButtonsLayout());
                $generalError = defined('MODULE_PAYMENT_PAYPAL_REST_GENERAL_ERROR') ? MODULE_PAYMENT_PAYPAL_REST_GENERAL_ERROR : "Can't create order";
                $generalError = htmlspecialchars($generalError);
                
                $createUrl = $this->createAbsoluteUrl(["ext/modules/payment/paypal_rest/index.php", 'action' => 'createOrder', 'page' =>$this->currentPage()]); //must return OrderId (from paypal)
                $retrieveUrl = $this->createAbsoluteUrl(["ext/modules/payment/paypal_rest/index.php", 'action' => 'retrieveOrder']);
                $shippingUrl = $this->createAbsoluteUrl(["ext/modules/payment/paypal_rest/index.php", 'action' => 'patchOrder']);
                
                $customersDetails = '';
                //$puoi_agreement = '';
                //if ($this->billing['country']['countries_iso_code_2'] == 'DE') {                }
                $puoi_agreement = str_replace(["\'", "\n"], ["\\'", ''], MODULE_PAYMENT_PAYPAL_REST_TEXT_UPON_INVOCE);


                //'frmCheckoutConfirm' : 'frmCheckout'
                if (!$this->getCustomerAssigned()) {
                    $customersDetails = ",{
                                        method: 'POST',
                                        headers: {'Content-Type':'application/x-www-form-urlencoded'},
                                        body: $('form[name=checkout_payment]').serialize()
                                    }";
                }
                
                $jsButtonsSettings = <<< EOD

                var pppButtonsParams = {
                    locale: '{$locale}',
                    style: {
                        size: '{$size}',
                        color: '{$color}',
                        shape: '{$shape}',
                        label: '{$label}',
                        layout: '{$layout}',
                    },
                    createOrder: function (data, actions) {
                                return fetch('{$createUrl}' {$customersDetails})
                          .then(function(res) {
                                return res.json();
                          }).then(function(data) {

                                try {
                                    if (data.hasOwnProperty('error') && data.hasOwnProperty('message') ) {
                                        alertMessage(data.message);
                                    }
                                    //return '';
                                } catch ( e ) {
                                  console.log(e);
                                  return false
                                }
                                return data.id;
                          });
                    },
                    onApprove: function (data, actions) {
                        $('body').append('<div class="hide-page" style="align-items: center;background-color: rgba(255, 255, 255, 0.7);height:100%; justify-content: center; left:0; position: fixed; top:0; width:100%; z-index: 100000; display:flex;"><div class="preloader"></div></div>');

                        return fetch('{$retrieveUrl}&id=' + data.orderID, {
                            method: 'POST'
                        }).then(function(res) {
                            return res.json();
                        }).then(function(res) {
                            if (!res.ok) {
                                $('.hide-page').remove();
                                alertMessage(res.error);
                            } else {
                                window.location.href = res.url;
                            }
                        });
                    },
                    onShippingChange: function(data,actions) {
                        return fetch('{$shippingUrl}', {
                                        method: 'POST',
                                        headers: {
                                           'Content-Type': 'application/x-www-form-urlencoded',
                                        },
                                        body: JSON.stringify(data)
                                    }).then(function(res) {
                                        return res.json();
                                    }).then(function(data) {
                                        if (data.length == 0) {
                                          return actions.reject();
                                        } else if (data.same == 1) {
                                          return actions.resolve();
                                        }
                                        return actions.order.patch(data);
                                    });
                    },
/*product page                    onClick: function(data, actions) {
                        var pppBtn = $('.paypal-button-container');
                        if (pppBtn && pppBtn.length>0){
                            var  that = pppBtn[0];
                            var theForm = $(that).parents('form');
                            if (theForm.attr('name') == 'cart_quantity' || theForm.hasClass('form-buy')){

                                var _data = theForm.serializeArray();
                                _data.push({name: 'popup', value: 'true'});
                                _data.push({name: 'purchase', value: '{$this->code}'});

                                $.ajax({
                                    url: theForm.attr('action'),
                                    data: _data,
                                    async: false,
                                    type: theForm.attr('method'),
                                    success: function(data){
                                        //just wait for adding product to cart
                                        }
                                });
                            }
                        }
                    },/*product page  */
                    onError: function(err){
                            if (false && err != '') {
                                 alert(err);
                            } else {
                                 alert("{$generalError}");
                            }
                    },
                    onCancel: function(err){

                    }
                };
EOD;


                return <<<EOD

                {$jsButtonsSettings}

window.pppRenderButtons = function() {

    document.querySelectorAll('.paypal-button-container').forEach(function(selector) {
        paypal.getFundingSources().forEach(function(fundingSource) {
            var pbParams = pppButtonsParams;
            if (fundingSource != 'paypal') {
                pbParams.style = {};
            }
            pbParams.fundingSource =  fundingSource;
            var button = paypal.Buttons(pbParams);
            if (button.isEligible()) {
                var container;
                container = document.createElement('span');
                container.classList.add('paypal-button-{$this->code}_' + fundingSource);
                selector.appendChild(container);
                button.render(container);
            }
        });
        if ($('.item-payment.paypal_rest_pui').length>0) {
            var container;
            container = document.createElement('span');
            container.classList.add('paypal-button-{$this->code}_pui');
            var agreement = btncontainer = '';
            agreement = document.createElement('span');
            btncontainer = document.createElement('span');
            container.classList.add('paypal-button-{$this->code}-agreement');
            agreement.innerHTML = '{$puoi_agreement}';

            container.appendChild(agreement);
            container.appendChild(btncontainer);

            selector.appendChild(container);
            container = btncontainer;
        }

    });
}

function paypalFieldsCallback(){

    if (paymentCollection.form) {

        var bodyData = '';
        if (paymentCollection.form) {
            bodyData = $(paymentCollection.form).serialize();
        }
        var method = paymentCollection.getCurrentPayment();
        method = method.substring(method.lastIndexOf('_')+1);

        if (typeof window['{$this->code}_callback_' + method] == 'function'){
            try {
                window['{$this->code}_callback_' + method]();
            } catch (e ) {
                console.log(e);
            }
        }


    }
    return false;

};

            if ($('.paypal-button-container').length>0){
                window.pppRenderButtons();
            }


            $(window).on('checkout_worker_complete', function (event) {
                var price = 0;
                try {
                    price = parseFloat($('.price-row.total.ot_total input.ot_total_clear').val()) || 0;
                    price = price.toFixed(2);
                } catch (e) { price = 0; }

                $('.pp-pay-later-message').attr('data-pp-amount', price);
            });

            $jsContainerAndButtonsInit
EOD;
/*
        fetch("$whSaveTmpOrderURL", {
                    method: "POST",
                    headers: {
                      "Content-Type": "application/json"
                    },
                    body: bodyData
                }).then(function(result) {
                    return result.json();
                })
                .then(function(data) {
                    //!payWithCard finish calback
                    if (!data || !data.clientSecret) {
                        if (data && data.error) {
                            showError(data.error);
                        } else {
                            showError('$general_error');
                        }
                    } else {
                        payWithCard(paymentCollection.stripes.stripe, paymentCollection.stripes.card, data.clientSecret, data.token);
                    }

                });*/
            } else {
                //paypal buttons with single payment module select

                $size = defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_SIZE') ? MODULE_PAYMENT_PAYPAL_REST_BUTTON_SIZE : 'small';
                $color = defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_COLOR') ? MODULE_PAYMENT_PAYPAL_REST_BUTTON_COLOR : 'gold';
                $shape = defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_SHAPE') ? MODULE_PAYMENT_PAYPAL_REST_BUTTON_SHAPE : 'pill';
                $label = defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_LABEL') ? MODULE_PAYMENT_PAYPAL_REST_BUTTON_LABEL : 'checkout';
                $layout = strtolower($this->currentPageButtonsLayout());
                $generalError = defined('MODULE_PAYMENT_PAYPAL_REST_GENERAL_ERROR') ? MODULE_PAYMENT_PAYPAL_REST_GENERAL_ERROR : "Can't create order";
                $generalError = htmlspecialchars($generalError);
                $createUrl = $this->createAbsoluteUrl(["ext/modules/payment/paypal_rest/index.php", 'action' => 'createOrder', 'page' => $this->currentPage()]); //must return OrderId (from paypal)
                $retrieveUrl = $this->createAbsoluteUrl(["ext/modules/payment/paypal_rest/index.php", 'action' => 'retrieveOrder', 'page' =>$this->currentPage()]);
                $shippingUrl = $this->createAbsoluteUrl(["ext/modules/payment/paypal_rest/index.php", 'action' => 'patchOrder']);
                $customersDetails = '';
                //'frmCheckoutConfirm' : 'frmCheckout'
                if (!$this->getCustomerAssigned()) {
                    //2do opc
                    $customersDetails = ",{
                                        method: 'POST',
                                        headers: {'Content-Type':'application/x-www-form-urlencoded'},
                                        body: $('form[name=checkout_payment]').serialize()
                                    }";
                }

                $jsButtonsSettings = <<< EOD

                var pppButtonsParams = {
                    locale: '{$locale}',
                    style: {
                        size: '{$size}',
                        color: '{$color}',
                        shape: '{$shape}',
                        label: '{$label}',
                        layout: '{$layout}',
                    },
                    createOrder: function (data, actions) {
                                return fetch('{$createUrl}' {$customersDetails})
                          .then(function(res) {
                                return res.json();
                          }).then(function(data) {

                                try {
                                    if (data.hasOwnProperty('error') && data.hasOwnProperty('message') ) {
                                        alertMessage(data.message);
                                    }
                                    //return '';
                                } catch ( e ) {
                                  console.log(e);
                                  return false
                                }
                                return data.id;
                          });
                    },
                    onApprove: function (data, actions) {
                        $('body').append('<div class="hide-page" style="align-items: center;background-color: rgba(255, 255, 255, 0.7);height:100%; justify-content: center; left:0; position: fixed; top:0; width:100%; z-index: 100000; display:flex;"><div class="preloader"></div></div>');
                        return fetch('{$retrieveUrl}&id=' + data.orderID, {
                            method: 'POST'
                        }).then(function(res) {
                            return res.json();
                        }).then(function(res) {
                            if (!res.ok) {
                                $('.hide-page').remove();
                                alertMessage(res.error);
                            } else {
                                window.location.href = res.url;
                            }
                        });
                    },
                    onShippingChange: function(data,actions) {
    //console.log("onShippingChange", data, actions);
                        return fetch('{$shippingUrl}', {
                                        method: 'POST',
                                        headers: {
                                           'Content-Type': 'application/x-www-form-urlencoded',
                                        },
                                        body: JSON.stringify(data)
                                    }).then(function(res) {
                                        return res.json();
                                    }).then(function(data) {
                                        if (data.length == 0) {
                                          return actions.reject();
                                        } else if (data.same == 1) {
                                          return actions.resolve();
                                        }
                                        return actions.order.patch(data);
                                    });
                    },
                    onClick: function(data, actions) {
                        var pppBtn = $('.paypal-button-container');
                        if (pppBtn && pppBtn.length>0){
                            var  that = pppBtn[0];
                        var theForm = $(that).parents('form');
                        if (theForm.attr('name') == 'cart_quantity' || theForm.hasClass('form-buy')){
                        /*
                            //theForm.removeClass('set-popup');
                            theForm.append('<input type="hidden" name="purchase" value="{$this->code}">').submit();
                            $(".popup-box:last").trigger("popup.close");$(".hide-page").remove();
                        */
                            var _data = theForm.serializeArray();
                            _data.push({name: 'popup', value: 'true'});
                            _data.push({name: 'purchase', value: '{$this->code}'});

                            $.ajax({
                                url: theForm.attr('action'),
                                data: _data,
                                async: false,
                                type: theForm.attr('method'),
                                success: function(data){
                                    //just wait for adding product to cart
                                    }
                            });
                            }
                        }
                    },
                    onError: function(err){
                            if (false && err != '') {
                                 alert(err);
                            } else {
                                 alert("{$generalError}");
                            }
                    },
                    onCancel: function(err){

                    }
                };
EOD;

                return <<<EOD

                {$jsButtonsSettings}

window.pppRenderButtons = function() {
    document.querySelectorAll('.paypal-button-container').forEach(function(selector) {
        paypal.Buttons(pppButtonsParams).render(selector);
    });
}


            if ($('.paypal-button-container').length>0){
                window.pppRenderButtons();
            }

            $(window).on('checkout_worker_complete', function (event) {
                var price = 0;
                try {
                    price = parseFloat($('.price-row.total.ot_total input.ot_total_clear').val()) || 0;
                    price = price.toFixed(2);
                } catch (e) { price = 0; }

                $('.pp-pay-later-message').attr('data-pp-amount', price);
            });

            $jsContainerAndButtonsInit
EOD;
            }
        }
    }


    /**
     * checkout - functions for onChange payment event (depends on current state etc)
     * @return string JS
     */
    public function getJS() {
        global $PHP_SELF;
        if (!in_array(basename($PHP_SELF), array(FILENAME_CHECKOUT_PAYMENT))) return; // not needed on shopping cart and product pages
            $quoteCheckout = 0;
            $unCheckedJSString = <<<EOD
                $('form[name="checkout_payment"] button[type="submit"]').show();
                $('form[name="checkout_payment"] button[type="submit"]').prop('disabled', false);
                $('[data-name^="checkout"] .paypal-button-container').hide();
                $('.paypal-button-container').parents('.add-buttons').prev('.or-text').hide();
EOD;
        if ($this->hasSession('ppartner_total_check')) {
            $this->resetSession('ppartner_total_check');
        }

        //usual checkout
        $checkedJSString = <<<EOD
            $('form[name="checkout_payment"] button[type="submit"]').hide();
            $('form[name="checkout_payment"] button[type="submit"]').prop('disabled', true);
            $('.paypal-button-container').show();
            //$('.paypal-button-container').parents('.add-buttons').prev('.or-text').show();
EOD;
        $unCheckedJSString = <<<EOD
            $('form[name="checkout_payment"] button[type="submit"]').show();
            $('form[name="checkout_payment"] button[type="submit"]').prop('disabled', false);
            $('.paypal-button-container').hide();
            //$('[data-name^="checkout"] .paypal-button-container').hide();
            //$('.paypal-button-container').parents('.add-buttons').prev('.or-text').hide();
EOD;

        $selectionInitJs = "

        var theForm = $('input[name=payment]:first').parents('form');

        window.paymentCollection = {
            form: theForm,
            getCurrentPayment:function(){
                return $('input[name=payment]:checked', this.form).val()||$('input[name=payment]:hidden', this.form).val();
            }
        };
        $(theForm).addClass('frmCheckout');
        $(theForm).on('submit', function() {
            var module = $('input[name=payment]:checked').val();
            if (module.indexOf('paypal_rest_') == 0) {
                return paypalFieldsCallback();
            }
            return true;
        });

        function hideFieldsTable() {
            var mainPPP = $('#pppTitleAfter').parents('tr');
            var fieldsTable = $(mainPPP).siblings().find('table');
            try  {
                $(fieldsTable).addClass('payment-method');
                $('tr', fieldsTable).addClass('moduleRow');
                $('tr', fieldsTable).addClass('item-radio');
                $('tr', fieldsTable).addClass('item-payment');
                $('tr', fieldsTable).each(function(){
                    $(this).addClass($('input[name=payment]', this).val());
                });


                $('tr:not(:first)', fieldsTable).hide();
                $('tr', fieldsTable).on('click', function(){
                    $('tr.moduleRowSelected').removeClass('moduleRowSelected').addClass('moduleRow');
                    $(this).addClass('moduleRowSelected');
                    $('input[name=payment]').prop('checked', false);
                    $('input[name=payment]', this).prop('checked', true);
                });
                $('tr', fieldsTable).on('mouseover', function(){
                    $(this).removeClass('moduleRow').addClass('moduleRowOver');
                });
                $('tr', fieldsTable).on('mouseout', function(){
                    $(this).removeClass('moduleRowOver').addClass('moduleRow');
                });


                $(mainPPP).remove();


            } catch ( e ) { console.log(e); }
//debug console.log(fieldsTable);

        }
        ";
        $apmsWithFields = array_keys(array_filter($this->getAPMTemplates()));
        $apmsWithFieldsStr = 'var apmsWithFields = ["' . implode('","', $apmsWithFields) . '"];';
        $apmsWithoutOwnButton = 'var apmsWithoutOwnButton = ["' . implode('","', $this->apmsWithoutOwnButton()) . '"];';
        $apmsWithAgreement = 'var apmsWithAgreement = ["' . implode('","', $this->apmsWithAgreement()) . '"];';

        $apmsWithCallbacks = array_keys(array_filter($this->getAPMJSCallbacks()));
        $apmsWithCallbacksStr = 'var apmsWithCallbacks = ["' . implode('","', $apmsWithCallbacks) . '"];';
        if ($this->checkCheckoutFields() && $this->isCheckout()) {
///2do visibility of APM extra fields
            return <<<EOD
            $selectionInitJs
    //APM show available APMs and render Marks (icons)
    //By default all APM's radio buttons are hidden and then is displayed according allowed fundingSource
    window.paypalShowElegableOptions = function () {
        paypal.getFundingSources().forEach(function(fundingSource) {

            var button = paypal.Buttons({
              fundingSource: fundingSource
            });
            if (button.isEligible()) {
                $('.item-radio.item-payment.{$this->code}_' + fundingSource).show();
            } else {
                //console.log('>> fundingSource ' + fundingSource + ' not eligible');
            }

            var mark = paypal.Marks({
                fundingSource: fundingSource,
            });
            if (mark.isEligible() && $('.paypal-marks-icon-' + fundingSource).length) {
                mark.render('.paypal-marks-icon-' + fundingSource);
            }

        });
    }
    window.toggleSubFields_{$this->code} = function () {
        {$apmsWithFieldsStr}{$apmsWithCallbacksStr}{$apmsWithoutOwnButton}{$apmsWithAgreement}
        $('input[name=payment][value^="{$this->code}_"]').not(':checked').each(function() {
            if ($('.paypal-fields-container.'+$(this).val()).length)
                $('.paypal-fields-container.'+$(this).val()).hide();
            $('.paypal-button-container .paypal-button-'+$(this).val()).hide();
        });
        var fundingSource = '';
        try {
            fundingSource = $('input[name=payment]:checked').val();
            if (fundingSource.indexOf('paypal_rest_')<0) {
                fundingSource = '';
            } else {
                fundingSource = fundingSource.replace('paypal_rest_', '');
            }
        } catch ( e ) {
        }
        if (typeof window['paypal_render_subfields_' + fundingSource] == 'function'){
            try {
                window['paypal_render_subfields_' + fundingSource]();
            } catch (e ) {
                console.log(e);
            }
        } else {
//console.log('not a function ' + 'paypal_render_subfields_' + fundingSource);
        }

        //custom fields - use standard osc checkout button, else - provided by PayPal
        if ($('input[name=payment][value^="{$this->code}_"]').is(':checked')){
            if ($('.paypal-fields-container.'+$('input[name=payment]:checked').val()).length) {
                $('.paypal-fields-container.'+$('input[name=payment]:checked').val()).show();
            }
            //$('.paypal-button-container .paypal-button-'+$('input[name=payment]:checked').val()).show();
        }
        //if ($('input[name=payment][value="{$this->code}_paypal"]').is(':checked')){
        //if (fundingSource != '' && apmsWithCallbacks.indexOf(fundingSource) < 0 ) {
        if (fundingSource != '' && apmsWithoutOwnButton.indexOf(fundingSource) < 0 ) {
            {$checkedJSString}
            $('[data-name^="checkout"] .paypal-button-container').removeClass('button-hide');
            //$('#frmCheckout button[type="submit"]').addClass('button-hide');
            $('.paypal-button-container .paypal-button-'+$('input[name=payment]:checked').val()).show();
        } else {
            {$unCheckedJSString}
            if (fundingSource != '' && apmsWithAgreement.indexOf(fundingSource) != -1 ) { //has agreement and no button there
                $('.paypal-button-container').show();
                $('[data-name^="checkout"] .paypal-button-container').removeClass('button-hide');
                $('.paypal-button-container .paypal-button-'+$('input[name=payment]:checked').val()).show();
            } else {
                $('[data-name^="checkout"] .paypal-button-container').addClass('button-hide');
            }
            //$('#frmCheckout button[type="submit"]').removeClass('button-hide');
        }

    }

        $(document).ready(function(){ //tl
            try {
                paypalShowElegableOptions();
            }catch (e ) {console.log(e ); }
            window.toggleSubFields_{$this->code}();
        })

EOD;
        } else {

            return <<<EOD
    window.toggleSubFields_{$this->code} = function () {
        if ($('input[name=payment][value="{$this->code}"]').is(':checked')){
        {$checkedJSString}
            $('[data-name^="checkout"] .paypal-button-container').removeClass('button-hide');
            //$('#frmCheckout button[type="submit"]').addClass('button-hide');
        } else {
        {$unCheckedJSString}
            $('[data-name^="checkout"] .paypal-button-container').addClass('button-hide');
            //$('#frmCheckout button[type="submit"]').removeClass('button-hide');
        }

    }

        $(document).ready(function(){ //tl
            window.toggleSubFields_{$this->code}();
            try {
                //2check checkout_payment_changed.set('window.toggleSubFields_{$this->code}');
            }catch (e ) {console.log(e ); }
        })

EOD;
        }
        
    }
    
    public function getHttpClient() {
        if ($this->getMode() == 'Live') {
            $environment = new \PayPalCheckoutSdk\Core\ProductionEnvironment($this->_getClientId(), $this->_getClientSecret());
        } else {
            $environment = new \PayPalCheckoutSdk\Core\SandboxEnvironment($this->_getClientId(), $this->_getClientSecret());
        }
        $client = new \PayPalCheckoutSdk\Core\PayPalHttpClient($environment);
        return $client;
    }

  /**
   *  format prices without currency formatting 
   * @param decimal $number
   * @param string $currency_code
   * @param float $currency_value
   * @return decimal
   */
  function formatRaw($number, $currency_code = '', $currency_value = '') {
    global $currencies;
    global $currency;

    if (empty($currency_code)) {
      $currency_code = $currency;
    }

    if (empty($currency_value) || !is_numeric($currency_value)) {
      $currency_value = $currencies->currencies[$currency_code]['value'];
    }

    return number_format(round($number * $currency_value, $currencies->currencies[$currency_code]['decimal_places']), $currencies->currencies[$currency_code]['decimal_places'], '.', '');
  }

/**
 * check whether to save order before payment
 * @param type $force
 * @return string
 */
    public function saveOrderBefore($force = false) {
        $orderClass = false;
        if ($force) {
            $orderClass = 'Order';
        } else {
            if (defined('MODULE_PAYMENT_PAYPAL_PARTNER_ORDER_BEFORE_PAYMENT') && MODULE_PAYMENT_PAYPAL_PARTNER_ORDER_BEFORE_PAYMENT != 'Yes') {
                $orderClass = 'Order';
            }
        }
        return $orderClass;
    }

/**
 * APMs are delayed payments (several days) order must be saved before payment
 * @return type
 */
    public function checkSaveTmpOrder() {
        $ret = false;
        if (!$this->saveOrderBefore()) {
            //$ret = $this->saveOrderBySettings('TmpOrder');
            $ret = $this->saveOrderBySettings('Order');
        }
        return $ret;
    }

/**
 * save order and save id in session (ppp_tmp_order, ppp_order_before)
 * @param string|false $force
 * @return string|false order id
 */
    public function saveOrderBySettings($force = false) {
        $ret = false;
        $orderClass = $this->saveOrderBefore($force);
        if ($orderClass) {
            if ($orderClass != 'TmpOrder') {
                $order = $this->getOrderInstance();
                $order->info['order_status'] = $this->getDefaultOrderStatusId();

            }
            $ret = $this->saveOrder($orderClass);
            if (!empty($ret)) {
                if ($orderClass == 'TmpOrder') {
                    $ret = 'tmp' . $ret;
                    $key = 'ppp_tmp_order';
                } else {
                    $key = 'ppp_order_before';
                }
                $this->setSession($key, $ret);
            }
        }
        return $ret;
    }


}
}
