<?php

/*
 * This file is part of osCommerce ecommerce platform.
 * 
 * osCommerce the ecommerce
 * 
 * @link https://www.oscommerce.com
 * @copyright Copyright 2000-2022 osCommerce LTD
 * 
 * Released under the GNU General Public License
 * For the full copyright and license information, please view the LICENSE.TXT file that was distributed with this source code.
 */

use PaypalPartner\models\Customer;

trait PaypalManagerTrait {

    /** @property order $order ref to order in session */
    var $orderInstance = null;
    var $contentType = 'phisical'; //not used
    /** @property Customer $_customer */
    private $_customer = null;
    /** @property shipping $shipping_modules (collection) */
    private $shipping_modules;
    /** @property array $shipping_cost */
    protected $shipping_cost;
    /** @property order_total $totals (collection) */
    private $totals;
    private $quotes = [];
    private $billingAddressChanged = false;
    private $deliveryAddressChanged = false;


    /**
     * oscok
     * @return boolean
     */
    public function isInstance() {

        $ret = false;
        if (!empty($this->orderInstance)) {
            $ret = true;
        }
        return $ret;
    }

    /**
     *
     * @return order
     */
    public function createOrderInstance($global = false) {
        global $order;
        if ($global!==true || !is_object($order)) {
            $instance = new order();
            $this->orderInstance = $instance;
            $order = $instance;
        } else {
            $this->orderInstance = $order;
        }
        return $this->orderInstance;
    }

/**
 * osc - recreate order with ID
 */
    public function getOrderInstanceWithId($id) {
        $instance = new order($id);
        $instance->info['orders_id'] = $id;
        $instance->order_id = $id;
        //$instance->info['order_status'] = $id;
        $this->orderInstance = $instance;
        global $order;
        if (is_object($instance)) {
            $order = $instance;
            //$this->setSession('order', $order);
        }
        return $this->orderInstance;
    }

    public function getOrderInstance() {
        if (!is_object($this->orderInstance))
            throw new \Exception('Order instance is not defined');
        return $this->orderInstance;
    }
    

/**
 * oscok
 * @staticvar array $address
 * @return array
 */
    public function getBillingAddress() {
        static $address = null;
        if (is_null($address) || $this->billingAddressChanged) {
            $address = $this->getAddressArray('bill');
        }
        return $address;
    }

/**
 * oscok
 * @staticvar array $address
 * @return array
 */
    public function getDeliveryAddress() {
        static $address = null;
        if (is_null($address) || $this->deliveryAddressChanged) {
            $address = $this->getAddressArray('ship');
        }
        return $address;
    }

    private function getAddressArray($which) {
        $address = [];
        if (!in_array($which, array('ship', 'bill'))) {
            $which = 'bill';
        }
        if ($which == 'bill') {
            $sessionAddress = $this->getBillto();
        } else {
            $sessionAddress = $this->getSendto();
        }
        if ($this->hasSession('estimate_'  . $which) || !$this->isCustomerAssigned()) {
            $address = $this->loadDefaultAddressValues('_' . $which);
        } else {
            
            if ($sessionAddress) {
                if (is_scalar($sessionAddress)) {
                    $address = $this->getCustomersAddress($sessionAddress, true, true);
                    if ($address) {
                        if (is_array($address['country'])) {
                            $address['country'] = self::addCountriesKey($address['country']);
                        }
                    }
                } else if (is_array($sessionAddress)) {
                    $address = $sessionAddress;
                }
            }
        }
        return $address;
    }

/**
 * load data either from estimate_* or  empty address (store country and zone only)
 * @param string $postfix
 * @return array
 */
    public function loadDefaultAddressValues($postfix) {
        if ($this->hasSession('estimate' . $postfix)) {
            $estimate = $this->getSession('estimate' . $postfix);
            $_country_info = tep_get_countries($estimate['country_id'], true);
            $address = [
                'street_address' => $estimate['street_address']??'',
                'suburb' => $estimate['suburb']??'',
                'city' => $estimate['city']??'',
                'postcode' => $estimate['postcode']??'',
                'zone_id' => (isset($estimate['zone']) && !empty($estimate['zone']) ? (is_int($estimate['zone']) ? $estimate['zone'] : self::get_zone_id($estimate['zone'], $estimate['country_id'])) : 0),
                'country_id' => $estimate['country_id'],
            ];
            foreach (['company_vat', 'company_vat_date', 'company_vat_status', 'customs_number', 'customs_number_date', 'customs_number_status', 'city'] as $k) {
              if (!empty($estimate[$k])) {
                $address[$k] = $estimate[$k];
              }
            }
        } else {
            $_country_info = tep_get_countries($this->getTaxCountry(), true);
            $address = [
                'street_address' => '',
                'suburb' => '',
                'city' => '',
                'postcode' => '',
                'zone_id' => $this->getTaxZone(),
                'country_id' => $this->getTaxCountry(),
            ];
        }

        $_country_info['title'] = $_country_info['countries_name'];
        $_country_info['iso_code_2'] = $_country_info['countries_iso_code_2'];
        $_country_info['iso_code_3'] = $_country_info['countries_iso_code_3'];
        $address['country'] = $_country_info;


        $address['countries_id'] = $address['country_id'];
        $address['entry_country_id'] = $address['country_id'];

        $address['address_format_id'] = $address['country_id'];
        $address['entry_state'] = $address['country_id'];
        $address['entry_country_id'] = $address['country_id'];
        //'' => $sendto['address_format_id'],
        //'entry_state' => $sendto['zone_name']
        foreach(array('countries_name', 'countries_iso_code_2', 'countries_iso_code_3') as $k) {
            $address[$k] = $_country_info[$k];
        }

        return $address;
    }

/**
 * oscok
 * @return int
 */
    public function getTaxCountry() {
        if ($this->isCustomerAssigned()) {
            $ret = $this->getSession('customer_country_id');//2check //getCustomersIdentity()->getDefaultAddress
        } else {
            $ret = STORE_COUNTRY;
        }
        return $ret;
    }

    public function getTaxZone() {
        if ($this->isCustomerAssigned()) {
            $ret = $this->getSession('customer_zone_id'); //getCustomersIdentity()->getDefaultAddress
        } else {
            $ret = STORE_ZONE;
        }
        return $ret;
    }

/**
 * oscok
 * @param int $abId
 * @param bool $toArray
 * @param bool $stripEntry
 * @return array|stdClass
 */
    public function getCustomersAddress($abId, $toArray = false, $stripEntry = false) {
        $address = null;
        if ($customer = $this->getCustomersIdentity()) {
            $address = $customer->getAddressBook($abId, $toArray);
            if ($address && $stripEntry && $toArray) {
                $address = self::skipEntry($address);
            }
        }
        return $address;
    }
//2check
    public function getBillto() {
        $billto = $this->getSession('billto') ?? false;
        if (!$billto || (is_array($billto) && empty($billto['country_id']))) {
            if ($customer = $this->getCustomersIdentity()) {
                $address = $customer->getDefaultAddress();
                if ($address) {
                    $billto = $address->address_book_id;
                    $this->setSession('billto', $billto);
                }
            }
        }
        return $billto;
    }


    public function getSendto() {
        $sendto = $this->getSession('sendto') ?? false;
        if (!$sendto || (is_array($sendto) && empty($sendto['country_id']))) {
            if ($customer = $this->getCustomersIdentity()) {
                $address = $customer->getDefaultAddress();
                if ($address) {
                    $sendto = $address->address_book_id;
                    $this->setSession('sendto', $sendto);
                }
            }
        }
        return $sendto;
    }
    

    public function resetBillingAddress(){
        $this->billingAddressChanged = true;
    }

    /**
     * oscok
     * @global int $customer_id (session)
     * @return bool
     */
    public function isCustomerAssigned() {
        global $customer_id;
        return (((int)$customer_id>0)?true:false);
    }

    /**
     * oscok
     * @global int $customer_id (session)
     * @return int|false
     */
    public function getCustomerAssigned() {
        global $customer_id;
        return (((int)$customer_id>0)?$customer_id:false);
    }

    public function getCustomerByEmail($email) {
        $this->_customer = Customer::findOne("customers_email_address like '" . $email . "'");
        if (is_null($this->_customer)) {
            $this->_customer = new Customer();
            $this->_customer->customers_email_address = $email;
        }
        return $this->_customer;
    }

/**
 *
 * @return Customer
 */
    public function getCustomersIdentity() {
        if ($this->getCustomerAssigned()) {
            if (is_object($this->_customer) && $this->getCustomerAssigned() == $this->_customer->customers_id) {
                return $this->_customer;
            }
            $this->_customer = Customer::findOne(['customers_id' => $this->getCustomerAssigned()]);
            if (is_null($this->_customer)) {
                $this->resetSession('customer_id');
                $this->_customer = new Customer();
                if ($this->hasSession('guest_email_address')) {
                    $this->_customer->customers_email_address = $this->getSession('guest_email_address');
                }
                if ($this->hasSession('guest_telephone')) {
                    $this->_customer->customers_telephone = $this->getSession('guest_telephone');
                }
            }
        } else {
            if (is_null($this->_customer)){
                $this->_customer = new Customer();
                if ($this->hasSession('guest_email_address')) {
                    $this->_customer->customers_email_address = $this->getSession('guest_email_address');
                }
                if ($this->hasSession('guest_telephone')) {
                    $this->_customer->customers_telephone = $this->getSession('guest_telephone');
                }
            }
        }
        return $this->_customer;
    }

    //has
    public function hasSession($key) {
        return tep_session_is_registered($key);
    }

    //set
    public function setSession($key, $val) {
        global $$key;
        if (!tep_session_is_registered($key)) {
            tep_session_register($key);
        }
        self::warning("setSession $key " . print_r($val, true));
        $$key = $val;
    }
    //remove
    public function resetSession($key) {
        if (tep_session_is_registered($key)) {
            global $$key;
            tep_session_unregister($key);
            $$key = null;
            self::warning("resetSession($key)");
        }
    }
    //get
    public function getSession($key) {
        $ret = null;
        if (tep_session_is_registered($key)) {
            global $$key;
            $ret = $$key;
        }
        return $ret;
    }
    public function setPayment($method) {
        $this->setSession('payment', $method);
    }

/**
 * strip entry_ prefix  from address array
 * @param array $address
 * @return array
 */
    public static function skipEntry($address = []){
        if (is_array($address) && !empty($address)) {
            $keys = array_keys($address);
            $keys = array_map(function($el){return str_replace("entry_", "", $el);}, $keys);
            $address = array_combine($keys, array_values($address));
        }
        return $address;
    }

/**
 * toggle countries_ prefix
 * @param array $address
 * @return array
 */
    public static function addCountriesKey(array $address){
        if ($address){
            foreach($address as $key => $value){
                if (strpos($key, 'countries_') !== false){
                    $address[substr($key, 10)] = $value;
                } else {
                    $address['countries_'.$key] = $value;
                }
            }
        }
        return $address;
    }

    public static function get_country_info_by_iso($iso_code, $which='iso-2') {
      if (!in_array($which, array('iso2', 'iso-2', 'iso3', 'iso-3'))) {
        $which = 'iso-2';
      }
      if ($which == 'iso-2' || $which == 'iso2') {
        $field = 'countries_iso_code_2';
      } else {
        $field = 'countries_iso_code_3';
      }

      $res = tep_db_query("select * from " . TABLE_COUNTRIES . " where $field = '" . tep_db_input($iso_code) . "'");
      $ret = array();

      if ($d = tep_db_fetch_array($res)) {
          $ret = array('id' => $d['countries_id'],
              'title' => $d['countries_name'],
              'iso_code_2' => $d['countries_iso_code_2'],
              'iso_code_3' => $d['countries_iso_code_3'],
              'address_format_id' => $d['address_format_id']
          );
      } else {
        $ret = $iso_code;
      }
      return $ret;
    }

    public static function get_zone_id($name, $country_id) {
        $tmp = self::get_zone_info_by_title($name, $country_id);
        if (!empty($tmp->zone_id)) {
            $ret = $tmp->zone_id;
        } else {
            $ret = 0;
        }
        return $ret;
    }

    public static function get_zone_info_by_title($name, $country_id) {
      $res = tep_db_query("select * from " . TABLE_ZONES . " 
             where  zone_country_id = '" . (int)$country_id . "' and (
                    zone_name='" . tep_db_input($name) . "' or zone_code='" . tep_db_input($name) . "' or
                    zone_name like '" . tep_db_input($name) . "' or zone_code like '" . tep_db_input($name) . "')
             order by zone_name='" . tep_db_input($name) . "'desc,  zone_code='" . tep_db_input($name) . "' desc,
                    zone_name like '" . tep_db_input($name) . "' desc, zone_code like '" . tep_db_input($name) . "' desc Limit 1");

      if ($d = tep_db_fetch_array($res)) {
          $ret = (object)$d;
      } else {
        $ret = false;
      }
      return $ret;
    }

    /**
     * oscok
     * @return bool
     */
    public function isShippingNeeded() {
        $needed = true;
        $needed = $needed && $this->orderInstance->content_type != 'virtual';
        $needed = $needed && ($this->contentType != 'virtual') && ($this->contentType != 'virtual_weight');
        $needed = $needed && !$this->skipShipping;
        
        if (!$needed) {
            $this->setSession('shipping', false);
            $this->setSession('sendto', false);
        }
        return $needed;
    }

/**
 * oscok
 * create class shipping and get quotes
 * @staticvar array $_quotes
 * @param bool $renew ??required if address changed
 * @return array|null
 */
    public function getShippingQuotesByChoice($renew = false) {
        static $_quotes = null;
        if ( $renew ) $_quotes = null;
        if (is_null($_quotes) && $this->isShippingNeeded()) {
            $_quotes = $this->getAllShippingQuotes();

            $exists = $this->checkExistedShippinInQuotes($_quotes);
            if (!$exists && !empty($this->shipping_modules)) {
                $cheapest = $this->shipping_modules->cheapest();
                if (is_array($cheapest)) {
                    $this->setSession('shipping', $cheapest);
                    $this->checkExistedShippinInQuotes($_quotes);
                }
            }
            $_quotes = $this->wrapQuotes($_quotes);
        }

        return $_quotes;
    }

/**
 * oscok
 * @param bool $requote
 * @return array
 */
    public function getAllShippingQuotes($requote=false) {
        if (empty($this->quotes) || $requote || empty($this->shipping_modules)) {
            $this->quotes = $this->getShippingCollection()->quote('', '');
        }
        $exists = $this->checkExistedShippinInQuotes($this->quotes);
        if (!$exists && !empty($this->shipping_modules)) {
            $cheapest = $this->shipping_modules->cheapest();
            if (is_array($cheapest)) {
                $this->setSession('shipping', $cheapest);
            }
        }
        return $this->quotes;
    }

/**
 * oscok
 * @param string $only_shipping
 * @return shipping
 */
    public function getShippingCollection($only_shipping = '') {
        global $shipping_modules;
        if (!is_object($this->shipping_modules) && !is_object($shipping_modules)) {
            $this->shipping_modules = new shipping($only_shipping);
        } elseif (is_object($shipping_modules)) {
            $this->shipping_modules = $shipping_modules;
        }

        return $this->shipping_modules;
    }

/**
 * oscok
 * @param array $quotes
 */
    public function checkExistedShippinInQuotes($quotes) {
        $select_shipping = $this->getSelectedShipping();
        $exist = false;
        if (is_array($quotes) && $select_shipping) {
            for ($i = 0, $n = sizeof($quotes); $i < $n; $i++) {
                if (!isset($quotes[$i]['error'])) {
                    for ($j = 0, $n2 = sizeof($quotes[$i]['methods']??[]); $j < $n2; $j++) {
                        if (($select_shipping == $quotes[$i]['id'] . '_' . $quotes[$i]['methods'][$j]['id'])) {
                            $exist = true;
                            $shipping = $this->_getShipingAsArray($quotes[$i], $quotes[$i]['id'], $quotes[$i]['methods'][$j]['id'], $j, false);
                            $this->setSession('shipping', $shipping);
                            $this->shipping_cost = $shipping;
                            $order = $this->getOrderInstance();
                            $order->info['shipping_method'] = $shipping['title'];
                            if ($order->info['shipping_cost']!=0) {
                                $order->info['total'] -= $order->info['shipping_cost'];
                            }
                            $order->info['shipping_cost'] = $shipping['cost'];
                            $order->info['total'] += $shipping['cost'];
                            /*if (DISPLAY_PRICE_WITH_TAX == 'true') {
                              $order->info['total'] = $order->info['subtotal'] + $order->info['shipping_cost'];
                            } else {
                              $order->info['total'] = $order->info['subtotal'] + $order->info['tax'] + $order->info['shipping_cost'];
                            }*/
                            break;
                        }
                    }
                }
            }
        }
        if (!$exist) {
            $this->resetSession('shipping');
            $this->shipping_cost = null;
        }
        return $exist;
    }
/**
 * oscok
 * @return string|false
 */
    public function getSelectedShipping() {
        $_selected = false;
        $_shipping = $this->getSession('shipping');
        $_selected = is_array($_shipping) ? $_shipping['id'] : false;
        if (!$_selected) {
            $_selected = $this->getSession('select_shipping') ?? false;
        }
        $exists = false;
        if ($_selected && (strpos($_selected, '_') !== false) && !empty($this->quotes)) {

            foreach ($this->quotes as $v) {
                
            }
        }

        return $_selected;
    }

/**
 * oscok
 * @param array $quote
 * @param string $module
 * @param string $method
 * @param int $method_index
 * @param bool $free_shipping
 * @return array
 */
    private function _getShipingAsArray($quote, $module, $method, $method_index, $free_shipping = false){
        $cost = (float) $quote['methods'][$method_index]['cost'];
        return [
            'module' => $module,
            'id' => $module . '_' . $quote['methods'][$method_index]['id'],
            'title' => (($free_shipping == true) ? $quote['methods'][$method_index]['title'] : $quote['module'] . (empty(trim($quote['methods'][$method_index]['title']))?'':' (' . $quote['methods'][$method_index]['title'] . ')')),
            'cost' => $cost,
            'cost_inc_tax' => tep_add_tax($cost, (isset($quote['tax']) ? $quote['tax'] : 0)),
        ];
    }

    /**
     * if shipping is required and $shippping has correct format then set or reset "shipping" in the manager itself
     * @param string $shipping value shippingCode_shippingMethod
     */
    public function setSelectedShipping($shipping) {
        if ($this->isShippingNeeded() && (strpos($shipping, '_') !== false)) {
            /*replace?*/$quotes = $this->getAllShippingQuotes();
            $this->resetSession('shipping');
            $this->setSession('select_shipping', $shipping);
            $this->checkExistedShippinInQuotes($quotes);
            /** /
            

            list($module, $method) = explode('_', $shipping);



            $quote = $this->getShippingCollection()->quote($method, $module);

            if (!isset($quote[0]['error'])) {
                if ((isset($quote[0]['methods'][0]['title'])) && (isset($quote[0]['methods'][0]['cost']))) {
                    $shipping = $this->_getShipingAsArray($quote[0], $module, $method, 0);
                }
                if (is_array($shipping)) {
                    $this->setSession('shipping', $shipping);
                } else {
                    $this->resetSession('shipping');
                }
            } else {
                $this->resetSession('shipping');
            } /**/
        }
    }

/**
 * oscok
 * @global type $currencies
 * @param array $quotes
 * @return array
 */
    public function wrapQuotes($quotes) {
        if (is_array($quotes)) {
            global $currencies;
            $select_shipping = $this->getSelectedShipping();
            for ($i = 0, $n = count($quotes); $i < $n; $i++) {
                if (!isset($quotes[$i]['error'])) {
                    for ($j = 0, $n2 = count($quotes[$i]['methods']??[]); $j < $n2; $j++) {
                        $quotes[$i]['methods'][$j]['code'] = $quotes[$i]['id'] . '_' . $quotes[$i]['methods'][$j]['id'];
                        $quotes[$i]['methods'][$j]['selected'] = $select_shipping === $quotes[$i]['methods'][$j]['code'];
                        $quotes[$i]['selected'] = $select_shipping === $quotes[$i]['methods'][$j]['code'];
                    }
                }
            }
        }
        return $quotes;
    }

/**
 * osc - recreate order from cart
 */
    public function checkoutOrderWithAddresses() {
        $this->orderInstance = null;
        $billto = $this->getBillingAddress();
        $this->setSession('billto', $billto);
        $sendto = $this->getDeliveryAddress();
        $this->setSession('sendto', $sendto);
        
        $this->scalarToAddresses();

        $this->createOrderInstance();
    }

    public function getTotalCollection($reconfig = false) {
        global $order_total_modules;
        if (!is_object($this->totals) && !is_object($order_total_modules)) {
            $order_total_modules = $this->totals = new order_total();
        } elseif (is_object($order_total_modules)) {
            $this->totals = $order_total_modules;
        }

        return $this->totals;
    }
/**
 * osc - ignore all params
 * @param type $wrap
 * @param type $context
 * @param type $all
 * @return array
 */
    public function getTotalOutput($wrap = true, $context = null, $all = false, $renew = false) {
        $order = $this->getOrderInstance();
        //if (!empty($order->info['orders_id']) && count($order->totals)) { //order from db
        if (count($order->totals) && (!empty($order->info['orders_id'])  || !$renew) ) { //order from db
            $order_total_output = $order->totals;
        } else {
            $order_total_output = $this->totalProcess();
        }
        return $order_total_output;
    }

/**
 * oscok
 * @return array
 */
    public function totalProcess() {
        $order = $this->getOrderInstance();
        $ret =  $this->getTotalCollection()->process();
        $order->totals = $ret;

        return $ret;
    }

    public function resetDeliveryAddress(){
        $this->deliveryAddressChanged = true;
    }


  /**
   * save order before redirect to payment gateway
   * @param bool|string $asType orderClass (for now TmpOrder only) or false (
   * @param bool  $updateStock default false
   * @param array $params extra order params default []
   * @return integer|null - orderId
   */
  protected function saveOrder($asType = false, $updateStock = false, $params = []) {
    global $customer_id, $languages_id, $currencies;

    global $payment, $sendto, $billto;

    $ret = null;

    switch ($asType) {

      case 'Order':
      default:
        $order = $this->getOrderInstance();
/** /
$_ee = new \Exception(); echo "#### <PRE>" . __FILE__ . ':' . __LINE__ . ' ' . print_r($_ee->getTraceAsString(), true) . "</PRE>\n";
echo "#### <PRE>"  . __FILE__ .':' . __LINE__ . ' ' . print_r($order, true) ."</PRE>"; die;
/**/
        if ($payment != $this->code) {
            $order->info['comments'] = "$payment prepare \n" . $order->info['comments'];
        }

        $order_totals = $order->totals;

        $sql_data_array = array('customers_id' => $customer_id,
                                'customers_name' => $order->customer['firstname'] . ' ' . $order->customer['lastname'],
                                'customers_company' => $order->customer['company'],
                                'customers_street_address' => $order->customer['street_address'],
                                'customers_suburb' => $order->customer['suburb'],
                                'customers_city' => $order->customer['city'],
                                'customers_postcode' => $order->customer['postcode'],
                                'customers_state' => $order->customer['state'],
                                'customers_country' => $order->customer['country']['title'],
                                'customers_telephone' => $order->customer['telephone'],
                                'customers_email_address' => $order->customer['email_address'],
                                'customers_address_format_id' => $order->customer['format_id'],
                                'delivery_name' => trim($order->delivery['firstname'] . ' ' . $order->delivery['lastname']),
                                'delivery_company' => $order->delivery['company'],
                                'delivery_street_address' => $order->delivery['street_address'],
                                'delivery_suburb' => $order->delivery['suburb'],
                                'delivery_city' => $order->delivery['city'],
                                'delivery_postcode' => $order->delivery['postcode'],
                                'delivery_state' => $order->delivery['state'],
                                'delivery_country' => $order->delivery['country']['title'],
                                'delivery_address_format_id' => $order->delivery['format_id'],
                                'billing_name' => $order->billing['firstname'] . ' ' . $order->billing['lastname'],
                                'billing_company' => $order->billing['company'],
                                'billing_street_address' => $order->billing['street_address'],
                                'billing_suburb' => $order->billing['suburb'],
                                'billing_city' => $order->billing['city'],
                                'billing_postcode' => $order->billing['postcode'],
                                'billing_state' => $order->billing['state'],
                                'billing_country' => $order->billing['country']['title'],
                                'billing_address_format_id' => $order->billing['format_id'],
                                'payment_method' => $order->info['payment_method'],
                                'cc_type' => $order->info['cc_type'],
                                'cc_owner' => $order->info['cc_owner'],
                                'cc_number' => $order->info['cc_number'],
                                'cc_expires' => $order->info['cc_expires'],
                                'date_purchased' => 'now()',
                                'orders_status' => $order->info['order_status'],
                                'currency' => $order->info['currency'],
                                'currency_value' => $order->info['currency_value']);
        tep_db_perform(TABLE_ORDERS, $sql_data_array);
        $insert_id = tep_db_insert_id();
        for ($i=0, $n=sizeof($order_totals); $i<$n; $i++) {
          $sql_data_array = array('orders_id' => $insert_id,
                                  'title' => $order_totals[$i]['title'],
                                  'text' => $order_totals[$i]['text'],
                                  'value' => $order_totals[$i]['value'],
                                  'class' => $order_totals[$i]['code'],
                                  'sort_order' => $order_totals[$i]['sort_order']);
          tep_db_perform(TABLE_ORDERS_TOTAL, $sql_data_array);
        }

        $customer_notification = (SEND_EMAILS == 'true') ? '1' : '0';
        $sql_data_array = array('orders_id' => $insert_id,
                                'orders_status_id' => $order->info['order_status'],
                                'date_added' => 'now()',
                                'customer_notified' => $customer_notification,
                                'comments' => $order->info['comments']);
        tep_db_perform(TABLE_ORDERS_STATUS_HISTORY, $sql_data_array);

      // initialized for the email confirmation
        $products_ordered = '';

        for ($i=0, $n=sizeof($order->products); $i<$n; $i++) {
      // Stock Update - Joao Correia
          if (STOCK_LIMITED == 'true') {
            if (DOWNLOAD_ENABLED == 'true') {
              $stock_query_raw = "SELECT products_quantity, pad.products_attributes_filename
                                  FROM " . TABLE_PRODUCTS . " p
                                  LEFT JOIN " . TABLE_PRODUCTS_ATTRIBUTES . " pa
                                   ON p.products_id=pa.products_id
                                  LEFT JOIN " . TABLE_PRODUCTS_ATTRIBUTES_DOWNLOAD . " pad
                                   ON pa.products_attributes_id=pad.products_attributes_id
                                  WHERE p.products_id = '" . tep_get_prid($order->products[$i]['id']) . "'";
      // Will work with only one option for downloadable products
      // otherwise, we have to build the query dynamically with a loop
              $products_attributes = (isset($order->products[$i]['attributes'])) ? $order->products[$i]['attributes'] : '';
              if (is_array($products_attributes)) {
                $stock_query_raw .= " AND pa.options_id = '" . (int)$products_attributes[0]['option_id'] . "' AND pa.options_values_id = '" . (int)$products_attributes[0]['value_id'] . "'";
              }
              $stock_query = tep_db_query($stock_query_raw);
            } else {
              $stock_query = tep_db_query("select products_quantity from " . TABLE_PRODUCTS . " where products_id = '" . tep_get_prid($order->products[$i]['id']) . "'");
            }
            if (tep_db_num_rows($stock_query) > 0) {
              $stock_values = tep_db_fetch_array($stock_query);
      // do not decrement quantities if products_attributes_filename exists
              if ((DOWNLOAD_ENABLED != 'true') || (!$stock_values['products_attributes_filename'])) {
                $stock_left = $stock_values['products_quantity'] - $order->products[$i]['qty'];
              } else {
                $stock_left = $stock_values['products_quantity'];
              }
              tep_db_query("update " . TABLE_PRODUCTS . " set products_quantity = '" . (int)$stock_left . "' where products_id = '" . tep_get_prid($order->products[$i]['id']) . "'");
              if ( ($stock_left < 1) && (STOCK_ALLOW_CHECKOUT == 'false') ) {
                tep_db_query("update " . TABLE_PRODUCTS . " set products_status = '0' where products_id = '" . tep_get_prid($order->products[$i]['id']) . "'");
              }
            }
          }

      // Update products_ordered (for bestsellers list)
          tep_db_query("update " . TABLE_PRODUCTS . " set products_ordered = products_ordered + " . sprintf('%d', $order->products[$i]['qty']) . " where products_id = '" . tep_get_prid($order->products[$i]['id']) . "'");

          $sql_data_array = array('orders_id' => $insert_id,
                                  'products_id' => tep_get_prid($order->products[$i]['id']),
                                  'products_model' => $order->products[$i]['model'],
                                  'products_name' => $order->products[$i]['name'],
                                  'products_price' => $order->products[$i]['price'],
                                  'final_price' => $order->products[$i]['final_price'],
                                  'products_tax' => $order->products[$i]['tax'],
                                  'products_quantity' => $order->products[$i]['qty']);
          tep_db_perform(TABLE_ORDERS_PRODUCTS, $sql_data_array);
          $order_products_id = tep_db_insert_id();

      //------insert customer choosen option to order--------
          $attributes_exist = '0';
          $products_ordered_attributes = '';
          if (isset($order->products[$i]['attributes'])) {
            $attributes_exist = '1';
            for ($j=0, $n2=sizeof($order->products[$i]['attributes']); $j<$n2; $j++) {
              if (DOWNLOAD_ENABLED == 'true') {
                $attributes_query = "select popt.products_options_name, poval.products_options_values_name, pa.options_values_price, pa.price_prefix, pad.products_attributes_maxdays, pad.products_attributes_maxcount , pad.products_attributes_filename
                                     from " . TABLE_PRODUCTS_OPTIONS . " popt, " . TABLE_PRODUCTS_OPTIONS_VALUES . " poval, " . TABLE_PRODUCTS_ATTRIBUTES . " pa
                                     left join " . TABLE_PRODUCTS_ATTRIBUTES_DOWNLOAD . " pad
                                      on pa.products_attributes_id=pad.products_attributes_id
                                     where pa.products_id = '" . (int)$order->products[$i]['id'] . "'
                                      and pa.options_id = '" . (int)$order->products[$i]['attributes'][$j]['option_id'] . "'
                                      and pa.options_id = popt.products_options_id
                                      and pa.options_values_id = '" . (int)$order->products[$i]['attributes'][$j]['value_id'] . "'
                                      and pa.options_values_id = poval.products_options_values_id
                                      and popt.language_id = '" . (int)$languages_id . "'
                                      and poval.language_id = '" . (int)$languages_id . "'";
                $attributes = tep_db_query($attributes_query);
              } else {
                $attributes = tep_db_query("select popt.products_options_name, poval.products_options_values_name, pa.options_values_price, pa.price_prefix from " . TABLE_PRODUCTS_OPTIONS . " popt, " . TABLE_PRODUCTS_OPTIONS_VALUES . " poval, " . TABLE_PRODUCTS_ATTRIBUTES . " pa where pa.products_id = '" . (int)$order->products[$i]['id'] . "' and pa.options_id = '" . (int)$order->products[$i]['attributes'][$j]['option_id'] . "' and pa.options_id = popt.products_options_id and pa.options_values_id = '" . (int)$order->products[$i]['attributes'][$j]['value_id'] . "' and pa.options_values_id = poval.products_options_values_id and popt.language_id = '" . (int)$languages_id . "' and poval.language_id = '" . (int)$languages_id . "'");
              }
              $attributes_values = tep_db_fetch_array($attributes);

              $sql_data_array = array('orders_id' => $insert_id,
                                      'orders_products_id' => $order_products_id,
                                      'products_options' => $attributes_values['products_options_name'],
                                      'products_options_values' => $attributes_values['products_options_values_name'],
                                      'options_values_price' => $attributes_values['options_values_price'],
                                      'price_prefix' => $attributes_values['price_prefix']);
              tep_db_perform(TABLE_ORDERS_PRODUCTS_ATTRIBUTES, $sql_data_array);

              if ((DOWNLOAD_ENABLED == 'true') && isset($attributes_values['products_attributes_filename']) && tep_not_null($attributes_values['products_attributes_filename'])) {
                $sql_data_array = array('orders_id' => $insert_id,
                                        'orders_products_id' => $order_products_id,
                                        'orders_products_filename' => $attributes_values['products_attributes_filename'],
                                        'download_maxdays' => $attributes_values['products_attributes_maxdays'],
                                        'download_count' => $attributes_values['products_attributes_maxcount']);
                tep_db_perform(TABLE_ORDERS_PRODUCTS_DOWNLOAD, $sql_data_array);
              }
              $products_ordered_attributes .= "\n\t" . $attributes_values['products_options_name'] . ' ' . $attributes_values['products_options_values_name'];
            }
          }
      //------insert customer choosen option eof ----
          $products_ordered .= $order->products[$i]['qty'] . ' x ' . $order->products[$i]['name'] . ' (' . $order->products[$i]['model'] . ') = ' . $currencies->display_price($order->products[$i]['final_price'], $order->products[$i]['tax'], $order->products[$i]['qty']) . $products_ordered_attributes . "\n";
        }

        if (!empty($params['notify'])) {
        // lets start with the email confirmation
          $email_order = STORE_NAME . "\n" .
                         EMAIL_SEPARATOR . "\n" .
                         EMAIL_TEXT_ORDER_NUMBER . ' ' . $insert_id . "\n" .
                         EMAIL_TEXT_INVOICE_URL . ' ' . tep_href_link(FILENAME_ACCOUNT_HISTORY_INFO, 'order_id=' . $insert_id, 'SSL', false) . "\n" .
                         EMAIL_TEXT_DATE_ORDERED . ' ' . strftime(DATE_FORMAT_LONG) . "\n\n";
          if ($order->info['comments']) {
            $email_order .= tep_db_output($order->info['comments']) . "\n\n";
          }
          $email_order .= EMAIL_TEXT_PRODUCTS . "\n" .
                          EMAIL_SEPARATOR . "\n" .
                          $products_ordered .
                          EMAIL_SEPARATOR . "\n";

          for ($i=0, $n=sizeof($order_totals); $i<$n; $i++) {
            $email_order .= strip_tags($order_totals[$i]['title']) . ' ' . strip_tags($order_totals[$i]['text']) . "\n";
          }

          if ($order->content_type != 'virtual') {
            $email_order .= "\n" . EMAIL_TEXT_DELIVERY_ADDRESS . "\n" .
                            EMAIL_SEPARATOR . "\n" .
                            tep_address_label($customer_id, $sendto, 0, '', "\n") . "\n";
          }

          $email_order .= "\n" . EMAIL_TEXT_BILLING_ADDRESS . "\n" .
                          EMAIL_SEPARATOR . "\n" .
                          tep_address_label($customer_id, $billto, 0, '', "\n") . "\n\n";
          if (is_object($$payment)) {
            $email_order .= EMAIL_TEXT_PAYMENT_METHOD . "\n" .
                            EMAIL_SEPARATOR . "\n";
            $payment_class = $$payment;
            $email_order .= $order->info['payment_method'] . "\n\n";
            if (isset($payment_class->email_footer)) {
              $email_order .= $payment_class->email_footer . "\n\n";
            }
          }
          tep_mail($order->customer['firstname'] . ' ' . $order->customer['lastname'], $order->customer['email_address'], EMAIL_TEXT_SUBJECT, $email_order, STORE_OWNER, STORE_OWNER_EMAIL_ADDRESS);

        // send emails to other people
          if (SEND_EXTRA_ORDER_EMAILS_TO != '') {
            tep_mail('', SEND_EXTRA_ORDER_EMAILS_TO, EMAIL_TEXT_SUBJECT, $email_order, STORE_OWNER, STORE_OWNER_EMAIL_ADDRESS);
          }

        }


        $ret = $insert_id;
        $this->getOrderInstanceWithId($insert_id);
        break;
    }

    return $ret;
  }


//2do
    /**
     * updates status/code/+comments of payment transaction (if exists) or add new transaction record
     * $transactionDetails['orders_id'] or "$this->getOrderInstance()" should be set to save new transaction
     * sets orderInstance if no orders_id in the instance (often payment gateways don't have order id in transactions details)
     * @param string $transaction_id
     * @param array $transactionDetails [status_code status amount suborder_id comments date parent_transaction_id ("external") orders_id fulljson]
     * @return OrdersPayment|bool|null null not found false - not changed; OrdersPayment - added/changed, to save order comments (by OrderPayment only)
     */
    public function updatePaymentTransaction($transaction_id, $transactionDetails) {
      $ret = null;
      $status = $transactionDetails['status']??'';
      $amount = $transactionDetails['amount']??'';
      $suborder_id = $transactionDetails['suborder_id']??'';
      $comments = $transactionDetails['comments']??'';
      $pp = null;
      /// first table
      if ($this->isInstance() ) {
        $order = $this->getOrderInstance();
      } else {
        $order = false;
      }
      /// new table
      $op = self::OrderPaymentSearchRecord($this->code, $transaction_id);
      $_changed = false;

      if ($op && !empty($op->orders_payment_id)) {
        // exists - update if required
        $ret = false;
        if ($op->orders_payment_transaction_status != $status
            || (!empty($transactionDetails['status_code']) && $op->orders_payment_status!=$transactionDetails['status_code'])
            || (!empty($transactionDetails['last_updated']) && $op->orders_payment_date_update!=$transactionDetails['last_updated'])
            ) {
            $op->orders_payment_transaction_status = $status;
            if (!empty($transactionDetails['status_code'])) {
              $_changed = $_changed || ($op->orders_payment_status != $transactionDetails['status_code']);
              $op->orders_payment_status = $transactionDetails['status_code'];
            }
            if (!empty($transactionDetails['amount'])) {
              $_changed = $_changed || ($op->orders_payment_amount != $transactionDetails['amount']);
              $op->orders_payment_amount = $transactionDetails['amount'];
            }
            if (!empty($transactionDetails['comments'])) {
              $_changed = $_changed || ($op->orders_payment_transaction_commentary != $transactionDetails['comments']);
              $op->orders_payment_transaction_commentary = $transactionDetails['comments'];
            }
            if (!empty($transactionDetails['last_updated'])) {
              $_changed = $_changed || ($op->orders_payment_date_update != $transactionDetails['last_updated']);
              $op->orders_payment_date_update = $transactionDetails['last_updated'];
            }
            if (!empty($transactionDetails['deferred'])) {
              $_changed = $_changed || ($op->deferred != $transactionDetails['deferred']);
              $op->deferred = $transactionDetails['deferred'];
            }
            if (!empty($transactionDetails['currency'])) {
              $_changed = $_changed || ($op->orders_payment_currency != $transactionDetails['currency']);
              $op->orders_payment_currency = $transactionDetails['currency'];
            }
            if (!empty($transactionDetails['transaction_date'])) {
              $_changed = $_changed || ($op->orders_payment_transaction_date != $transactionDetails['transaction_date']);
              $op->orders_payment_transaction_date = $transactionDetails['transaction_date'];
            }
            if (empty($op->orders_payment_order_id) && !empty($transactionDetails['orders_id'])) {
                $_changed = $_changed || ($op->orders_payment_order_id != $transactionDetails['orders_id']);
                $op->orders_payment_order_id = $transactionDetails['orders_id'];
            }
            if (empty($op->orders_payment_module_name) && !empty($transactionDetails['payment_method'])) {
                $_changed = $_changed || ($op->orders_payment_module_name != $transactionDetails['payment_method']);
                $op->orders_payment_module_name = $transactionDetails['payment_method'];
            }
            if ($_changed) {
              if (!empty($transactionDetails['fulljson'])) {
                $op->orders_payment_transaction_full = $transactionDetails['fulljson'];
              }
              self::OrderPaymentSave($op);
              $ret = true;
            }
        }
        if (!$this->isInstance() || empty($order->info['orders_id'])) {
          $this->getOrderInstanceWithId($op->orders_payment_order_id);
        }

      } else {

        $parent_id = 0;
        $replaceOrderInstance = false;
        if (!empty($transactionDetails['parent_transaction_id'])) {
          $pp = self::OrderPaymentSearchRecord($this->code, $transactionDetails['parent_transaction_id']);
          if ($pp) {
            $parent_id = $pp->orders_payment_id;
          }
        }
        if (!empty($transactionDetails['orders_id'])) {
          $order = new order($transactionDetails['orders_id']);
          $order->info['orders_id'] = $transactionDetails['orders_id'];
          $replaceOrderInstance = true;
        } elseif ($pp) {
          $order = new order($pp->orders_payment_order_id);
          $order->info['orders_id'] = $pp->orders_payment_order_id;
          $replaceOrderInstance = true;
        }
        if (!empty($order->info['orders_id'])) {
            $data = [
                'orders_payment_id_parent' => $parent_id,
                'orders_payment_order_id' => $order->info['orders_id'],
                'orders_payment_module' => $transactionDetails['payment_class']??($this->code??''),
                //'orders_payment_module_name'
                'orders_payment_is_credit' => 0,
                'orders_payment_status' =>  (int)$transactionDetails['status_code'],
                'orders_payment_amount' => $amount,
                'orders_payment_currency' => $order->info['currency'],
                'orders_payment_currency_rate' => (float)$order->info['currency_value'],
              //
                'orders_payment_transaction_id' => $transaction_id,
                'orders_payment_transaction_status' => $status,
                'orders_payment_transaction_commentary' => $comments,
                'orders_payment_transaction_date' => (isset($transactionDetails['date']) ? $transactionDetails['date'] : '0000-00-00 00:00:00'),
                'orders_payment_transaction_full' => $transactionDetails['fulljson'],
                'deferred' => $transactionDetails['deferred']??0
                
              ];
              
              if (!empty($order->info['payment_method']) ) {
                   $data['orders_payment_module_name'] = $order->info['payment_method'];
              }
              self::OrderPaymentSave( $data );
        }

      }
      if (!empty($order->info['orders_id'])) {
        if ($replaceOrderInstance ?? null) {
          $this->getOrderInstanceWithId($order->info['orders_id']);
        }
      }

      return $ret;
    }

    public static function OrderPaymentSave($data) {
        global $login_id;
        if (!is_array($data)) {
            $data = (array)$data;
        }
        if (empty($data['orders_payment_id'])) {
            $data['orders_payment_date_create'] = date('Y-m-d H:i:s');
            $data['orders_payment_admin_create'] = (int)$login_id;
            tep_db_perform('orders_payment', $data);
            
        } else {
            $data['orders_payment_admin_update'] = (int)$login_id;
            tep_db_perform('orders_payment', $data, 'update', "orders_payment_id='" . $data['orders_payment_id'] . "'");
        }
    }

    public static function OrderPaymentSearchRecord($class, $id) {
        $res = tep_db_fetch_array(tep_db_query("select * from orders_payment where orders_payment_module='" . tep_db_input($class) . "' and orders_payment_transaction_id='" . tep_db_input($id) . "'"));

        if (empty($res)) {
            $res = new stdClass();
            $res->orders_payment_module = $class;
            $res->orders_payment_transaction_id = $id;
            $res->orders_payment_status = self::OPYS_PENDING;
            global $login_id;
            $res->orders_payment_admin_create = (int)$login_id;

        } else {
            $res = (object)$res;
        }
        return $res;

    }

    public function getBillingIso2() {
        $addr = $this->getBillingAddress();
        return $addr['country']['countries_iso_code_2']??false;

    }

    /**
     * shipping module availability is checked in constructor. Shipping module should be recreated to change status on address update
     * (initial order is created with empty address)
     * @global string $language
     * @global string $PHP_SELF
     * @param string $module
     */
    public function update_shipping_modules($module = '') {
      global $language, $PHP_SELF;

      if (defined('MODULE_SHIPPING_INSTALLED') && tep_not_null(MODULE_SHIPPING_INSTALLED)) {
        $modules = explode(';', MODULE_SHIPPING_INSTALLED);

        $include_modules = array();

        if ( (tep_not_null($module)) && (in_array(substr($module['id'], 0, strpos($module['id'], '_')) . '.' . substr($PHP_SELF, (strrpos($PHP_SELF, '.')+1)), $this->modules)) ) {
          $include_modules[] = array('class' => substr($module['id'], 0, strpos($module['id'], '_')), 'file' => substr($module['id'], 0, strpos($module['id'], '_')) . '.' . substr($PHP_SELF, (strrpos($PHP_SELF, '.')+1)));
        } else {

          foreach($modules as $value) {
            $class = substr($value, 0, strrpos($value, '.'));
            $include_modules[] = array('class' => $class, 'file' => $value);
          }
        }

        for ($i=0, $n=sizeof($include_modules); $i<$n; $i++) {
          include_once(DIR_WS_LANGUAGES . $language . '/modules/shipping/' . $include_modules[$i]['file']);
          include_once(DIR_WS_MODULES . 'shipping/' . $include_modules[$i]['file']);

          $GLOBALS[$include_modules[$i]['class']] = new $include_modules[$i]['class'];
        }
      }
    }


/**
 * replace sendto/billto arrays with addressbook ids in it (or reset if array doesn't contain it)
 * @global int|array $billto
 * @global int|array $sendto
 */
    public function scalarToAddresses() {
        global $billto, $sendto;
        if (is_array($billto)) {
            if (!empty($billto['address_book_id'])) {
                $billto = $billto['address_book_id'];
            } else {
                $billto = null;
            }
        }
        if (is_array($sendto)) {
            if (!empty($sendto['address_book_id'])) {
                $sendto = $sendto['address_book_id'];
            } else {
                $sendto = null;
            }
        }
    }

}
