<?php
/**
 * Transactional Midle Ware for Paypal modules
 * This file is part of osCommerce ecommerce platform.
 * osCommerce the ecommerce
 *
 * @link https://www.oscommerce.com
 * @copyright Copyright (c) 2000-2022 osCommerce LTD
 *
 * Released under the GNU General Public License
 * For the full copyright and license information, please view the LICENSE.TXT file that was distributed with this source code.
 */

namespace PaypalPartner;

class ExtraConfig {

    /** @prop models\SellerInfo $seller */
    public $seller;
    /** @prop paypal_rest $module */
    public $module;
    public $url;
    public $errors;
    public $messages;
    public $mode = 'form';
    public $link_params;

    public function draw($settings) {
        foreach($settings as $k => $setting) {
            $this->$k = $setting;
        }
        $this->run();
    }

    public function run() {

        $ret = '';
        $next = ($_GET['ppp_next']??0);
        $urlParams = ['ext/modules/payment/paypal_rest/index.php']; //the path is the same for both admin and frontend callbacks (see 3rd param in create URL method
        $mode = $this->module->getMode();

        $unBoardLink = $this->module->createAbsoluteUrl($urlParams + ['action' => 'deleteSeller', 'ck' => $this->seller->tracking_id], null);
        $whSubscribeLink = $this->module->createAbsoluteUrl($urlParams + ['action' => 'subscribeSeller', 'ck' => $this->seller->tracking_id], null);

        $checkBoardLink = $this->module->createAbsoluteUrl($urlParams + ['action' => 'checkOnBoarded', 'result' => 'show'], null, true);
        $checkLink = $this->module->createAbsoluteUrl($urlParams + ['action' => 'processOnBoard', 'subaction'=> 'checkDetails', 'result' => 'show'], null);
        $saveExtraUrl = $this->module->createAbsoluteUrl($urlParams + ['action' => 'saveExtra'], null);


        //$fetchKeysUrl = $this->module->createAbsoluteUrl(['action' => 'sellerDetails'], null, true);

        $base_url = $cor = false;
        $base_url = HTTPS_SERVER;


        $ownAPILink = $this->module->createAbsoluteUrl([FILENAME_MODULES, 'set' => 'payment', 'module' => $this->module->code, 'ppp_next' => 4, 'test_mode' => ($mode != 'Live')]) . '#extra';
        $editLink = $this->module->createAbsoluteUrl([FILENAME_MODULES, 'set' => 'payment', 'module' => $this->module->code]) . '#extra';

        if ($this->mode == 'form' && !empty($this->seller->payer_id) && $this->seller->status>0 ) {
            $tmp = $this->module->getBoardingDetails($this->seller);
        }
        $ccpActive = false;
        $sellerBoardingDetails = [];

        if (is_array($tmp ?? null)) {
            foreach (['errors', 'warnings', 'info'] as $group) {
                if (!empty($tmp[$group])) {
                    $sellerBoardingDetails[$group] = $tmp[$group];
                }
            }
            $ccpActive = $this->module->CCPActive($this->seller);
            $seller3DS = $this->module->get3DSSettings($this->seller);
        }
        //1 - sandbox 2 - live 3 - both partner's keys
        $installPPP = $this->module->getInstallOptions(true);
        $boardingActive = (($mode != 'Live' && in_array($installPPP, [1,3])) || ($mode == 'Live' && in_array($installPPP, [2,3])));

        $whListNames = $this->module->getSubscribedWebHooks();

        if (empty($whListNames)) {
            $whListNames = false;
        }

        $neededWebHooks = $this->module->getNeededWebHooks();

        if (is_array($whListNames) && !empty($whListNames) && is_array($whListNames) && is_array($neededWebHooks)) {
            $subEvents = array_diff($neededWebHooks, $whListNames);
        } else {
            $subEvents = $neededWebHooks;
        }

        switch ($this->mode) {
            case 'form':
                foreach([
                  'seller' => $this->seller,
                  'boardViaLink' => $this->module->boardViaLink(),
                  'boardingActive' => $boardingActive,
                  'platform_base_url' => $base_url,
                  'module' => $this->module,//ok
                  'errors' => $this->errors,
                  'messages' => $this->messages,
                  'gologin' => ($next == 1),
                  'checkBoardLink' => $checkBoardLink,
                  'unBoardLink' => $unBoardLink,
                  'ownAPILink' => $ownAPILink,
                  'editLink' => $editLink,
                  'sellerBoardingDetails' => $sellerBoardingDetails,//ok
                  'seller3DS' => $seller3DS,//ok
                  'ccpActive' => $ccpActive,//ok
                  'ownAPIDetails' => ((!empty($this->seller->own_client_id) && !empty($this->seller->own_client_secret)) || $next == 4),
                  'boardingMode' => 1,
                  //'fetchKeysUrl' => $fetchKeysUrl,
                  'webhooksList' => $whListNames,
                  'missedWh' => !empty($subEvents),//ok
                  'webhooksRequiredList' => $neededWebHooks,//ok
                  'whSubscribeLink' => $whSubscribeLink,//ok
                  'ppp_mode' => ($this->module->getPartnerId('Sandbox') == $this->seller->partner_id)?'Sandbox':'Live',//ok
                  'checkVal' => md5(DB_SERVER_PASSWORD),//ok
                ] as $k =>$v) {
                    $$k = $v;
                }
                include('views/extra.php');
                break;
            case 'info':
                $ret = $this->render('info', [
                  'errors' => $this->errors,
                  'messages' => $this->messages,
                  'boardingMode' => 1,
                  //'fetchKeysUrl' => $fetchKeysUrl,
                ]);
                break;
            /*case 'signup': // grant permissions
                $ret = $this->render('signup', [
                  'url' => $this->url,
                  'errors' => $this->errors,
                  'boardingMode' => $this->module::BOARDING_MODE,
                  //'fetchKeysUrl' => $fetchKeysUrl,
                ]);
                break;
            case 'signupLink': // grant permissions
                $ret = $this->render('signup_link', [
                  'errors' => $this->errors,
                  'boardingMode' => $this->module::BOARDING_MODE,
                  'link_params' => $this->link_params,
                  //'fetchKeysUrl' => $fetchKeysUrl,
                  'mode' => $mode
                ]);
                break;
            case 'login':
                $ret = $this->render('extra', [
                  'login' => 1,
                  'seller' => $this->seller,
                  'cor' => $cor,
                  'platform_base_url' => $base_url,
                  'address' => $this->address,
                  'module' => $this->module,
                  'errors' => $this->errors,
                  'messages' => $this->messages,
                  'boardingMode' => $this->module::BOARDING_MODE,
                  //'fetchKeysUrl' => $fetchKeysUrl,
                ]);
                break;*/
        }
        return $ret;
    }


    public function saveExtra($seller) {
        $seller->load($_POST);
        if (!empty($seller->own_client_id) && !empty($seller->own_client_secret) && !empty($seller->payer_id)) {
            $seller->is_onboard = 1;
            $seller->status = 1;
        }

        $a3dsOn = ($_POST['paypal_partner_ccp_status']??0);
        $a3dsSettings = [];
        if ($a3dsOn) {
            $a3dsSettings['status'] = 0;
            foreach ($this->module::$threeDSDefaults as $k => $v) {
                $a3dsSettings[$k] = ($_POST['paypal_partner_3ds_' . $k]??0);
            }
        }
        $seller->paypal_partner_ccp_status = $a3dsOn;
        $seller->three_ds_settings = json_encode($a3dsSettings);

        $seller->save(false);

        // cleanup
        models\SellerInfo::deleteAll(
          ' partner_id="' . $seller->partner_id . '"' .
          ' and platform_id=' . (int)$seller->platform_id .
          ' and psi_id<>' . (int)$seller->psi_id
        );

    }

    public function extra_params() {
        $psi_id = ($_POST['psi_id']??($_GET['psi_id']??false));
        if ($psi_id) {
            $seller = $this->module->getSellerById((int)$psi_id);
        } else {
            $seller = $this->module->getSeller();
        }

        if ($_SERVER['REQUEST_METHOD'] === 'POST') {
            self::saveExtra($seller);
        } else {
            if (empty($seller->psi_id)) {
                try {
                    $seller->save(false); // save tracking id else won't be able to get boarding results.
                } catch (\Exception $e) {
                    $this->module::error('seller tracking id is not saved ' . $e->getMessage(), 'paypal_partner');
                }
                if (defined('STORE_OWNER_EMAIL_ADDRESS')) {
                    $seller->email_address = STORE_OWNER_EMAIL_ADDRESS;
                }
            }
        }

        return $this->draw([
              'module' => $this->module,
              'seller' => $seller,
              'errors' => $this->errors,
        ]);
    }



}
