<?php

namespace PaypalPartner\models;

use Yii;

 
class Customer extends \stdClass
{
    private static $attr = [
            'customers_id',
            'customers_gender',
            'customers_firstname',
            'customers_lastname',
            'customers_dob',
            'customers_email_address',
            'customers_default_address_id',
            'customers_telephone',
            'customers_fax',
            'customers_password',
            'customers_newsletter'
      ];
    private $dummy_address_id = null;

    public function __construct($config=[]) {
        $this->load($config);
        return $this;
    }
    
    public function load($config) {
        foreach (self::$attr as $a ) {
            $this->$a = (isset($config[$a])?$config[$a]: (!empty($this->$a)?$this->$a:'') );
        }
        return $this;
    }

    public function loadDefaultValues() {

    }

    public static function tableName()
    {
        return TABLE_CUSTOMERS;
    }

    public function beforeSave($newRecord) {
        $ret = true;
        if (empty($this->customers_email_address)) {
            $ret = false;
        }
        return $ret;
    }

    public function setAttribute($attr, $val) {
        $this->$attr = $val;
        return $this;
    }

    public function deleteAll($where = '') {
        tep_db_query("delete from " . self::tableName() . " where 1 " . (!empty($where)?'and ' . $where : ''));
    }

    public function save($validate=false) {
        //$validate always ignored
        //record could be deleted add check
        $newRecord = empty($this->customers_id);
        if (!$newRecord) {
            $check = tep_db_fetch_array(teep_db_query("select 1 from " . self::tableName() . " where customers_id='" . $this->customers_id . "'"));
            if (empty($check)) {
                $this->customers_id = null;
                $newRecord = true;
            }
        }
        if ($this->beforeSave(empty($this->customers_id)) ) {
            $data = (array)$this;
            $data = array_intersect_key($data, array_flip(self::$attr));

            if (empty($this->customers_id)) {
                unset($data['customers_id']);
                tep_db_perform(self::tableName(), $data);
                $this->customers_id = tep_db_insert_id();

            } else {
                tep_db_perform(self::tableName(), $data, 'update', "customers_id='" . $this->customers_id . "'");
            }
            $res = $this;
            if ($newRecord) {
                // add empty default address book and customers info record
                tep_db_perform(TABLE_ADDRESS_BOOK, array(
                  'customers_id' => $this->customers_id,
                  'entry_country_id' => STORE_COUNTRY,
                  'entry_zone_id' => STORE_ZONE
                ));
                $this->customers_default_address_id = tep_db_insert_id();
                $this->dummy_address_id = $this->customers_default_address_id;
                tep_db_perform(self::tableName(), array(
                        'customers_default_address_id' => $this->customers_default_address_id
                    ),
                    'update', "customers_id='" . $this->customers_id . "'");

                tep_db_perform(TABLE_CUSTOMERS_INFO, array(
                  'customers_info_id' => $this->customers_id,
                  'customers_info_date_account_created' => 'now()',
                  'customers_info_number_of_logons' => 0
	            ));
            }

        } else {
            $res = false;
        }
        return $res;
    }

/**
 * oscok
 * @param string $key
 * @return type
 */
    public function get($key){
        $ret = null;
        if (isset($this->$key)) {
            $ret = $this->$key;
        }
        return $ret;
    }

    public static function findOne($where) {
        $ret = null;
        if (is_array($where)) {
            $tmp = [];
            foreach($where as $k => $v) {
               $tmp[] = $k . "='" . tep_db_input($v) . "'";
            }
            $where = implode(' and ', $tmp);
        }
        $r = tep_db_fetch_array(tep_db_query("select * from " . TABLE_CUSTOMERS . " where 1 and " . $where . " limit 1"));
        if (is_array($r)) {
            $instance = new self;
            $ret = $instance->load($r);
        }
        return $ret;
    }

    public function getAddressBookQuery($where = ''){
        $res = tep_db_query("select * from " . TABLE_ADDRESS_BOOK . " where customers_id='" . (int)$this->customers_id  . "'" .
            (!empty($where)? "and $where":'') );
        return $res;
    }

    public function getAddressBook($abId, $toArray=false){
        $res = tep_db_fetch_array($this->getAddressBookQuery("address_book_id='" . (int)$abId . "'"));
        $country = tep_get_countries_with_iso_codes($res['entry_country_id']);
        $res['country'] = $country;
        $res['country']['id'] = $res['entry_country_id'];
        if (!$toArray) {
            $res = (object)$res;
        }
        return $res;
    }

    public function getDefaultAddress() {

        if ($this->customers_default_address_id>0) {
            $res = $this->getAddressBook($this->customers_default_address_id);
        } 
        if (empty($res)) {
            $res = tep_db_fetch_array($this->getAddressBookQuery());
            if ($res) {
                $res = (object)$res;
            }
        }
        return $res;
    }

    public function findAddress($where) {
        $ret = null;
        if (is_array($where)) {
            $tmp = [];
            foreach($where as $k => $v) {
               $tmp[] = $k . "='" . tep_db_input($v) . "'";
            }
            $where = implode(' and ', $tmp);
        }
        $r = tep_db_fetch_array($this->getAddressBookQuery($where));
        if (is_array($r)) {
            $ret = (object)$r;
        }
        return $ret;
    }

    public function saveAddressBook($data=[]){
        if (!empty($this->customers_id)) {
            $resetDummy = false;
            if (!empty($data['address_book_id'])) {
                $res = tep_db_fetch_array($this->getAddressBookQuery("address_book_id='" . (int)$data['address_book_id'] . "'"));
                if (!empty($res) && !empty($this->dummy_address_id) && $this->dummy_address_id == $data['address_book_id']) {
                    $resetDummy = true;
                }

            } elseif (!empty($this->dummy_address_id)) {
                $res = tep_db_fetch_array($this->getAddressBookQuery("address_book_id='" . (int)$this->dummy_address_id . "'"));
                if (empty($res)) {
                    $this->dummy_address_id = null;
                } else {
                    $resetDummy = true;
                }
            }

            if (empty($res)) {
                $res = [];
            }
            $res['customers_id'] = $this->customers_id;

            foreach (array(
              'address_book_id',
              'customers_id',
              'entry_gender',
              'entry_company',
              'entry_firstname',
              'entry_lastname',
              'entry_street_address',
              'entry_suburb',
              'entry_postcode',
              'entry_city',
              'entry_state',
              'entry_country_id',
              'entry_zone_id'
            ) as $allowedKey) {
                $res[$allowedKey] = (isset($data[$allowedKey])?$data[$allowedKey]: (!empty($res[$allowedKey])?$res[$allowedKey]:'') );
            }
        }

        $ret = false;

        if (!empty($res['customers_id'] . $res['entry_country_id'])) {

            if (empty($res['address_book_id'])) {
                unset($res['address_book_id']);
                tep_db_perform(TABLE_ADDRESS_BOOK, $res);
                $id = tep_db_insert_id();

            } else {
                tep_db_perform(TABLE_ADDRESS_BOOK, $res, 'update', "address_book_id='" . $res['address_book_id'] . "'");
                if ($resetDummy) {
                    $this->dummy_address_id = null;
                }
                $id = $res['address_book_id'];
            }
            $res = tep_db_fetch_array($this->getAddressBookQuery("address_book_id='" . (int)$id . "'"));
            if ($res) {
                $ret = (object)$res;
            }
        }
        return $ret;
    }

}
