<?php
/**
 * Transactional Midle Ware for Paypal modules
 * This file is part of osCommerce ecommerce platform.
 * osCommerce the ecommerce
 *
 * @link https://www.oscommerce.com
 * @copyright Copyright (c) 2000-2022 osCommerce LTD
 *
 * Released under the GNU General Public License
 * For the full copyright and license information, please view the LICENSE.TXT file that was distributed with this source code.
 */
namespace PaypalPartner\models;

use Yii;

/**
 * This is the model class for table "paypal_seller_info".
 *
 * @property int $psi_id
 * @property int $platform_id
 * @property string $partner_id
 * @property float $fee_percent
 * @property string $tracking_id
 * @property string $payer_id
 * @property string $entry_company
 * @property string $email_address
 * @property string $entry_firstname
 * @property string $entry_lastname
 * @property string $entry_street_address
 * @property string $entry_suburb
 * @property string $entry_postcode
 * @property string $entry_city
 * @property string $entry_state
 * @property int $entry_country_id
 * @property int $entry_zone_id
 * @property int $is_onboard
 * @property string $entry_telephone
 * @property int $fee_editable flag
 * @property string $own_client_id
 * @property string $own_client_secret
 * @property text $boarding_json
 * @property date $boarding_date
 */
 
class SellerInfo extends \stdClass
{
    private static $attr = [
            'psi_id',
            'platform_id',
            'partner_id',
            'fee_percent',
            'tracking_id',
            'payer_id',
            'entry_company',
            'email_address',
            'entry_firstname',
            'entry_lastname',
            'entry_street_address',
            'entry_suburb',
            'entry_postcode',
            'entry_city',
            'entry_state',
            'entry_country_id',
            'entry_zone_id',
            'is_onboard',
            'entry_telephone',
            'fee_editable',
            'own_client_id',
            'own_client_secret',
            'status',
            'paypal_partner_ccp_status',
            'three_ds_settings',
            'boarding_json',
            'boarding_date'];

    public function __construct($config) {
        $this->load($config);
        return $this;
    }
    
    public function load($config) {
        foreach (self::$attr as $a ) {
            $this->$a = (isset($config[$a])?$config[$a]: (!empty($this->$a)?$this->$a:'') );
        }
        return $this;
    }

    public function loadDefaultValues() {

    }

    public static function tableName()
    {
        return 'paypal_seller_info';
    }
    
    public static function generateTrackingId(){
        if (function_exists('random_int') && function_exists('mb_strlen')) {
            $ret = self::random_str(127);
        } else {
            $ret = sha1(uniqid());
        }
        return $ret;
    }

    private static function random_str(int $length = 64, string $keyspace = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-'): string
    {
      if ($length < 1) {
          throw new \RangeException("Length must be a positive integer");
      }
      $pieces = [];
      $max = mb_strlen($keyspace, '8bit') - 1;
      for ($i = 0; $i < $length; ++$i) {
          if (function_exists('random_int')) {
            $pieces []= $keyspace[random_int(0, $max)];
          } else {
            $pieces []= $keyspace[rand(0, $max)];
          }
      }
      return implode('', $pieces);
    }

    public function isOnBoarded(){
        return !!$this->is_onboard;
    }
    
    public function updateMerchantId($merchantId){
        if ($this->psi_id){
            $this->setAttribute('payer_id', $merchantId);
            if (!$this->save()){
                throw new \Exception("Seller: payer_id can't be updated");
            }
            return true;
        }
        trigger_error("updateMerchantId Invalid Seller");
    }
    
    public function setOnBoarded(){
        if ($this->psi_id){
            $this->setAttribute('is_onboard', 1);
            if (!$this->save()){
                throw new \Exception("Seller can't be saved as onBoarded");
            }
            return true;
        }
        trigger_error("setOnBoarded Invalid Seller");
    }

    public function setAttribute($attr, $val) {
        $this->$attr = $val;
        return $this;
    }

    public function deleteAll($where = '') {
        tep_db_query("delete from " . self::tableName() . " where 1 " . (!empty($where)?'and ' . $where : ''));
    }

    public function beforeSave($insert) {
        if ($insert || empty($this->tracking_id)) {
            $this->tracking_id = $this->random_str(127);
        }
        return true;
    }

    public function save($validate=false) {
        //$validate always ignored
        if ($this->beforeSave(empty($this->psi_id)) ) {
            $data = (array)$this;
            $data = array_intersect_key($data, array_flip(self::$attr));

            if (empty($this->psi_id)) {
                unset($data['psi_id']);
                tep_db_perform(self::tableName(), $data);
                $this->psi_id = tep_db_insert_id();

            } else {
                tep_db_perform(self::tableName(), $data, 'update', "psi_id='" . $this->psi_id . "'");
            }
            $res = $this;
            
        } else {
            $res = false;
        }
        return $res;
    }
}
