<?php

use PaypalPartner\api\Merchant;

trait PaypalPartnerTrait {
    /**
     * oscok
     * @return PaypalPartner\ExtraConfig
     */
    public function getExtraConfigClass(){
        return PaypalPartner\ExtraConfig::class;
    }

    /**
     * oscok
     * @return bool
     */
    protected function liveConfigurationExists(){
        $ret = false;
        $partnerId = self::PARTNER_MERCHANT_ID;

        $seller = tep_db_fetch_array(tep_db_query("select * from paypal_seller_info where partner_id = '" . $partnerId . "' and status=1 "));
        $ret = !empty($seller['is_onboard']);
        return $ret;
    }

    /**
     * oscok
     * @return bool
     */
    protected function ownSandboxConfigExists(){
        $ret = false;
        $partnerId = self::PARTNER_MERCHANT_SANDBOX_ID;
        $seller = tep_db_fetch_array(tep_db_query("select * from paypal_seller_info where partner_id = '" . $partnerId . "' and status=1 "));
        $ret = (
                !empty($seller['is_onboard'])
            &&  !empty($seller['own_client_id'])
            &&  !empty($seller['own_client_secret'])
            &&  $seller->own_client_id !== 'Ad1UJlMj9BKThV6rWyktLD5ggqp2uL4iFnFtYIKwAcKULWu3OoaGHw4NO2v8bmx86-6RIcWB_LoSOGsr'
            );
        return $ret;
    }

    /**
     * oscok
     * @param int $id
     * @return PaypalPartner\models\SellerInfo
     */
    public function getSellerById($id){
        $seller_array = tep_db_fetch_array(tep_db_query("select * from paypal_seller_info where psi_id = '" . $id . "' /*and status=1 */"));
        if (!empty($seller_array)){
            $seller = new PaypalPartner\models\SellerInfo($seller_array);
        } else {
            $seller = $this->getSeller();
        }
        return $seller;
    }

/**
 * oscok
 * @param string $force force mode
 * @param bool $forceNew
 * @return \paypal_rest\PaypalPartner\models\SellerInfo
 */
    public function getSeller($force = false, $forceNew = false){
        $partnerId = $this->getPartnerId($force); // depends on sandbox/live mode
        $seller = null;
        $seller_array = tep_db_fetch_array(tep_db_query("select * from paypal_seller_info where partner_id = '" . $partnerId . "' and status='" . ($forceNew?-1:1) . "' "));
        if (!empty($seller_array)){
            $seller = new PaypalPartner\models\SellerInfo($seller_array);
        }
        if (!$seller || ($forceNew && $seller->status != -1) ){
            $seller = new PaypalPartner\models\SellerInfo([
              'partner_id' => $partnerId,
              'fee_percent' => 0,
              'tracking_id' => PaypalPartner\models\SellerInfo::generateTrackingId(),
              'is_onboard' => 0,
              'status' => -1
              ]);
            if (empty($seller->psi_id)) {
                $res = $seller->save();
            }
        }

        return $seller;
    }
    
    public function getInstaller(){
        return new PaypalPartner\Installer();
    }

 /**
  * create partner referral structure to SIGN-IN existing PayPal seller.
  * @param PaypalPartner\models\SellerInfo $seller
  * @throws \Exception
  */
    public function signinPartner(PaypalPartner\models\SellerInfo $seller){
      if (!$seller->tracking_id) {
          throw new \Exception("Tracking ID is undefined");
      }
      /*'{
    "operations": [
      {
        "operation": "API_INTEGRATION",
        "api_integration_preference": {
          "rest_api_integration": {
            "integration_method": "PAYPAL",
            "integration_type": "FIRST_PARTY",
            "first_party_details": {
              "features": [
                "PAYMENT",
                "REFUND"
              ],
              "seller_nonce": " Seller-Nonce"
            }
          }
        }
      }
    ],
    "products": [
      "EXPRESS_CHECKOUT"
    ],
    "legal_consents": [
      {
        "type": "SHARE_DATA_CONSENT",
        "granted": true
      }
    ]
}'

       */
      try {
        $partner = new PaypalPartner\api\Partner();

        //$partner->setPartnerConfigOverride(new PaypalPartner\api\PartnerConfigOverride());
        $partner->setTrackingId($seller->tracking_id);

        $integration = new PaypalPartner\api\Integration(["integration_method" => "PAYPAL"]);
        $integration->setFirstPartyDetails(new PaypalPartner\api\FirstPartyDetails(['seller_nonce' => $seller->tracking_id]));  //self::random_str(100)

        $preference = new PaypalPartner\api\Preference();
        $preference->setRestApiIntegration($integration);

        $operation = new PaypalPartner\api\Operation();
        $operation->setApiIntegrationPreference($preference);

        $partner->setOperations($operation);
      } catch (\Exception $ex) {
        $this->sendDebugEmail($ex);
        return false;
      }

      return $partner;


    }
    private static function random_str(int $length = 64, string $keyspace = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'): string
    {
      if ($length < 1) {
          throw new \RangeException("Length must be a positive integer");
      }
      $pieces = [];
      $max = mb_strlen($keyspace, '8bit') - 1;
      for ($i = 0; $i < $length; ++$i) {
          if (function_exists('random_int')) {
            $pieces []= $keyspace[random_int(0, $max)];
          } else {
            $pieces []= $keyspace[rand(0, $max)];
          }
      }
      return implode('', $pieces);
    }

    /**
     * create partner referral structure to SIGN-UP new PayPal seller.
     * @param \paypal_rest\PaypalPartner\models\SellerInfo $seller
     * @param string $seller_type allowed ['i', 'b']
     * @return \paypal_rest\PaypalPartner\api\Partner
     * @throws \Exception
     */
    public function signupPartner(PaypalPartner\models\SellerInfo $seller, $seller_type = 'b'){

        if (!$seller->tracking_id) {
            throw new \Exception("Tracking ID is undefined");
        }

        if (!in_array($seller_type, ['i', 'b'])) {
            throw new \Exception("Incorrect seller type '{$seller_type}' allowed ['i', 'b']");
        }

        global $languages_id;
        //$country = \common\helpers\Country::get_countries($seller->entry_country_id, true);
        $country = \common\models\Countries::findOne(['countries_id' => $seller->entry_country_id, 'language_id' => $languages_id])->toArray();
        $address = new PaypalPartner\api\Address([
            'address_line_1' => $seller->entry_street_address,
            'address_line_2' => $seller->entry_suburb,
            'admin_area_2' => $seller->entry_city,
            'admin_area_1' => $seller->entry_state,
            'postal_code' => $seller->entry_postcode,
            'country_code' => $country['countries_iso_code_2'],
        ]);
        $address->setType(PaypalPartner\api\PartnerConstants::ADDRESS_TYPE_HOME);
        
        $partner = new PaypalPartner\api\Partner();
        
        $iOwner = new PaypalPartner\api\IndividualOwners();
        $iOwner->setNames(new PaypalPartner\api\Name([
            'given_name' => $seller->entry_firstname,
            'surname' => $seller->entry_lastname,
        ]));

        $iOwner->setAddresses($address);
// not in SDK
        /**/
        $_pn = trim(preg_replace('/[^0-9]/', '', $seller->entry_telephone), '0');
        $_cdp = substr(trim($country['dialling_prefix'], ' +'), 0, 3);
        $sellerPhoneOk = false;
        if (strlen($_pn)>0 && strlen($_pn)<15 && strlen($_cdp)>0 && strlen($_cdp.$_pn)<16 ) {
            $sellerPhoneOk = true;
            $iOwner->phones = [(object)[
              'national_number' => $_pn,
              'country_code' => $_cdp
            ]];
        }
        /**/
        $iOwner->setCitizenship($country['countries_iso_code_2']);
        
        $partner->setIndividualOwners($iOwner);

        if ($seller_type!='i') {
          $bEntity = new PaypalPartner\api\BusinessEntity();
          $bEntity->setBusinessType(new PaypalPartner\api\BusinessType);
          $bAddress = new PaypalPartner\api\Address([
              'address_line_1' => $seller->entry_street_address,
              'address_line_2' => $seller->entry_suburb,
              'admin_area_2' => $seller->entry_city,
              'admin_area_1' => $seller->entry_state,
              'postal_code' => $seller->entry_postcode,
              'country_code' => $country['countries_iso_code_2'],
          ]);
          $bAddress->setType(PaypalPartner\api\PartnerConstants::ADDRESS_TYPE_WORK);
          $bEntity->setAddresses($bAddress);
          $bEntity->setWebsite($this->createAbsoluteUrl(['/'], 'SSL', true));

  // not in SDK
          $bEntity->names = [(object)[
            'business_name' => $seller->entry_company,
            "type" => "LEGAL_NAME"
          ]];
          if ($sellerPhoneOk) {
            $bEntity->phones = [(object)[
              'national_number' => $_pn,
              'country_code' => $_cdp
            ]];
          }

          $bEntity->emails = [(object)[
            'email' => $seller->email_address,
            "type" => "CUSTOMER_SERVICE" // nothing else in ENUM now
          ]];


          $partner->setBusinessEntity($bEntity);
        }
        
        $partner->setEmail($seller->email_address);
        $partner->setPreferredLanguageCode($this->getLocale());
        //$partner->setPreferredLanguageCode(\Yii::$app->language);
        
        $partner->setTrackingId($seller->tracking_id);
        
        //$partner->setPartnerConfigOverride(new PaypalPartner\api\PartnerConfigOverride);
        
        $operation = new PaypalPartner\api\Operation();
        $preference = new PaypalPartner\api\Preference();
        $integration = new PaypalPartner\api\Integration();
        $integration->setThirdPartyDetails(new PaypalPartner\api\ThirdPartyDetails());
        $preference->setRestApiIntegration($integration);
        $operation->setApiIntegrationPreference($preference);
        
        $partner->setOperations($operation);

        return $partner;
    }
        
    public function getMerchant($holbiMerchantId = null){
        $merchant = new PaypalPartner\api\Merchant();
        if ($holbiMerchantId){
            $merchant->setPartnerId($holbiMerchantId);
        }
        return $merchant;
    }

  /**
   * get Webhooks for application
   * @return array|false
   */
  public function getWebHooks() {
    $ret = false;
    try {
      $ret = \PayPal\Api\Webhook::getAllWithParams([], $this->getApiContext());// [ anchor_type => (APPLICATION|ACCOUNT)]
    } catch (\Exception $ex) {
      self::warning("List all webhooks exception " . $ex->getMessage(), $this->code);
    }
    return $ret;
  }

  /**
   * set webhooks
   * @param array $subEvents
   * @return null|array|object
   */
  public function setWebHooks($subEvents) {
    $output = null;
    if (is_array($subEvents) && !empty($subEvents)) {
      $webhook = new \PayPal\Api\Webhook();
      $url = $this->getWebHookUrl();

      $webhook->setUrl($url);

      $webhookEventTypes = [];
      foreach ($subEvents as $event) {
        $webhookEventTypes[] = new \PayPal\Api\WebhookEventType(
          '{"name":"' . $event . '"}'
        );
      }
      $webhook->setEventTypes($webhookEventTypes);

      try {
        $apiContext = $this->getApiContext();
      } catch (\Exception $ex) {
        self::warning($ex->getMessage(), 'paypal_rest');
      }

      try {
        $output = $webhook->create($apiContext);
      } catch (Exception $ex) {
        if ($ex instanceof \PayPal\Exception\PayPalConnectionException) {
          $data = $ex->getData();
          if (strpos($data, 'WEBHOOK_NUMBER_LIMIT_EXCEEDED') !== false) {
            try {
              foreach ($this->getWebHooks() as $dw) {
                $dw->delete($apiContext);
              }
            } catch (\Exception $ex) {
              self::warning($ex->getMessage(), 'paypal_rest');
            }
            try {
              $output = $webhook->create($apiContext);
            } catch (Exception $ex) {
              self::warning($ex->getMessage(), 'paypal_rest');
            }
          }
        }
        if (empty($output)) {
          self::warning($ex->getMessage(), 'paypal_rest');
        }
      }
    }
    return $output;
  }

  public function onProcessWebhook() {
    /** @var String $bodyReceived */
    $bodyReceived = file_get_contents('php://input');
    /** @var Array $headers */
    $headers = array();
    if (function_exists('getallheaders')) {
        $headers = getallheaders();
    } elseif (function_exists('apache_request_headers')) {
        $headers = apache_request_headers();
    } else {
       foreach ($_SERVER as $name => $value) {
           if (substr($name, 0, 5) == 'HTTP_')
           {
               $headers[str_replace(' ', '-', ucwords(strtolower(str_replace('_', ' ', substr($name, 5)))))] = $value;
           }
       }
    }
    /*success     [Correlation-Id] => d2bfc31ba6ce1
    [User-Agent] => PayPal/AUHR-214.0-54656248
    [Content-Type] => application/json
    [Paypal-Auth-Algo] => SHA256withRSA
    [Paypal-Cert-Url] => https://api.sandbox.paypal.com/v1/notifications/certs/CERT-360caa42-fca2a594-1d93a270
    [Paypal-Auth-Version] => v2
    [Paypal-Transmission-Sig] => ZTCNDTsZOWPyfgNsx0qhylVrxuqUjTVMcxTOInOhBxaD7lxQm5hLNOIlAPA+R1+UVilQxgaEIM/m2V+s+XcuUgVDfqHL9NZUNyz9V4wwZo+RERXcvoVpLvPI9s+WbcNA1IWxSNNyWBVAVKQkcsuaDkU2IWYmK+rCoBVI+/oG+vhTA32ulZOWxrQCnfAbDXmRwW3qbN2kIn1L
11h7YIfG7W+k5z9xciWOL//ZHXsn5Y/tiYkl/n//NfwTdxkxYqewpXRrZK1KzDw2oBAzdn2vADodHIVn1F2no2CRiRqaNYB0wG+S6YDmIg2gHVPzqwoeU63nvjZvTb/EOrQC/7/biQ==
    [Paypal-Transmission-Time] => 2020-08-05T14:52:27Z
    [Paypal-Transmission-Id] => 47964f20-d72b-11ea-9349-29bf1500b5a9
    [Accept] => *<>/*
    [Content-Length] => 1597
    [Connection] => close
    [X-Forwarded-Proto] => https
    [X-Forwarded-For] => 173.0.82.126
    [Host] => dev5.trueloaded.co.uk
*/

    $whDetails = json_decode($bodyReceived);

    if (/*1 ||*/ $this->whVerify($headers, $bodyReceived, $whDetails)) {
      self::debug(print_r($whDetails, true));

      switch ($whDetails->event_type) {
        case "MERCHANT.ONBOARDING.COMPLETED":
          $this->whBoardingComplete($whDetails);
          break;
        case "MERCHANT.PARTNER-CONSENT.REVOKED":
          $this->whBoardingRevoked($whDetails);
          break;
        case 'PAYMENT.CAPTURE.COMPLETED':
        case 'CHECKOUT.ORDER.APPROVED':
          $this->whCaptured($whDetails);
          break;
        case 'PAYMENT.AUTHORIZATION.VOIDED':
        case 'PAYMENT.CAPTURE.DENIED':
        case 'PAYMENT.CAPTURE.REFUNDED':
        case 'PAYMENT.CAPTURE.REVERSED':
        case 'CHECKOUT.PAYMENT-APPROVAL.REVERSED':
          $this->whCancelled($whDetails);
          break;
        case 'PAYMENT.REFERENCED-PAYOUT-ITEM.COMPLETED':
        case 'PAYMENT.REFERENCED-PAYOUT-ITEM.FAILED':
          //self::warning(" #### " .print_r($whDetails, 1), 'TLDEBUG');
          break;
        case 'CUSTOMER.DISPUTE.CREATED':
        case 'CUSTOMER.DISPUTE.UPDATED':
        case 'CUSTOMER.DISPUTE.RESOLVED':
          $this->whInfo($whDetails);
          break;
      }
    } else {
        self::debug(print_r('whDetails not valid', true));
    }
  }


  protected function whInfo($whDetails) {
    if (is_array($whDetails->resource->disputed_transactions)) { // add info to all transactions/orders
      $processedOrders = [];
      $history = [];
      $history['comments'] = $whDetails->summary;
      $history['comments'] .= "\n" . $whDetails->resource->status;

      foreach ($whDetails->resource->disputed_transactions as $trans) {
        $op = \common\helpers\OrderPayment::searchRecord($this->code, $trans->seller_transaction_id);
        
        if ($op && !empty($op->orders_payment_order_id) && !in_array($op->orders_payment_order_id, $processedOrders)) {
          $processedOrders[] = $op->orders_payment_order_id;
          self::setStatus($op->orders_payment_order_id, 0, $history);
        }
      }
    }
  }

/**
 * 
 * @param stdClass $whDetails transaction details
 * @param string $parent_id PP parent transaction id If empty - suppose webHook and update order details (totals and status) else leave as is (do the same in orders controller)
 */
  protected function whCancelled($whDetails, $parent_id = '') {
    /** @var \common\services\PaymentTransactionManager $tm */
    $tm = $this->manager->getTransactionManager($this);
    if ($this->isInstance() ) {
      $order = $this->getOrderInstance();
    } elseif (!empty($whDetails->resource->invoice_id)) {
        $order = $this->getOrderInstanceWithId($whDetails->resource->invoice_id);
    }
    if (!empty($whDetails->resource)) {
      $res = $whDetails->resource;
    } elseif (!empty($whDetails->result)) {
      $res = $whDetails->result;
    } else {
      $res = $whDetails;
    }
    //$updateOrder = empty($parent_id);
    $updateOrder = true;
    if (empty($parent_id) && !empty($res->links) && is_array($res->links)) {
      foreach ( $res->links as $link) {
        if (!empty($link->rel) && $link->rel=='up') {
          if (!empty($link->href) ) {
            $parent_id = substr($link->href,
                (strpos($link->href, '/captures/')??0)+10);
            if (strpos($parent_id, '?')) {
              $parent_id = substr($parent_id, 0, strpos($parent_id, '?'));
            }
          }
          break;
        }
      }
    }
    $parsed = $this->parseTransactionDetails($whDetails);

    $ret = $tm->updatePaymentTransaction($res->id,
        [
          'fulljson' => json_encode($whDetails),
          'status_code' => self::OPYS_REFUNDED,
          'status' => $res->status,
          'amount' => $res->amount->value,
          'comments'  => ($whDetails->summary??$parsed['comments']),
          'date'  => date('Y-m-d H:i:s', strtotime($res->update_time)),
          'parent_transaction_id' => $parent_id,
          'orders_id' => (($order && $order->info['orders_id'])?$order->info['orders_id']:0)
        ]);
    
    self::updatePaidTotalsAndNotify();

  }

/**
 * @deprecated 1 should be enough
 * 2check if transaction ID is the same as capture then refunded total could be incorrect
 * @param obj $whDetails
 */
  protected function whReversed($whDetails) {
    if ($this->isInstance() ) {
      $order = $this->getOrderInstance();
    }
    $ret = $this->updatePaymentTransaction($whDetails->resource->id,
        [
          'fulljson' => json_encode($whDetails),
          'status_code' => self::OPYS_REFUNDED,
          'status' => $whDetails->resource->status,
          'amount' => $whDetails->resource->amount->value,
          'comments'  => $whDetails->summary,
          'date'  => date('Y-m-d H:i:s', strtotime($whDetails->resource->update_time)),
          'parent_transaction_id' => $whDetails->resource->id,
          'orders_id' => (($order && $order->info['orders_id'])?$order->info['orders_id']:0)
        ]);
    if ($this->isInstance() ) {
      $order = $this->getOrderInstance();
    }

    if ($order && $order->info['orders_id']) {
      //$order->update_piad_information(true);
      //$order->save_details();
      /// order matters
      if (is_numeric(MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID) && ( MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID > 0 )) {
        $history = [];
        if ($ret)  {
          $history['comments'] = $whDetails->summary;
        }
        self::setStatus($order->info['orders_id'], MODULE_PAYMENT_PAYPAL_REST_TRANSACTIONS_ORDER_STATUS_ID, $history);

      }
    }
  }

  protected function whCaptured($whDetails, $parent_id = '') {
    if ($this->isInstance() ) {
      $order = $this->getOrderInstance();
    } elseif (!empty($whDetails->resource->invoice_id)) {
        $order_id = (int)$whDetails->resource->invoice_id;
        $order = $this->getOrderInstanceWithId($order_id);
    }
    if (!empty($whDetails->resource)) {
      $res = $whDetails->resource;
    } elseif (!empty($whDetails->result)) {
      $res = $whDetails->result;
    } else {
      $res = $whDetails;
    }
    $updateOrder = empty($parent_id); // web hook - requires to update order, in admin - it'll be done in controller
    $updateOrder = true;
    if (empty($parent_id) && !empty($res->links) && is_array($res->links)) {
      foreach ( $res->links as $link) {
        if (!empty($link->rel) && $link->rel=='up') {
          if (!empty($link->href) ) {
            $parent_id = substr($link->href,
                (strpos($link->href, '/authorizations/')??0)+strlen('/authorizations/'));
            if (strpos($parent_id, '?')) {
              $parent_id = substr($parent_id, 0, strpos($parent_id, '?'));
            }
          }
          break;
        }
      }
    }
    $parsed = $this->parseTransactionDetails($whDetails);
    self::debug(print_r($parsed, true), 'parsed parent transaction ' . $parent_id);

    $ret = $this->updatePaymentTransaction($res->id,
        [
          'fulljson' => json_encode($whDetails),
          'status_code' => self::OPYS_SUCCESSFUL,
          'status' => $res->status,
          'amount' => $res->amount->value,
          'comments'  => ($whDetails->summary??$parsed['comments']),
          'date'  => date('Y-m-d H:i:s', strtotime($res->update_time)),
          'parent_transaction_id' => $parent_id,
          'orders_id' => (($order && $order->info['orders_id'])?$order->info['orders_id']:0)
        ]);
    if ($this->isInstance() ) {
      $order = $this->getOrderInstance();
    }

    if ($updateOrder && $order && !empty($order->info['orders_id'])) {
        $_comments = ($whDetails->summary??$parsed['comments']);
        self::updatePaidTotalsAndNotify($_comments);
/** /
      $updated = $order->updatePaidTotals();
      if ($updated) { //update order status and notify customer if required
        $status = '';
        if (isset($updated['paid']) ) {
          //if ($updated['details']['status']>0) {// has due
          if (abs(
              round($updated['details']['total'], 2)-
              round($updated['details']['debit'], 2)
              ) < 0.01) {
            $status = $this->paidOrderStatus();
          } else {
            $status = $this->partlyPaidOrderStatus();
          }
        } elseif (isset($updated['refund']) && $updated['details']['credit']>0) {
          if (abs(
              round($updated['details']['total'], 2)-
              round($updated['details']['credit'], 2)
              ) < 0.01) {
            $status = $this->refundOrderStatus();
          } else {
            $status = $this->partialRefundOrderStatus();
          }
        } else {// auth orders have full paid amount - not changed
          $orderPaymentStatusArray = \common\helpers\OrderPayment::getTotalStatusArray($order->getOrderId(), round($order->info['total_inc_tax'],2));
          if (abs(
              round($orderPaymentStatusArray['total'], 2)-
              round($orderPaymentStatusArray['debit'], 2)
              ) < 0.01) {
            $status = $this->paidOrderStatus();
          } else {
            $status = $this->partlyPaidOrderStatus();
          }
        }
        if (1 && !empty($status) && $status != $order->info['order_status']) {
          $order->update_status_and_notify($status);
        }

      }
      /**/
    }

  }

  private function whVerify($headers, $requestBody, $whDetails) {
/**/    $whList = $this->getWebHooks();
    $tmp = parse_url($this->getWebHookUrl());
    $url = $tmp['host'] . $tmp['path'];
    $whId = false;


    if (!empty($whList->webhooks) && is_array($whList->webhooks)) {
      foreach ($whList->webhooks as $el) {
        
        self::debug(print_r(self::map($el->event_types, 'name', 'name'), true));
        self::debug(print_r("type {$whDetails->event_type} " . strpos($el->url, $url), true), 'event URL ok');
        $_ret = [];
        if (is_array($el->event_types)) {
            foreach ($el->event_types as $et) {
                $_ret[] = $et->name;
            }
        }
        if (strpos($el->url, $url)!==false && is_array($el->event_types) && 
            ( $el->event_types[0]->name=='*' ||
              in_array($whDetails->event_type, $_ret))
            ) {
          $whId = $el->id;
          break;
        }
      }
    }

    $output = false;
    if ($whId) { //do not send verification if wbhook Id wasn't found.
      $headers = array_change_key_case($headers, CASE_UPPER);
      $signatureVerification = new \PayPal\Api\VerifyWebhookSignature();
      $signatureVerification->setAuthAlgo($headers['PAYPAL-AUTH-ALGO']);
      $signatureVerification->setTransmissionId($headers['PAYPAL-TRANSMISSION-ID']);
      $signatureVerification->setCertUrl($headers['PAYPAL-CERT-URL']);
      $signatureVerification->setWebhookId($whId); //0FV983392A0531201 Note that the Webhook ID must be a currently valid Webhook that you created with your client ID/secret.
      $signatureVerification->setTransmissionSig($headers['PAYPAL-TRANSMISSION-SIG']);
      $signatureVerification->setTransmissionTime($headers['PAYPAL-TRANSMISSION-TIME']);

      $signatureVerification->setRequestBody($requestBody);
      $request = clone $signatureVerification;

      self::debug(print_r($signatureVerification, true), '$signatureVerification');
      self::debug(print_r($this->getApiContext(), true), 'getApiContext');

      try {
        /** @var \PayPal\Api\VerifyWebhookSignatureResponse $output */
        $output = $signatureVerification->post($this->getApiContext());
      } catch (\Exception $ex) {
        self::warning($ex, 'paypal_rest');
        self::warning('request ' . print_r($request, true), 'paypal_rest');
      }
    } else {
        self::debug(print_r('not my webhook id $url' . $url, true));
        self::debug(print_r($whList, true));
    }
    self::debug(print_r($output, true), 'signatureVerification Response');

    return ($output && $output->getVerificationStatus() == "SUCCESS");
  }

  private function whBoardingComplete($whDetails) {
    /*
MERCHANT.ONBOARDING.COMPLETED
{
    "id":"WH-68Y618304D271921B-2TF22540HP7127708",
    "event_version":"1.0",
    "create_time":"2017-11-23T15:46:44.521Z",
    "resource_type":"merchant-onboarding",
    "event_type":"MERCHANT.ONBOARDING.COMPLETED",
    "summary":"The merchant account setup is completed",
    "resource":{
        "partner_client_id":"{partner-client-id}",
        "links":[
            … …
        ],
        "merchant_id":"{seller-payer-id}",
        "tracking_id":"{tracking-id}"
    },
    "links":[
        … …
    ]
}     */
    try {
      $check = PaypalPartner\models\SellerInfo::find()->where([
        'tracking_id' => $whDetails->resource->tracking_id,
        'is_onboard ' => 0
      ])->one();
      if ($check) {
        $check->is_onboard = 1;
        $check->save(false);
      }
    } catch (\Exception $e) {
      self::warning($e->getMessage(), 'paypal_rest');
    }
  }

  private function whBoardingRevoked($whDetails) {
    /*
{
   "id":"WH-02N37238V82187421-5PF6629796965892K",
   "event_version":"1.0",
   "create_time":"2017-09-07T08:56:32.812Z",
   "resource_type":"partner-consent",
   "event_type":"MERCHANT.PARTNER-CONSENT.REVOKED",
   "summary":"The Account setup consents has been revoked or the merchant account is closed",
   "resource":{
       "merchant_id":"{seller-payer-id}",
       "tracking_id":"{tracking-id}"
   },
   "links":[
      … …
   ]
}    */
    if (!empty($whDetails->resource->merchant_id) || !empty($whDetails->resource->tracking_id)) {
      try {

        $check = PaypalPartner\models\SellerInfo::find()->filterWhere([
          'payer_id' => $whDetails->resource->merchant_id,
          'tracking_id' => $whDetails->resource->tracking_id, // same merchant could be on several platforms (probably doesn't matter)
          'is_onboard' => 1
        ])->one();

        if ($check) {
          $check->is_onboard = 0;
          $check->save(false);
        }
      } catch (\Exception $e) {
        self::warning($e->getMessage(), 'paypal_rest');
      }
    }
  }


  /**
   * parse getTransactionDetails into $this->transactionInfo
   * @param array $transactionDetails
   */
  public function parseTransactionDetails($transactionDetails) {
/**
 PayPalHttp\HttpResponse Object
(
    [statusCode] => 200
    [result] => stdClass Object
        (
            [id] => 5NY87752NN160122A
            [amount] => stdClass Object
                (
                    [currency_code] => USD
                    [value] => 63.33
                )

            [final_capture] => 1
            [seller_protection] => stdClass Object
                (
                    [status] => ELIGIBLE
                    [dispute_categories] => Array
                        (
                            [0] => ITEM_NOT_RECEIVED
                            [1] => UNAUTHORIZED_TRANSACTION
                        )

                )

            [seller_receivable_breakdown] => stdClass Object
                (
                    [gross_amount] => stdClass Object
                        (
                            [currency_code] => USD
                            [value] => 63.33
                        )

                    [paypal_fee] => stdClass Object
                        (
                            [currency_code] => USD
                            [value] => 2.45
                        )

                    [net_amount] => stdClass Object
                        (
                            [currency_code] => USD
                            [value] => 60.88
                        )

                    [receivable_amount] => stdClass Object
                        (
                            [currency_code] => GBP
                            [value] => 46.09
                        )

                    [exchange_rate] => stdClass Object
                        (
                            [source_currency] => USD
                            [target_currency] => GBP
                            [value] => 0.757046355
                        )

                )

            [status] => COMPLETED
            [create_time] => 2020-09-01T09:55:07Z
            [update_time] => 2020-09-01T10:00:02Z
            [links] => Array
                (
                    [0] => stdClass Object
                        (
                            [href] => https://api.sandbox.paypal.com/v2/payments/captures/5NY87752NN160122A
                            [rel] => self
                            [method] => GET
                        )

                    [1] => stdClass Object
                        (
                            [href] => https://api.sandbox.paypal.com/v2/payments/captures/5NY87752NN160122A/refund
                            [rel] => refund
                            [method] => POST
                        )

                    [2] => stdClass Object
                        (
                            [href] => https://api.sandbox.paypal.com/v2/checkout/orders/6BV62129B0245622C
                            [rel] => up
                            [method] => GET
                        )

                )

        )
 */
    $this->transactionInfo = [];
    if (!empty($transactionDetails->resource)) {
      $res = $transactionDetails->resource;
    } elseif (!empty($transactionDetails->result)) {
      $res = $transactionDetails->result;
    }
    if ($res && !empty($res->id)) {

      if (!empty($res->status_details->reason)) {
        $comment =  $res->status_details->reason;
      } else {
        $comment =  $res->status . ' ' . $this->transactionType($transactionDetails) . ' ' . $res->amount->value . $res->amount->currency_code;
      }
      $this->transactionInfo['status'] = $res->status;
      $this->transactionInfo['status_code'] = $this->getStatusCode($transactionDetails);
      $this->transactionInfo['transaction_id'] = $res->id;
      $this->transactionInfo['amount'] = $res->amount->value;
      $this->transactionInfo['fulljson'] = json_encode($transactionDetails);
      //$this->transactionInfo['last_updated'] = str_replace(['T', 'Z'], [' ', ''], $res->update_time);

      $this->transactionInfo['comments'] = $comment;
    }

    return $this->transactionInfo;
  }

    public function parseBoardingDetails($response, $date = '') {
        if (empty($date)) {
            $date = date('Y-m-d');
        }
        $info = $errors = $warnings = [];
        if (!empty($response)) {
            $oAuth = $response->getOauthIntegrations();
            $scopes = $oAuth[0]->oauth_third_party[0]->scopes??'';

            if (!$response->getPaymentsReceivable() ) {
                $errors[] = MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_ERROR_RECEIVABLE;
            }
            if (!$response->getPrimaryEmailConfirmed()) {
                $errors[] = MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_ERROR_EMAIL;
            }
            
            $prods = $response->getProducts();
            $_capabilities = $response->getСapabilities();


            if (!empty($prods) && is_array($prods)) {
                foreach ( $prods as $prod  ) {
                    $cCap = $this->findCapabilities($_capabilities, 'CUSTOM_CARD_PROCESSING');
                    $cWithdraw = $this->findCapabilities($_capabilities, 'WITHDRAW_MONEY');
                    $cSend = $this->findCapabilities($_capabilities, 'SEND_MONEY');
                    /*
If "products[name==PPCP_CUSTOM].vetting_status:" is SUBSCRIBED
and "capabilities[name==CUSTOM_CARD_PROCESSING].status" is ACTIVE
and "capabilities[name==CUSTOM_CARD_PROCESSING].limits[0].type" is GENERAL
and "capabilities[name==WITHDRAW_MONEY].limits" is anything but undefinied
and "capabilities[name==SEND_MONEY].limits" is anything but undefinited
                     */
                    if ($prod->name == 'PPCP_CUSTOM' && $prod->vetting_status == 'SUBSCRIBED' &&
                         !empty($cCap['status']) && $cCap['status'] == 'ACTIVE' &&
                         !empty($cCap['limits'][0]['type']) && $cCap['limits'][0]['type'] == 'GENERAL' &&
                         isset($cWithdraw['limits']) &&
                         isset($cSend['limits'])
                        ){
                        $warnings[] = MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_500_WARNING_ACTIVE;
                    }

/*
"products[name].vetting_status:" is SUBSCRIBED and
"capabilities[name==CUSTOM_CARD_PROCESSING].status" is ACTIVE and
"capabilities[name==CUSTOM_CARD_PROCESSING].limits[0].type" is GENERAL and
"capabilities[name==WITHDRAW_MONEY].limits" is undefinied and
"capabilities[name==SEND_MONEY].limits" is undefinited
*/
                    elseif ($prod->name == 'PPCP_CUSTOM' && $prod->vetting_status == 'SUBSCRIBED' &&
                         !empty($cCap['status']) && $cCap['status'] == 'ACTIVE' &&
                         !empty($cCap['limits'][0]['type']) && $cCap['limits'][0]['type'] == 'GENERAL' &&
                         !isset($cWithdraw['limits']) &&
                         !isset($cSend['limits'])
                        ){
                        $warnings[] = MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_500_WARNING;
                    }

                    elseif ($prod->name == 'PPCP_CUSTOM' && $prod->vetting_status == 'NEED_MORE_DATA' ){
                        $warnings[] = MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_CCF_VETTING;
                    }
                    elseif ($prod->name == 'PPCP_CUSTOM' && $prod->vetting_status == 'IN_REVIEW' ){
                        $warnings[] = MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_CCF_VETTING_REVIEW;
                    }
                    elseif ($prod->name == 'PPCP_CUSTOM' && $prod->vetting_status == 'DENIED' ){
                        $warnings[] = sprintf(MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_CCF_VETTING_DENIED, tep_date_short($date));
                    }
                    
                }
            }

            $_tmp = false;
            if (!empty($scopes)) {
                $_tmp = $this->checkScopes($scopes, true);
            }
            if (is_array($_tmp) && !empty($_tmp['recommended'])) {
                $warnings = array_merge($warnings, $_tmp['recommended']);
            }

            /// info
            $info = [];
            foreach ([
                'MODULE_PAYMENT_PAYPAL_REST_LEGAL_NAME' => ['Legal name','getLegalName'],
                'MODULE_PAYMENT_PAYPAL_REST_SELLER_MERCHANT_ID' => ['Merchant id','getMerhantId'],
                'MODULE_PAYMENT_PAYPAL_REST_PAYMENTS_RECEIVABLE' => ['Payments receivable','getPaymentsReceivable'],
                'MODULE_PAYMENT_PAYPAL_REST_PRIMARY_EMAIL' => ['Primary e-mail','getPrimaryEmail'],
                'MODULE_PAYMENT_PAYPAL_REST_PRIMARY_CURRENCY' => ['Primary currency','getPrimaryCurrency'],
                //'MODULE_PAYMENT_PAYPAL_REST_' => ['',''],
              ] as $k => $inf
            ) {
                if (!empty($inf)) {
                    $val = $response->{$inf[1]}();
                    if ($val!='') {
                        $info[defined($k)?constant($k):$inf[0]] = $val;
                    }
                }
            }

            $products = [];
            if (false && !empty($prods) && is_array($prods)) {
                foreach ( $prods as $prod  ) {
                    $str = "<b>{$prod->name}</b><br />\n"
                        . "Vetting Status: <b>{$prod->vetting_status}</b><br />\n";
                     if (!empty($prod->capabilities) && is_array($prod->capabilities)) {
                         $str .= implode(" ", $prod->capabilities); // <BR />\n
                     }
                    $products[] = $str ;
                }
                $info['Products'] = '<small>' . implode("<BR /><BR />\n", $products) . "<BR /></small>\n";
            }

            $capabilities = [];
            if (!empty($_capabilities) && is_array($_capabilities)) {
                foreach ( $_capabilities as $_capability ) {
                    if (!empty($_capability['status'])) {
                        $status = '<span class="paypal-capabilities glyphicon ' . strtolower($_capability['status']) . '">' . $_capability['status'] . '</span>';
                    } else {
                        $status = $_capability['status']??'';
                    }

                    if (!empty($_capability['name'])) {
                        $str =
                          (defined('MODULE_PAYMENT_PAYPAL_REST_' . strtoupper($_capability['name']) . '_TEXT')?
                          constant('MODULE_PAYMENT_PAYPAL_REST_' . strtoupper($_capability['name']) . '_TEXT'):
                          $_capability['name']) . ": " . $status . (!empty($_capability['limits']) && is_scalar($_capability['limits']) ? $_capability['limits'] : '') . "";
                        $capabilities[] = $str;
                    }
                }
                $info[
                  defined('MODULE_PAYMENT_PAYPAL_REST_CAPABILITIES')?MODULE_PAYMENT_PAYPAL_REST_CAPABILITIES:'Capabilities'
                  ] = '<div>' . implode("</div><div>\n", $capabilities) . "</div>\n";
            }
            /// info end


        } else {
            $errors[] = MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_ERROR;
        }

        return [
                'info' => $info,
                'errors' => $errors,
                'warnings' => $warnings,
                ];

    }

    public function findCapabilities($_capabilities, $name) {
        if (!empty($_capabilities) && is_array($_capabilities)) {
                foreach ( $_capabilities as $_capability ) {
                    if ($_capability['name'] == $name) {
                        return $_capability;
                    }
                }
        }
        return [];
    }

    /**
     * oscok
     * @param type $force
     * @return type
     */
    public static function getPartnerId($force = false) {
        if (isset($this) && $this instanceof self) {
            $mode = $this->getMode();
        } else {
            $mode = self::getMode();
        }
        if (!empty($force)) {
            $mode = $force;
        }
        if ($mode == 'Live') {
            $ret = self::PARTNER_MERCHANT_ID;
        } else {
            $ret = self::PARTNER_MERCHANT_SANDBOX_ID;
        }
        return $ret;
    }


    public static function getAttributionId() {
        $ret = self::PARTNER_ATTRIBUTION_ID;
        if (defined('MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER') && MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER == 'Sandbox') {
            $ret = self::SANDBOX_PARTNER_ATTRIBUTION_ID;
        }
        if (empty($ret)) {
            $ret = false;
        }
        return $ret;
    }

    protected function _getClientId() {
        if ($this->getMode() == 'Live') {
            $ret = self::PARTNER_APP_CLIENT_ID;
        } else {
            $ret = self::PARTNER_APP_SANDBOX_CLIENT_ID;
        }
        if ($this->hasOwnKeys()) {
            $platformId = $this->getPlatformId();
            $seller = $this->getSeller($platformId);
            $ret = $seller->own_client_id;
        }

        return $ret;
    }

    protected function _getClientSecret() {
        if ($this->getMode() == 'Live') {
            $ret = self::PARTNER_APP_CLIENT_SECRET;
        } else {
            $ret = self::PARTNER_APP_SANDBOX_CLIENT_SECRET;
        }

        $platformId = $this->getPlatformId();
        $seller = $this->getSeller($platformId);
        if (!empty($seller->own_client_id) && !empty($seller->own_client_secret)) {
            $ret = $seller->own_client_secret;
        }

        return $ret;
    }

    protected function _validateKey($key, $platform_id) {
        $ret = false;
        return $ret;
    }

    protected function _getIntent() {
        if (defined('MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_METHOD')) {
            $ret = MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_METHOD;
        } else {
            $ret = 'sale';
        }
        return strtolower($ret);
    }

    public static function get3DSSettings($seller) {
        $ret = \paypal_rest::$threeDSDefaults;
        if (!empty($seller->three_ds_settings)) {
            $ret = json_decode($seller->three_ds_settings, true);
        }
        $ret['defaults'] = \paypal_rest::$threeDSDefaults;
        $ret['3dsa'] = [];
        if (is_array(\paypal_rest::$threeDSDefaults)) {
            foreach(\paypal_rest::$threeDSDefaults as $k => $dfVal) {
                if (substr($k, 0, 5) == '3dsa_') {
                    $tmp = explode('_', substr($k, 5));
                    $row = [];
                    if (!empty($tmp[0])) {
                        $row['e_desc'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_ENROLLMENT_' . strtoupper($tmp[0]). '_DESCRIPTION')?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_ENROLLMENT_' . strtoupper($tmp[0]). '_DESCRIPTION') : strtoupper($tmp[0]);
                        $row['e'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_ENROLLMENT_' . strtoupper($tmp[0]))?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_ENROLLMENT_' . strtoupper($tmp[0])) : strtoupper($tmp[0]);
                    } else {
                        $row['e_desc'] = $row['e'] ='';
                    }
                    if (!empty($tmp[1])) {
                        $row['a_desc'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_AUTHENTICATION_' . strtoupper($tmp[1]). '_DESCRIPTION')?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_AUTHENTICATION_' . strtoupper($tmp[1]). '_DESCRIPTION') : strtoupper($tmp[1]);
                        $row['a'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_AUTHENTICATION_' . strtoupper($tmp[1]))?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_AUTHENTICATION_' . strtoupper($tmp[1])) : strtoupper($tmp[1]);
                    } else {
                        $row['a_desc'] = $row['a'] ='';
                    }
                    if (!empty($tmp[2])) {
                        switch ($tmp[2]) {
                            case 1:
                                $row['l_desc'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_POSSIBLE_DESCRIPTION')?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_POSSIBLE_DESCRIPTION') : 'POSSIBLE';
                                $row['l'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_POSSIBLE')?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_POSSIBLE') : 'POSSIBLE';
                                break;
                            case 2:
                                $row['l_desc'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_NO_DESCRIPTION')?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_NO_DESCRIPTION') : 'NO';
                                $row['l'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_NO')?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_NO') : 'NO';
                                break;
                            case 3:
                                $row['l_desc'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_UNKNOWN_DESCRIPTION')?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_UNKNOWN_DESCRIPTION') : 'UNKNOWN';
                                $row['l'] = defined('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_UNKNOWN')?constant('TEXT_MODULE_PAYMENT_PAYPAL_REST_3DSA_LIABILITY_UNKNOWN') : 'UNKNOWN';
                                break;
                        }
                    } else {
                        $row['l_desc'] = $row['l'] ='';
                    }

                    $row['def_state_class'] = (!empty($ret['defaults'][$k])?'def-checked':'def-unchecked');
                    $row['checked'] = (!empty($ret[$k]) || (!isset($ret[$k]) && $ret['defaults'][$k]));
                    $row['key'] = $k;
                    
                    $ret['3dsa'][] = $row;
                }
            }
        }

        return $ret;
    }

    public function getFundings() {
        $ret = [];
        if (defined('MODULE_PAYMENT_PAYPAL_REST_BUTTON_FUNDING') && !empty(MODULE_PAYMENT_PAYPAL_REST_BUTTON_FUNDING)) {
            $fundings = array_map('trim', explode(',', MODULE_PAYMENT_PAYPAL_REST_BUTTON_FUNDING));
            if (!in_array('', $fundings) && !in_array('all', $fundings) && !(in_array('--none--', $fundings) && count($fundings)==1)) {
                $enabled = array_values(array_diff($fundings, ['', 'all', '--none--']));
                $disabled = array_diff(array_keys(self::$possibleFundings), $enabled);

                if (!empty($enabled)) {
                    $ret['enabled'] = implode(',', array_map('trim', $enabled));
                }
                if (!empty($disabled)) {
                    $ret['disabled'] = implode(',', $disabled);
                }
            }
            //$ret['enabled']
        }
        return $ret;

    }

    //Custom Card processing status
    public function CCPActive($seller) {
        $ret = false;
        if (!empty($seller->boarding_json)) {
            $merchant = new PaypalPartner\api\Merchant();
            $response = $merchant->fromJson($seller->boarding_json);


            //Settings should be shown only if the following values are returned from the Show Seller Status API call:
            //"payments_receivable": true,
            //"primary_email_confirmed": true
            //"products[name==PPCP_CUSTOM].vetting_status:":SUBSCRIBED
            //"capabilities[name==CUSTOM_CARD_PROCESSING].status": ACTIVE

            if ($response && $response->getPaymentsReceivable() && $response->getPrimaryEmailConfirmed()) {
                $prods = $response->getProducts();
                $_capabilities = $response->getСapabilities();
                $cCap = $this->findCapabilities($_capabilities, 'CUSTOM_CARD_PROCESSING');
                if (!empty($prods) && is_array($prods) && !empty($cCap['status']) &&  $cCap['status'] == 'ACTIVE') {

                    foreach ( $prods as $prod  ) {
                        if ($prod->name == 'PPCP_CUSTOM') {
                            $ret = true;
                            break;
                        }
                    }
                }
            }
        }

        return $ret;
    }

    public function generateCCToken() {

    }

    public function generateCCCustomerDetails() {
        $ret = [];
        $post_data = $_POST;
        $postbill = $post_data['billing_ab_id']??false;
        if ($postbill) {
            $tmp = $this->getCustomersAddress($postbill, true, true);
        }
        if (!array($tmp) || empty($tmp['street_address'])) {
            if (!$this->isInstance()) {
                $order = $this->createOrderInstance();
            } else {
                $order = $this->getOrderInstance();
            }
            $tmp = $order->billing??null;
        }
        if (!array($tmp) || empty($tmp['street_address'])) {
            $tmp = $this->getBillingAddress();
        }
        if (!array($tmp) || empty($tmp['street_address']) && !empty($post_data['Billing_address'])) {
            foreach (['firstname', 'lastname', 'street_address', 'city', 'postcode'] as $k) {
                if (empty($tmp[$k]) && !empty($post_data['Billing_address'][$k])) {
                    $tmp[$k] = strip_tags($post_data['Billing_address'][$k]);
                }
            }
            if ($post_data['Billing_address']['country'] != $tmp['country']['id']) {
                $_country = \common\helpers\Country::get_country_info_by_id($post_data['Billing_address']['country']);
                if (!empty($_country['iso_code_2'])) {
                    $tmp['country'] = $_country;
                }
            }
        }
        
        if ($tmp['zone_id'] > 0) {
            if ($tmp['country']['iso_code_2'] == 'US') {
                $_state = tep_get_zone_code($tmp['country']['id'], $tmp['zone_id'], '');
            } else {
                $_state = tep_get_zone_name($tmp['country']['id'], $tmp['zone_id'], $tmp['state']);
            }
            if (!empty($_state)) {
                $tmp['state'] = $_state;
            }
        }

        if (!empty($tmp['firstname']) && !empty($tmp['lastname'])) {
            $ret['cardholderName'] = substr($tmp['firstname'] . ' ' . $tmp['lastname'], 0, 140);
        }
        if (!empty($tmp['street_address']) && !empty($tmp['postcode']) ) {
            $ret['billingAddress'] = [
              'streetAddress' => substr($tmp['street_address'], 0, 300),
              'extendedAddress' => substr($tmp['suburb'], 0, 300),
              'locality' => substr($tmp['city'], 0, 120),
              'region' => substr($tmp['state'], 0, 300),
              'postalCode' => substr($tmp['postcode'], 0, 60),
              'countryCodeAlpha2' => substr($tmp['country']['countries_iso_code_2']??$tmp['country']['iso_code_2'], 0, 60),//
            ];
        }
        $seller = $this->getSeller();
        $contingencies = 'SCA_WHEN_REQUIRED';
        if (!empty($seller->three_ds_settings)) {
            $tmp = json_decode($seller->three_ds_settings, true);
            if (!empty($tmp['contingencies'])) {
                $contingencies = $tmp['contingencies'];
            }
        }
        $ret['contingencies'] = [$contingencies];
        $this->scalarToAddresses();
        return $ret;
    }

/**
 *
 * @param array $data ["transaction_id" => "8MC585209K746392H"{, "tracking_number" => "443844607820", "carrier" => "FEDEX", "status" => "SHIPPED", 'carrier_name_other'}]
 * @return obj|array|false
 */
    public function addTracking($data) {
        $request = new \PayPalCheckoutSdk\Shipping\TrackersBatchRequest();
        if (false && $tmp = self::getAttributionId()) {
            $request->payPalPartnerAttributionId($tmp);
        }
        $trackers = [];
        if (!empty($data) && is_array($data)) {
            foreach ($data as $d) {
                $tmp = ["status" => "SHIPPED", "carrier" => "OTHER"];
                foreach (['transaction_id', 'tracking_number', 'carrier', 'carrier_name_other', 'status'] as $key) {
                    if (!empty($d[$key])) {
                        $tmp[$key] = $d[$key];
                    }
                }
                if (!empty($tmp['transaction_id'])) {
                    $trackers[] = $tmp;
                }
            }
        }
        if (empty($trackers)) {
            return ['error' => 1, 'message' => TEXT_ERROR_TRANSACTIONID_REQUIRED];
        }

        $request->body = ['trackers' => $trackers];
        if ($this->debug) {
            self::warning(" #### " .print_r($request->body, true), 'TLDEBUG' . $this->code);
        }

        try {
            return $this->getHttpClient()->execute($request);
        } catch (\Exception $ex) {

            self::error('body ' . print_r($request->body, true), 'paypal_rest_exception');
            self::error('message ' . $ex->getMessage(), 'paypal_rest_exception');
            if ($tmp = $this->parseJsonMessage($ex->getMessage())) {
                return ['error' => 1, 'message' => $tmp];
            }
        }
        return false;
    }

/**
 *
 * @param string $traker_id
 * @param array $data ["transaction_id" => "8MC585209K746392H"{, "tracking_number" => "443844607820", "carrier" => "FEDEX", "status" => "CANCELLED", 'carrier_name_other'}]
 * @return obj|array|false
 */
    public function cancelTracking($tracker_id, $data) {
        $request = new \PayPalCheckoutSdk\Shipping\TrackersPatchRequest($tracker_id);
        if (false && $tmp = self::getAttributionId()) {
            $request->payPalPartnerAttributionId($tmp);
        }
        $trackers = [];
        if (!empty($data) && is_array($data)) {
            $tmp = ["status" => "CANCELLED"];
            foreach (['transaction_id', 'tracking_number', 'carrier', 'carrier_name_other', 'status'] as $key) {
                if (!empty($data[$key])) {
                    $tmp[$key] = $data[$key];
                }
            }
            if (!empty($tmp['transaction_id'])) {
                $trackers = $tmp;
            }
        }
        if (empty($trackers)) {
            return ['error' => 1, 'message' => TEXT_ERROR_TRANSACTIONID_REQUIRED];
        }

        $request->body = $trackers;
        if ($this->debug) {
            self::warning(" #### " .print_r($request->body, true), 'TLDEBUG' . $this->code);
        }

        try {
            return $this->getHttpClient()->execute($request);
        } catch (\Exception $ex) {

            self::error('body ' . print_r($request->body, true), 'paypal_rest_exception');
            self::error('message ' . $ex->getMessage(), 'paypal_rest_exception');
            if ($tmp = $this->parseJsonMessage($ex->getMessage())) {
                return ['error' => 1, 'message' => $tmp];
            }
        }
        return false;
    }

    public function createAbsoluteUrl($params = [], $connection = 'NONSSL', $catalog = false){
        $page = $params[0];
        if (empty($page)) {
            $page = '/';
        }
        if (!in_array($connection, array('NONSSL', 'SSL'))) {
            $connection = 'SSL';
        }

        unset($params[0]);
        if ($catalog && function_exists('tep_catalog_href_link')) {
            $ret = tep_catalog_href_link($page, http_build_query($params), $connection);
        } else {
            $ret = tep_href_link($page, http_build_query($params), $connection);
        }

        return str_replace('&amp;', '&', $ret);
    }

    public function fetchOwnKeys() {
        $get = $_GET;
        $post = $_POST;
        if (empty($post)) {
            $payload = @file_get_contents('php://input');
            if (!empty($payload )) {
               // parse_str($payload, $post);
                $post = json_decode($payload, true);
            }
        }
        $platformId = $post['platform_id']??$this->getPlatformId();
        try {
            if (!empty($post['ppp_mode']) && in_array($post['ppp_mode'], ['Live', 'Sandbox'])) {
                $mode = $post['ppp_mode'];
                $seller = $this->getSeller($mode, true);
                $partner_id = $this->getPartnerId($mode);
            }
            if (!$seller) {
                $seller = $this->getSeller();
                $mode = $this->getMode();
                $partner_id =$this->getPartnerId();
            }

            if ($mode != MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER) {
                tep_db_query("update " . TABLE_CONFIGURATION . " set configuration_value='" . $mode . "' where configuration_key='". MODULE_PAYMENT_PAYPAL_REST_TRANSACTION_SERVER . "'");
            }
            
            $details = Merchant::getCustomerToken($post['authCode'], $post['sharedId'], $seller->tracking_id, $partner_id, $mode);
            if (!empty($details['client_id'])) {
                $seller->own_client_id = $details['client_id'];
                $seller->own_client_secret = $details['client_secret'];
                $seller->payer_id = $details['payer_id'];
                $seller->is_onboard = 1;
                $seller->save(false);
            }
        } catch (\Exception $ex) {
            $this->sendDebugEmail($ex);
            $this->errors[] = $ex->getMessage();
        }
    }

    public static function getPlatformId( ){
        return 0;
    }

    public static function multiOptionCheckbox($select_array, $key_value, $key = '') {
        static $counter = 1;
        $string = '';
        $key_values = array_map('trim', explode(',', $key_value));

        $name = ((tep_not_null($key)) ? 'configuration[' . $key . ']' : 'configuration_value');
        $nm = ((tep_not_null($key)) ? 'configuration' . $key . '' : 'configuration_value' . $counter);
        foreach ($select_array as $k => $v) {


          $string .= '<br /><label><input type="checkbox" class="a' . $nm . '" name="a' . $nm . '[]" value="' . $k . '"';

          if (in_array($k, $key_values)) $string .= ' checked="checked"';

          $string .= ' /> ' . $v . '</label>';

        }
        $string .= '<input type="hidden" class="' . $nm . '" name="' . $name . '" value="' . $key_value . '">';

        $js = <<<EOD
        <script> 
            $(".a{$nm}").on("click", function(){
                if ($(this).val() == '' || $(this).val() == 'all') {
                    $(".a{$nm}").prop('checked', $(this).prop('checked'));
                }
                var val = [];
                $(".a{$nm}:checked").each(function(i, e){
                    if ($(e).val() != '') {
                        val.push($(e).val());
                    }
                });
                $(".{$nm}").val(val.join());
            });
        </script>
EOD;
        $string .= $js;
        
        $counter++;

        return $string;
    }

    public static function get_country_info_by_id($id = null) {
        if (empty($id)) {
            $id = (int)STORE_COUNTRY;
        }
        $countries_query = tep_db_query("select countries_id, countries_name, countries_iso_code_2 from " . TABLE_COUNTRIES . " where countries_id = '". (int)$id . "'");
        return tep_db_fetch_array($countries_query);
    }
    
    public static function warning($message, $code='') {
        if (self::$debug) {
            $_ee = new \Exception(); 
            error_log("[" . date('d-m-Y H:i:s') . "][" . $code . "]: " . "$message\n" . $_SERVER['QUERY_STRING'] . "\n" . $_ee->getTraceAsString() . "\n", 3, self::$warningsFile);
        }
    }

    public static function debug($message, $code='') {
        if (self::$debug) {
            $_ee = new \Exception();
            error_log("[" . date('d-m-Y H:i:s') . "][" . $code . "]: " . "$message\n" . $_SERVER['QUERY_STRING'] . "\n" . $_ee->getTraceAsString() . "\n", 3, self::$warningsFile);
        }
    }
    public static function error($message, $code='') {
        $_ee = new \Exception(); 
        error_log("[" . date('d-m-Y H:i:s') . "][" . $code . "]: " . "$message\n GET" . print_r($_GET, true) . "\nPOST"  . print_r($_POST, true) . "\n" . $_ee->getTraceAsString() . "\n", 3, self::$errorFile);
    }
/**
 *  check possible options to get PP account and keys
 * @param bool $only code
 * @return int|array $code: 1 - sandbox 2 - live 3 - both partner's keys 4 - only own keys
 */
    public function getInstallOptions($only = false) {
        $ret = [];
        $code = 0;//rudiment
        if (!empty(self::PARTNER_MERCHANT_ID)) { //!empty(self::PARTNER_APP_CLIENT_ID) &&
           $code += 2;
        }
        if ( !empty(self::PARTNER_MERCHANT_SANDBOX_ID) ) { //!empty(self::PARTNER_APP_SANDBOX_CLIENT_ID) &&
           $code += 1;
        }
        if (!$code) {
            $code = 4;
        }
        if ($only) {
            return $code;
        }


        $ret['activePPP'] = 0;
        $installed = false;
        if (defined('MODULE_PAYMENT_PAYPAL_REST_STATUS')) {
            $ret['activePPP'] = (int)(MODULE_PAYMENT_PAYPAL_REST_STATUS == 'True');
            $installed = $this->getInstaller()->check_install();
        }

        $for_seller = true; // always new sellers;
        $wh_urlParams = ['ext/modules/payment/paypal_rest/index.php', 'action' => 'callback'];
        //$wh_urlParams = tep_href_link('ext/modules/payment/paypal_rest.php', '=');
        $boardingOptionsUrl = tep_href_link('ext/modules/payment/paypal_rest/index.php', 'action=links');
        $ret['titlePPP'] = defined('ADD_PAYPAL')?ADD_PAYPAL:'PayPal Quick Setup';
        $ret['installPPP'] = $code;// in any mode
        $ret['golivePPP'] = false; // sandbox exists
        $ret['boardingOptionsUrl'] = $boardingOptionsUrl;

        if ($installed) {
            $urlParams = [
              'set' => 'payment', 'module' => $this->code,
            ];

            if (($_GET['action']??false) == 'edit' && ($_GET['psi_id']?? false) > 0) {
                $filename = 'modules.php';
                unset($urlParams[0]);
                $ret['cancelUrl'] =  tep_href_link($filename, http_build_query($urlParams));
            }


            if ($this->liveConfigurationExists()) {
                $ret['installPPP'] = 0;

            } else {

                if ($this->ownSandboxConfigExists()
                    && !(($_GET['action']??false) == 'edit' && $this->getMode() == 'Sandbox' && ($_GET['psi_id']??false) > 0)
                    ) { //
                    $ret['titlePPP'] = defined('MODULE_PAYMENT_PAYPAL_REST_GO_LIVE')?MODULE_PAYMENT_PAYPAL_REST_GO_LIVE:'PayPal Go Live';
                    $ret['golivePPP'] = true;// allow own live account only
                } else {
                    $ret['titlePPP'] = defined('ADD_PAYPAL')?ADD_PAYPAL:'PayPal Quick Setup';
                    $ret['activePPP'] = false; //not configured at all
                }

                if ($this->boardViaLink()) {

                    //$logoUrl = tep_catalog_href_link(FILENAME_DEFAULT, '', 'SSL', true);
                    $logoUrl = tep_catalog_href_link('/', '', 'SSL', true);
                    if (strpos($logoUrl, '127.0.0.1') || strpos($logoUrl, 'localhost')) {
                        $logoUrl = 'https://paypal.oscommerce.host';
                    }
                    $logoUrl = rtrim($logoUrl, '/') . '/admin/images/oscommerce.png';
                    $seller = $this->getSeller('Live', $for_seller);
                    $seller->save();
                    //$seller->refresh();
                    $sSeller = $this->getSeller('Sandbox', $for_seller);
                    $sSeller->save();
                    //$sSeller->refresh();

                    $tmp = self::get_country_info_by_id();

                    $_params = $wh_urlParams;
                    $_params['action'] = 'sellerDetails';
                    $fetchKeysUrl = self::createAbsoluteUrl($_params, 'SSL', true);
                    if ($seller) {
                        $tmp = $urlParams;
                        $tmp[0] = 'ext/modules/payment/paypal_rest/index.php';
                        $tmp['action'] = 'save_merchant';
                        $returnUrl = self::createAbsoluteUrl($tmp+['psi_id' => $seller->psi_id]);
                        $ret['link_params'] = [
                                'partnerClientId' => self::PARTNER_APP_CLIENT_ID,
                                'partnerId' => self::PARTNER_MERCHANT_ID,
                                'sellerNonce' => $seller->tracking_id,
                                'country' => $tmp['countries_iso_code_2']??'',
                                'locale' => str_replace('_', '-', locale_get_default()),
                                'return_url' => rawurlencode($returnUrl),
                                'partnerLogoUrl' => rawurlencode($logoUrl)
                              ];
                        $ret['psi_id'] = $seller->psi_id;
                    }
                    if ($sSeller) {
                        $tmp = $urlParams;
                        $tmp[0] = 'ext/modules/payment/paypal_rest/index.php';
                        $tmp['action'] = 'save_merchant';
                        $returnUrl = self::createAbsoluteUrl($tmp+['psi_id' => $sSeller->psi_id]);
                        $ret['link_params_sandbox'] = [
                                'partnerClientId' => self::PARTNER_APP_SANDBOX_CLIENT_ID,
                                'partnerId' => self::PARTNER_MERCHANT_SANDBOX_ID,
                                'sellerNonce' => $sSeller->tracking_id,
                                'country' => $tmp['countries_iso_code_2']??'',
                                'locale' => str_replace('_', '-', locale_get_default()),
                                'return_url' => rawurlencode($returnUrl),
                                'partnerLogoUrl' => rawurlencode($logoUrl)
                              ];
                        $ret['spsi_id'] = $sSeller->psi_id;
                    }
                    $ret['fetchKeysUrl'] = $fetchKeysUrl;
                }

            }
        }
        return $ret;
    }

    public function getBoardingDetails(PaypalPartner\models\SellerInfo $seller) {
        $ret = [];

        if (!empty($seller->boarding_json) && !$this->hasOwnKeys() ) {
            $merchant = new Merchant();
            $ret = $this->parseBoardingDetails($merchant->fromJson($seller->boarding_json), $seller->boarding_date);
        } elseif($this->_isReady()) {
            $merchant = $this->getMerchant();
            $get = $_GET;
            if (empty($seller->payer_id) && !empty($get['merchantIdInPayPal'])) {
                $seller->payer_id = $get['merchantIdInPayPal'];
            }
            if ($response = $merchant::checkStatus($this->getPartnerId(), $seller->payer_id, $this->getApiContext())) {
                $ret = $this->parseBoardingDetails($response);
                try {
                    $sEmail = $response->getPrimaryEmail();
                    if (!empty($sEmail) && $sEmail != $seller->email_address) {
                        $seller->email_address = $response->getPrimaryEmail();
                    }
                    $seller->boarding_json = $response->json;
                    $seller->boarding_date = date('Y-m-d');
                    $seller->save(false);
                } catch (\Exception $e) {
                    self::warning(print_r($e->getMessage(), true), 'TLDEBUG_' . $this->code);
                }
            } else {
                $ret['errors'][] = MODULE_PAYMENT_PAYPAL_REST_SELLER_BOARDED_ERROR;
            }
        }
        return $ret;
    }


    /**
     * get current webhooks for application
     */
    public function getNeededWebHooks() {
        return $this->webHooks;
    }

    public function getWebHookUrl() {
        $url = $this->createAbsoluteUrl(['ext/modules/payment/paypal_rest/index.php', 'action' => 'processWebhook'], 'SSL', true);
        return $url;
    }

    public function getSubscribedWebHooks() {
        //get webhooks (could be several, groupped by listener URL)
        $whList = $this->getWebHooks();
        $whListNames = []; // subscribed names only
        $url = $this->getWebHookUrl();

        if (!empty($whList->webhooks) && is_array($whList->webhooks)) {
            $whListNames = array_reduce(array_map(
                    function ($el) use ($url) {
                        $ret = [];
                        if ($url == $el->url && is_array($el->event_types)) {
                            foreach ($el->event_types as $et) {
                                $ret[] = $et->name;
                            }
                        }
                        return $ret;
                    }, $whList->webhooks), 'array_merge', array());
        }
        return $whListNames;
    }
    /**
     * subscribe to required webhooks for application
     */
    public function addWebHooks() {
        if (!empty($this->webHooks)) {
            $whListNames = $this->getSubscribedWebHooks();
            $neededWebHooks = $this->getNeededWebHooks();

            if (is_array($whListNames) && !empty($whListNames) && is_array($whListNames)) {
                $subEvents = array_diff($neededWebHooks, $whListNames);
            } else {
                $subEvents = $neededWebHooks;
            }

            if (is_array($subEvents) && !empty($subEvents)) {
                try {
                    $this->setWebHooks($subEvents);
                } catch (\Exception $ex) {
                    self::warning("Subscribe webhooks exception " . $ex->getMessage(), $this->code);
                }
            }
        }
    }

    // port - partly
    public static function map($array, $from, $to) {
        $ret = [];
        if (is_array($array)) {
            foreach ($array as $d) {
                if (is_array($d) && array_key_exists($from, $d)){
                    $ret[$d[$from]] = isset($d[$to])?$d[$to]:null;
                } elseif (is_object($d) && property_exists($d, $from)){
                    $ret[$d->$from] = isset($d->$to)?$d->$to:null;
                }
            }
        }
        return $ret;
    }

    public static function extraConfigDescription(){
        $ret = '';
        $set = 'payment';
        global $cfgModules, $PHP_SELF;
        $module_directory = $cfgModules->get($set, 'directory');

        $file_extension = substr($PHP_SELF, strrpos($PHP_SELF, '.'));
        $class = 'paypal_rest';
        if (!class_exists($class)) {
          include_once($module_directory . $class . $file_extension);
        }
        $module = new $class;
        $extra = new PaypalPartner\ExtraConfig([]);
        $extra->module = $module;
        $ret = $extra->extra_params();
        return $ret;
    }

    public static function js_escape($str) {
        return str_replace(["\n", "\r"], '', htmlspecialchars($str));
    }

    public function hasOwnKeys() {
        $seller = $this->getSeller();
        return (!empty($seller->own_client_id) && !empty($seller->own_client_secret));
    }

    public function validateCredentials($data) {
        $ret = ['html' => '...'];
        if (!empty($data['checkVal']) && $data['checkVal'] == md5(DB_SERVER_PASSWORD)) {
            $mode = $this->getMode();
            if (in_array(($data['mode']??''), ['Live', 'Sandbox'])) {
                $mode = $data['mode'];
            }
            $auth = new PayPal\Auth\OAuthTokenCredential($data['own_client_id'], $data['own_client_secret']);
            $apiContext = new PayPal\Rest\ApiContext($auth);
            $apiContext->setConfig(['mode' => $mode]);
            $merchant = $this->getMerchant();
            if ($response = $merchant::checkStatus($this->getPartnerId($mode), $data['payer_id'], $apiContext)) {
                $ret = $this->parseBoardingDetails($response);
                if (!empty($response) && $response->getMerhantId() == $data['payer_id']) {
                    $ret = ['html' => MODULE_PAYMENT_PAYPAL_REST_API_OK, 'status' => 1];
                } else {
                    $ret = ['html' => MODULE_PAYMENT_PAYPAL_REST_API_FAIL_DATA, 'status' => 0];
                }
            }
        }
        return $ret;
    }
//-------------------- Yii

    public function addJsFile($src, $options=[]){
        $ret = '<script type="text/javascript" src="' . $src . '" ';
        if (!empty($options) && is_array($options)) {
            foreach ($options as $k=>$v) {
                $ret .= $k . '="' . $v . '"';
            }
        }
        $ret .= ' ></script>';
        $this->html_buffer .= "\n" . $ret;
        return $ret;
    }
    public function addJs($script){
        $ret = '';
        $ret .= "\n<!--{$this->code} js -->\n";
        $ret .= '<script type="text/javascript">';
        $ret .= "\n" . $script . "\n";
        $ret .= '</script>';
        $ret .= "\n<!--{$this->code} js -->\n";
        $this->html_buffer .= "\n" . $ret;
        return $ret;
    }

    public function addCss($css){
        $ret = '';
        $ret .= "\n<!--{$this->code} css -->\n";
        $ret .= '<style>';
        $ret .= "\n" . $css . "\n";
        $ret .= '</style>';
        $this->html_buffer .= "\n" . $ret;
        return $ret;
    }

    public function getCartDetails() {
        $items = [];
        $order = $this->getOrderInstance();
        global $currency;
        global $currencies;

        $masterTotal = $this->formatRaw($order->info['total']);

        $totalPayPal = $totalPayPalTax = 0;
        $ii = 0;
        // ex VAT prices and tax to avoid items patch (export orders) - not working via JS :(.
        foreach ($order->products as $product) {

            if ($this->sendExVat) {
                $_val = $product['final_price'];
            } else {
                $_val = tep_add_tax($product['final_price'], $product['tax']);
            }
            $val = $this->formatRaw($_val);
            $qty = $product['qty'];
            $virtual_qty = '';

            $items[$ii] = [
              'name' => substr($virtual_qty . $product['name'], 0, 127),
              'unit_amount' => [
                'currency_code' => $currency,
                'value' => $val,
              ],
              'quantity' => $qty,
            ];

            if ($this->requireItemVat) {
                //PayPal /USA aproach could give huge descrepancy in TAX/Total values because of floating math.
                // on site for EU we need price inc Tax and the same price inc VAT should be for 1 and 10 pcs
                // it means net priceEach is changed for 1 and 10 products (EU). USA approach don't allow marketing (NN.99) prices.
                // ex VAT 20% priceEachGross 0.15 NetPriceEach 0.12 VAT 0.03 (USA approach Net 0.12 Tax 0.02 Gross 0.14)
                // 10 pcs price Gross 1.50 (Net 1.25 Tax 0.25)  USA approach Net 1.20 Tax 0.20 Gross 1.40
                $_tax = 0;
                if ($product['tax']>0 && $qty>0) {
                    if ($this->sendExVat) {
                        $_eachGross = $val + tep_calculate_tax($val, $product['tax']);
                    } else {
                        $_eachGross = $val;
                    }
                    $_gross = $this->formatRaw($_eachGross * $qty, $currency, 1);
                    $_nett = $this->formatRaw(self::get_untaxed_value($_gross, $product['tax']), $currency, 1);
                    $_tax = $this->formatRaw(($_gross - $_nett)/$qty, $currency, 1);
                    $totalPayPalTax += $this->formatRaw($_tax*$qty, $currency, 1);
                }
                $items[$ii]['tax'] = [
                    'currency_code' => $currency,
                    'value' => $_tax,
                ];
                $items[$ii]['tax_rate'] = round($product['tax'], 2);
                $items[$ii]['category'] = 'PHYSICAL_GOODS';//DIGITAL_GOODS
            }
            $ii++;
        }
        $shipping = [];
        if ($this->isShippingNeeded()) {
            $shipping_inc = $shipping_exc = 0;
            if (!empty($this->shipping_cost['cost'])) {
                $shipping_exc = $this->shipping_cost['cost'];
            }
            if (!empty($this->shipping_cost['cost_inc_tax'])) {
                $shipping_inc = $this->shipping_cost['cost_inc_tax'];
            }
            if (!$this->requireItemVat) { //shipping as total - else as product
                $shipping = [
                  'currency_code' => $currency,
                  'value' =>
                    $this->formatRaw(
                        $this->display_price_clear($this->sendExVat ? $shipping_exc : $shipping_inc, 0, 1)
                                                  , $currency, 1
                    )
                ];
                $totalPayPal += $shipping['value'];
            } else {// no shipping tax and Should equal sum of (tax * quantity) across all items for a given purchase_unit (/purchase_units/@reference_id=='default'/amount/breakdown/tax_total/value )
                $_tax = $this->formatRaw($this->formatRaw(
                        $this->display_price_clear($shipping_inc, 0, 1), $currency, 1
                    ) - $this->formatRaw(
                        $this->display_price_clear($shipping_exc, 0, 1), $currency, 1
                    ), $currency, 1);
                $items[$ii] = [
                    'name' => TEXT_SHIPPING,
                    'unit_amount' => [
                        'currency_code' => $currency,
                        'value' => $this->formatRaw(
                            $this->display_price_clear($shipping_exc, 0, 1), $currency, 1
                        )
                    ],
                    'quantity' => 1,
                    'tax' => [
                        'currency_code' => $currency,
                        'value' => $_tax,
                    ]
                ];
                $items[$ii]['tax_rate'] = round($_tax/$this->formatRaw(
                        $this->display_price_clear($shipping_exc, 0, 1), $currency, 1
                    ), 2);
                $items[$ii]['category'] = 'PHYSICAL_GOODS';
                $ii++;
                $totalPayPalTax += $_tax;
            }
        }

        $details = [
          'totals' => [
            'item_total' => [
              'currency_code' => $currency,
              'value' => 0
            ],
          ]
        ];
        //if ($this->sendExVat) {
            $details['items'] = $items;
        //}

        if (!empty($shipping)) {
            $details['totals']['shipping'] = $shipping;
        }

        $handlingValue = 0;
        $discountValue = 0;
        $tax = 0;
        $totalCollection = $this->getTotalCollection();
        foreach ($this->getTotalOutput(false) as $total) {
            if ($total['code'] == 'ot_tax') {
                $tax =  $total['value'];
/*2do if required
            } elseif (!in_array($total['code'], ['ot_shipping', 'ot_subtotal', 'ot_total'])) {
                if ($totalCollection->get($total['code'])->credit_class) {
                    $_tmpTitle = defined('MODULE_ORDER_TOTAL_COUPON_TOTAL')?constant('MODULE_ORDER_TOTAL_COUPON_TOTAL'):'';
                    if ($total['code']=='ot_coupon' && $total['title'] == $_tmpTitle . ':') {
                        continue; //multicoupon with total discount line dirty hack;
                    }
                    $discountValue += ($this->sendExVat ? $total['value_exc_vat'] : $total['value_inc_tax']);
                } else {
                    $handlingValue += ($this->sendExVat ? $total['value_exc_vat'] : $total['value_inc_tax']);
                }*/
            }
        }

        $tax = $this->formatRaw($tax);
        if ($this->requireItemVat) {
            $tax = $this->formatRaw($totalPayPalTax, $currency, 1);
        }
        if ($this->sendExVat) {
            $totalPayPal += $tax;
        }

        foreach ($items as $item) {
            $details['totals']['item_total']['value'] += $this->formatRaw($item['unit_amount']['value'] * $item['quantity'], $currency, 1); //already multiplied currency value
        }
        $details['totals']['item_total']['value'] = $this->formatRaw($details['totals']['item_total']['value'], $currency, 1);

        $totalPayPal += $details['totals']['item_total']['value'];

        if ($handlingValue) {
            $details['totals']['handling'] = [
              'currency_code' => $currency,
              'value' => $this->formatRaw($handlingValue),
            ];
        }
        $totalPayPal += $details['totals']['handling']['value'] ?? 0;
        $totalPayPal = $this->formatRaw($totalPayPal, $currency, 1);


        if ($discountValue) {
            //escape few penny rounding issue with discounts
            if (abs($totalPayPal - $this->formatRaw($discountValue) - $masterTotal)<0.05) {
                if ( ($totalPayPal - $this->formatRaw($discountValue) < $masterTotal) ) {
                    $discountValue -= abs($totalPayPal - $this->formatRaw($discountValue) - $masterTotal);
                } else {
                    $discountValue += abs($totalPayPal - $this->formatRaw($discountValue) - $masterTotal);
                }
            }
            $details['totals']['discount'] = [
              'currency_code' => $currency,
              'value' => $this->formatRaw($discountValue),
            ];
        }

///rounding issues
        $totalPayPal -= $details['totals']['discount']['value'] ?? 0;
        $totalPayPal = $this->formatRaw($totalPayPal, $currency, 1);
        if ($totalPayPal > $masterTotal) {
            $dscnt = $details['totals']['discount']['value']??0;
            $details['totals']['discount'] = [ //shipping_discount
              'currency_code' => $currency,
              'value' => $this->formatRaw($dscnt + $totalPayPal - $masterTotal, $currency, 1),
            ];
        } else if ($totalPayPal < $masterTotal) {
            $handling = $details['totals']['handling']['value']??0;
            $details['totals']['handling'] = [//insurance
              'currency_code' => $currency,
              'value' => $this->formatRaw($handling + $masterTotal - $totalPayPal, $currency, 1),
            ];
        }

        if ($this->sendExVat && $tax > 0) {
            $details['totals']['tax_total'] = [
              'currency_code' => $currency,
              'value' => $tax,
            ];
        }

        return $details;
    }

    public function createOrder($data = []) {
        global $cart, $cartID;

        if ($cart->count_contents() < 1) {
            $this->enabled = false;
        }
//2check protection (cart/shipping change while on confirmaiotn page)
        if (($cartID != $cart->cartID) && tep_session_is_registered('shipping')) {
            //2check
            $data['restart'] = true;
            $this->resetSession('shipping');
        }

        $post = $data;

        if (!empty($data['restart'])) {
        //if (false && !empty($post)) {//osc - never sends address and shipping details here (opc).
            $this->resetSession('estimate_ship');
            $this->resetSession('estimate_bill');
            $this->resetSession('payment');
            $this->resetSession('select_shipping');
            $this->resetSession('partner_order_id');
            $this->resetSession('partner_order_pending');
            $this->resetSession('partner_transaction_id');
            $this->resetSession('ppartner_total_check');
            $this->resetSession('ppp_order_before');
            //$this->resetSession('billto');
        } 
        
        global $order; /// shipping calculated on it
        /** @var order $order */
/*
        if (!$this->isInstance()) {
            $order = $this->createOrderInstance();
        } else {
            $order = $this->getOrderInstance();
        }*/

        /** @var paypal_rest $this */
        $this->checkoutOrderWithAddresses();// set default addresses and create order

//set  $this->shipping_modules session['shipping'], $this->shipping_cost $order->info['shipping*']
        $this->getShippingQuotesByChoice();
        
        $this->totalProcess();

        $order = $this->getOrderInstance();

        $request = new \PayPalCheckoutSdk\Orders\OrdersCreateRequest();
        $request->prefer('return=representation');
        if ($tmp = self::getAttributionId()) {
            $request->payPalPartnerAttributionId($tmp);
        }
        if (!empty($post['payment']) ) {
            $this->setAPMSetting($post);
        }

        self::debug(print_r($this, true), 'createOrder');
        //self::debug(print_r($order, true), 'createOrder');
        $request->body = $this->ppBuildOrderDetails($order, $data);
        /*if ($_GET['option']=='card') {
            $post['payment'] = 'paypal_rest_card';
        }*/


        if (!empty($post['payment']) ) {
            $this->addAPMDetails($request, $order, $post);
        }

        

        if (!empty($request->body['payment_source']) ) { //&& !isset($request->body['payment_source']['paypal'])
            $request->headers['PayPal-Request-Id'] = $this->getPayPalOrderAPIRequestId();
            $request->headers['PayPal-Client-Metadata-Id'] = $this->getFraudNetSessionId();
        }

        self::debug(print_r($request, true), 'createOrder');

        try {
            return $this->getHttpClient()->execute($request);
        } catch (\Exception $ex) {

            self::error('body ' . print_r($request->body, true), 'paypal_rest_exception');
            self::error('message ' . $ex->getMessage(), 'paypal_rest_exception');
            if (!empty($request->body['payment_source'])) {
                $this->resetFraudNetSessionId();
            }
            if ($tmp = $this->parseJsonMessage($ex->getMessage())) {
                return ['error' => 1, 'message' => $tmp];
            }
        }
        return false;
    }

    private function ppBuildOrderDetails($order, $post_data = []) {
        global $currency;
        global $currencies;
        $currency_code = $currency;
        if (!empty($order->info['currency'])) {
            $currency_code = $order->info['currency'];
        }

        $cartDetails = $this->getCartDetails();
        $applicationContext = [
          'return_url' => tep_href_link("ext/modules/payment/paypal_rest/index.php", 'action=return', 'SSL', true, false),
          'cancel_url' => tep_href_link("ext/modules/payment/paypal_rest/index.php", 'action=cancel', 'SSL', true, false),
        ];

        if (!$this->isShippingNeeded()) {
            $applicationContext['shipping_preference'] = 'NO_SHIPPING';
        }

        if (true) { //checkout page/logged in customer: all details are available.
            if (!empty($order->billing)) {
                $tmp = $order->billing;
                if (!empty($tmp['street_address']) && !empty($tmp['postcode']) && !empty($tmp['country']['iso_code_2'])) {
                } else {
                    $tmp = $this->getBillingAddress();
                }
            } else {
                $tmp = $this->getBillingAddress();
            }
            if ($tmp['zone_id'] > 0) {
                if ($tmp['country']['iso_code_2'] == 'US') {
                    $_state = tep_get_zone_code($tmp['country']['id'], $tmp['zone_id'], '');
                } else {
                    $_state = tep_get_zone_name($tmp['country']['id'], $tmp['zone_id'], $tmp['state']);
                }
                if (!empty($_state)) {
                    $tmp['state'] = $_state;
                }
            }

            if (!empty($post_data['Billing_address'])) {// opcUpdate
                foreach (['firstname', 'lastname', 'street_address', 'city', 'postcode', 'suburb'] as $k) {
                    if (empty($tmp[$k]) && !empty($post_data['Billing_address'][$k])) {
                        $tmp[$k] = strip_tags($post_data['Billing_address'][$k]);
                    }
                }
                if ($post_data['Billing_address']['country'] != $tmp['country']['id']) {
                    $_country = \common\helpers\Country::get_country_info_by_id($post_data['Billing_address']['country']);
                    if (!empty($_country['iso_code_2'])) {
                        $tmp['country'] = $_country;
                    }
                }
            } elseif (empty($tmp['firstname']) && empty($tmp['lastname'])) {
                $tmp['firstname'] = $order->customer['firstname']??'';
                $tmp['lastname'] = $order->customer['lastname']??'';
            }
            $payer = [];
            if (!empty($tmp['firstname']) && !empty($tmp['lastname'])) {
                $payer['name'] = [
                  'given_name' => substr($tmp['firstname'], 0, 140),
                  'surname' => substr($tmp['lastname'], 0, 140),
                ];
            }
            if (!empty($tmp['firstname']) && !empty($tmp['lastname'])) {
                if (!empty($order->customer['email_address'])) {
                    $payer['email_address'] = $order->customer['email_address'];
                } elseif (!empty($post_data['checkout']['email_address'])) {
                    $payer['email_address'] = $post_data['checkout']['email_address'];
                }
                if (!empty($post_data['checkout']['telephone'])) {
                    $payer['telephone'] = $post_data['checkout']['telephone'];
                } elseif (!empty($order->customer['telephone'])) {
                    $payer['telephone'] = $order->customer['telephone'];
                }

            }
            if (!empty($tmp['street_address']) && !empty($tmp['postcode']) && !empty($tmp['country']['iso_code_2'])) {
                $payer['address'] = [//billing
                  'address_line_1' => substr($tmp['street_address'], 0, 300),
                  'address_line_2' => substr($tmp['suburb'], 0, 300),
                  'admin_area_2' => substr($tmp['city'], 0, 120),
                  'admin_area_1' => substr($tmp['state'], 0, 300),
                  'postal_code' => substr($tmp['postcode'], 0, 60),
                  'country_code' => substr($tmp['country']['iso_code_2'], 0, 2),
                ];
            }
        }

        $purchaseUnits = [[
        //"reference_id" => "default",
            "amount" => [
              "value" => $this->formatRaw($order->info['total']),
              "currency_code" => $currency_code,
              "breakdown" => $cartDetails['totals'],
            ],
            "items" => $cartDetails['items'],
            'description' => (strlen(defined('STORE_NAME')?constant('STORE_NAME'):'')>127  ?
                  substr((defined('STORE_NAME')?constant('STORE_NAME'):''), 0, 123) . ' ...' :
                  (defined('STORE_NAME')?constant('STORE_NAME'):'')),
            'invoice_id' => $this->estimateOrderId() . '-e-' . date('ymdHis'),
// custom_id to hide from customer
        ]];

        //if (true) { //checkout page/logged in customer: all details are available.
        if ($this->isShippingNeeded()) {
            //$applicationContext['user_action'] = 'PAY_NOW';
            //$applicationContext['shipping_preference'] = 'SET_PROVIDED_ADDRESS';
            $tmp = $order->delivery;
//echo " order#### <PRE>"  . __FILE__ .':' . __LINE__ . ' ' . print_r($tmp, true) ."</PRE>";

            //2check - partial incorrect
            $tmp = $this->getDeliveryAddress();//patchOrder, custom opc
//echo "#### from m<PRE>"  . __FILE__ .':' . __LINE__ . ' ' . print_r($tmp , true) ."</PRE>"; die;

            if ($tmp['zone_id'] > 0) {
                if ($tmp['country']['iso_code_2'] == 'US') {
                    $_state = tep_get_zone_code($tmp['country']['id'], $tmp['zone_id'], '');
                } else {
                    $_state = tep_get_zone_name($tmp['country']['id'], $tmp['zone_id'], $tmp['state']);
                }
                if (!empty($_state)) {
                    $tmp['state'] = $_state;
                }
            }
            if (!empty($post_data['Shipping_address'])) {//2docustom - opc
                foreach (['firstname', 'lastname', 'street_address', 'city', 'postcode'] as $k) {
                    if (empty($tmp[$k]) && !empty($post_data['Shipping_address'][$k])) {
                        $tmp[$k] = strip_tags($post_data['Shipping_address'][$k]);
                    }
                }
                if ($post_data['Shipping_address']['country'] != $tmp['country']['id']) {
                    $_country = tep_get_countries($post_data['Shipping_address']['country'], true);
                    if (!empty($_country['countries_iso_code_2'])) {
                        $tmp['country'] = $_country;
                    }
                }
            }
            $purchaseUnits[0]['shipping'] = [];
            $options = $this->getShippingOptions($currencies, $currency_code);
            if (!empty($options)) {
                $purchaseUnits[0]['shipping']['options'] = $options;
            }

            if (false && isset($purchaseUnits[0]['shipping']['options'][0]['type']) && $purchaseUnits[0]['shipping']['options'][0]['type'] == 'PICKUP') {
                //2docustom pickup - not an option in osc by default
                $purchaseUnits[0]['shipping']['name'] = [];
                foreach ($purchaseUnits[0]['shipping']['options'] as $option) {
                    if (!empty($option['selected'])) {
                        $addr = []; //2docustom pickup - store country
                        $purchaseUnits[0]['shipping'] = [
                            'name' => [
                              'full_name' => substr('S2S ' . $option['label'], 0, 300),
                            ],
                            'type' => 'PICKUP_IN_PERSON',
                            'address' => [
                              'address_line_1' => substr($addr['street_address']??'', 0, 300),
                              'address_line_2' => substr($addr['suburb']??'', 0, 300),
                              'admin_area_2' => substr($addr['city']??'', 0, 120),
                              'admin_area_1' => substr($addr['state']??'', 0, 300),
                              'postal_code' => substr($addr['postcode']??'', 0, 60),
                              'country_code' => substr($addr['country_iso_code_2']??'', 0, 60),
                            ]
                        ];
                        $applicationContext['shipping_preference'] = 'NO_SHIPPING';
                        break;
                    }
                }


            } elseif (!empty($tmp['street_address']) && !empty($tmp['postcode']) && !empty($tmp['country']['iso_code_2'])) {
                $purchaseUnits[0]['shipping'] = [
                  'name' => [
                    'full_name' => substr($tmp['firstname'] . ' ' . $tmp['lastname'], 0, 300),
                  ],
                  //'type' => 'SHIPPING',
                  'address' => [
                    'address_line_1' => substr($tmp['street_address'], 0, 300),
                    'address_line_2' => substr($tmp['suburb'], 0, 300),
                    'admin_area_2' => substr($tmp['city'], 0, 120),
                    'admin_area_1' => substr($tmp['state'], 0, 300),
                    'postal_code' => substr($tmp['postcode'], 0, 60),
                    'country_code' => substr($tmp['country']['iso_code_2'], 0, 60),
                  ]
                ];
            } else {
                self::warning(" empty shipping address " .print_r($tmp, true), 'TLDEBUG');
            }
            //}
        }


        if (!$this->hasOwnKeys() && $this->_getIntent() != 'authorize') {
            $purchaseUnits[0]["payment_instruction"] = [//only in capture mode
                "disbursement_mode" => "INSTANT",
                "platform_fees" => [[
                    "amount" => [
                      "currency_code" => $currency_code,
                      "value" => 0, // $this->getFee($order->info['total_inc_tax']),
                    ]
                ]]
            ];
        }

        $seller = $this->getSeller();
        if (!empty($seller->email_address)) {
            $purchaseUnits[0]['payee'] =
                [
                  "email_address" => $seller->email_address
                ];
        }
        $applicationContext['locale'] = $this->getLocale();

        $ret = [
                "intent" => ($this->_getIntent() == 'authorize' ? "AUTHORIZE" : "CAPTURE"),
                "purchase_units" => $purchaseUnits,
                "application_context" => $applicationContext
            ];

        if (!empty($payer)) {
            if (false && empty($ret['payment_source'])) {
                if (!empty($payer['telephone'])) {
                    $payer['phone']['phone_number']['national_number'] = $payer['telephone'];
                    unset($payer['telephone']);
                }
                $ret['payment_source']['paypal'] = $payer;
            } else {
                $ret['payer'] = $payer;
            }
        }

        return $ret;
    }


    public function getLocale($replace = true){
        $locale = 'en_US';
        try {
            $currentLocale = setlocale(LC_CTYPE, 0);
            if (strpos($currentLocale, '.')) {
                $currentLocale = substr($currentLocale, 0, strpos($currentLocale, '.'));
            }
            if (!empty($currentLocale) && strlen($currentLocale)>4) {
                $locale = $currentLocale;
            } else {
            }
        } catch (\Exception $e ) {

        }
        if ($replace) {
            $locale = str_replace('_', '-', $locale);
        }
        return $locale;
    }

    public function isCheckout() {
        return $this->currentPage() == 'checkout';
    }

    public function currentPage() {
        global $PHP_SELF;
        $ret = '';
        switch (basename($PHP_SELF)) {
            case FILENAME_CHECKOUT_SHIPPING:
                $ret = 'shipping';
                break;
            case FILENAME_CHECKOUT_PAYMENT:
                $ret = 'checkout';
                break;
            case FILENAME_CHECKOUT_CONFIRMATION:
                $ret = 'confirmation';
                break;
            case FILENAME_SHOPPING_CART:
                $ret = 'shopping-cart';
                break;
            case FILENAME_LOGIN:
                $ret = 'login';
                break;

        }
        return $ret;
    }

    private function currentPageButtonsLayout() {
        $ret = 'False';
        if ($this->page == 'callback') {
            $ret = 'True';
        } elseif (defined('MODULE_PAYMENT_PAYPAL_REST_BUY_IMMEDIATELLY') && $this->currentPage() == 'catalog') {
            $ret = MODULE_PAYMENT_PAYPAL_REST_BUY_IMMEDIATELLY;
        } elseif (defined('MODULE_PAYMENT_PAYPAL_REST_SHOPPING_CART') && $this->currentPage() == 'shopping-cart') {
            $ret = MODULE_PAYMENT_PAYPAL_REST_SHOPPING_CART;
        } elseif (defined('MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT_LOGIN') && $this->currentPage() == 'login') {
            $ret = MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT_LOGIN;
        } elseif (defined('MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT') &&
                    (($this->currentPage() == 'checkout' && $this->currentPage() != 'login') )
                )
        {
            $ret = MODULE_PAYMENT_PAYPAL_REST_AT_CHECKOUT;
        }
        if ($ret == 'Fields') {
            $ret = 'Vertical';
        }

        return $ret;
    }

/**
 * oscok
 * @return string
 */
    function estimateOrderId() {
        $ret = '';
        try {
            $d = tep_db_fetch_array(tep_db_query("select max(orders_id) +1 as oid from " . TABLE_ORDERS));
            if ($d['oid']) {
                $ret .= $d['oid'];
            } else {
                $ret .= '1';
            }
        } catch (\Exception $e) {
            self::warning(" #### " . print_r($e->getMessage(), 1), 'TLDEBUG-' . $this->code??'payment');
        }
        return $ret;
    }

/**
 * oscok
 * @param type $currencies
 * @param type $currency_code
 * @return type
 */
    public function getShippingOptions($currencies, $currency_code) {

        $options = [];
        if ($this->isShippingNeeded()) {
            foreach ($this->getShippingQuotesByChoice() as $shipping_quote_item ) {
                if (empty($shipping_quote_item['error'])) {

                    foreach ($shipping_quote_item['methods'] as $shipping_quote_item_method) {
                        $label = html_entity_decode(strip_tags($shipping_quote_item['module'] . ' '. $shipping_quote_item_method['title']));
                        $mxl = 128;
                        if (strlen($label) > $mxl) {
                            $label = substr($label, 0, $mxl-3) . '...';
                        }
                        if ($this->sendExVat) {
                            $cost = $this->display_price_clear($shipping_quote_item_method['cost'], 0 , 1);
                        } else {
                            $cost = $this->display_price_clear($shipping_quote_item_method['cost'], $shipping_quote_item['tax'], 1);
                        }
                        $row = ['id' => $shipping_quote_item_method['code'],
                                'label' => $label,
                                'type' => ($shipping_quote_item['id'] != 'collect'? "SHIPPING" :"PICKUP"),
                                'selected' => $shipping_quote_item_method['selected']? true : false,
                                'amount' => [
                                  'value' =>  $this->formatRaw(
                                      $cost
                                      , $currency_code, 1
                                          ),
                                  'currency_code' => $currency_code
                                  ]
                               ];
                        $options[] = $row;
                    }
                }
            }
        }
        return $options;
    }

    public function display_price_clear($products_price, $products_tax, $quantity = 1) {
      global $currencies;
      global $currency;

      $number= $currencies->calculate_price($products_price, $products_tax, $quantity);

      $calculate_currency_value = true; 
      $currency_type = $currency_value = '';

      if (empty($currency_type)) $currency_type = $currency;

      $format_string = number_format(tep_round($number, $currencies->currencies[$currency_type]['decimal_places']), $currencies->currencies[$currency_type]['decimal_places'], $currencies->currencies[$currency_type]['decimal_point'], $this->currencies[$currency_type]['thousands_point']);

      return $format_string;
    }

    private function parseJsonMessage($msg) {
        $ret = false;
        try {
            $tmp = json_decode($msg, true);
            if (is_array($tmp['details'])) {
                $ret = '';
                foreach ( $tmp['details'] as $err) {
                    if (!empty($err['description'])) {
                        $ret .= $err['description'];
                        if (!empty($err['field'])) {
                            $ret .= ' (' . $err['field'];
                            if (!empty($err['value'])) {
                                $ret .= ' - ' . $err['value'];
                            }
                            $ret .= ')';
                        }
                        $ret .= "\n";
                    }
                }
            }
        } catch (\Exception $ex) {
            self::error('parseJsonMessage: ' . $ex->getMessage(), 'paypal_rest_exception');
        }
        return $ret;
    }
    
    public static function get_untaxed_value($price, $tax) {
        if ($tax == 100) return $price / 2;
        return $price * 100 / ($tax + 100);
    }

/**
 * get order details from PayPal by id
 * @param string $orderId
 * @return stdClass|false
 */
    public function getOrder($orderId) {
        $request = new \PayPalCheckoutSdk\Orders\OrdersGetRequest($orderId);

        try {
            return $this->getHttpClient()->execute($request);
        } catch (\Exception $ex) {
            self::error($ex->getMessage(), 'paypal_rest');
        }
        return false;
    }


/**
 * patch order at PayPal
 * @param string $orderId PayPal order Id
 * @return bool
 */
    public function updateOrder($orderId) {
        $request = new \PayPalCheckoutSdk\Orders\OrdersPatchRequest($orderId);
        $order = $this->getOrderInstance(); //?? VL2check wasn't init value 0

        $cartDetails = $this->getCartDetails();
        $details = $this->ppBuildOrderDetails($order);
        $data = [[
                'op' => 'replace',
                'path' => '/purchase_units/@reference_id==\'default\'',
                'value' => $details['purchase_units'][0]
                ]
        ];

        $request->body = $data;

        try {
            return $this->getHttpClient()->execute($request);
        } catch (\Exception $ex) {
            self::error($ex->getMessage(), 'paypal_rest');
        }
        return false;
    }

/**
 * patch order at PayPal with invoice id (local order id)
 * @param string $orderId PayPal order Id
 * @return bool
 */
    public function updateOrderInvoiceId($orderId, $invoiceId) {
        $request = new \PayPalCheckoutSdk\Orders\OrdersPatchRequest($orderId);
        $data = [[
                'op' => 'replace',
                'path' => '/purchase_units/@reference_id==\'default\'/invoice_id',
                'value' => $invoiceId
                ]
        ];

        $request->body = $data;

        try {
            return $this->getHttpClient()->execute($request);
        } catch (\Exception $ex) {
            self::error($ex->getMessage(), 'paypal_rest');
        }
        return false;
    }

    public function authorizeOrder($orderId) {

        $request = new \PayPalCheckoutSdk\Orders\OrdersAuthorizeRequest($orderId);
        $request->prefer('return=representation');
        if (self::getFraudnetSI()) {
            $metaData = $this->getFraudNetSessionId();
            if (!empty($metaData) ) {
                $request->payPalClientMetadataId($metaData);
            }
        }

        try {
            return $this->getHttpClient()->execute($request);
        } catch (\Exception $ex) {
            self::error($ex->getMessage(), 'paypal_partner');
        }
        return false;
    }

    public function captureOrder($orderId) {

        $request = new \PayPalCheckoutSdk\Orders\OrdersCaptureRequest($orderId);
        if ($this->getFraudnetSI()) {
            $metaData = $this->getFraudNetSessionId();
            if (!empty($metaData) ) {
                $request->payPalClientMetadataId($metaData);
            }
        }

        try {
            return $this->getHttpClient()->execute($request);
        } catch (\Exception $ex) {
            self::error($ex->getMessage(), 'paypal_rest');
        }
        return false;
    }

    function getCheckoutUrl(array $params, int $checkoutPage = 0) {
        switch ($checkoutPage) {
          case self::CONFIRMATION_PAGE :
            $url = defined('FILENAME_CHECKOUT_CONFIRMATION') ? FILENAME_CHECKOUT_CONFIRMATION : '';
            break;
          case self::PROCESS_PAGE :
            $url = defined('FILENAME_CHECKOUT_PROCESS') ? FILENAME_CHECKOUT_PROCESS : '';
            break;
          case self::PAYMENT_PAGE :
          default:
            $url = defined('FILENAME_CHECKOUT_PAYMENT') ? FILENAME_CHECKOUT_PAYMENT : '';
            break;
        }
        return $this->createAbsoluteUrl(array_merge([$url], $params), ((ENABLE_SSL == true) ? 'https' : 'http'));

    }

    public function getStatusCode($transaction) {
        $type = $this->transactionType($transaction);
        $sk = self::OPYS_PENDING;
        if (!empty($transaction->result->status))
          switch (strtoupper($transaction->result->status)) {
            case 'COMPLETED':
            case 'PARTIALLY_REFUNDED':
                if (in_array($type, ['authorize', 'autorization'])) { //always pending
                    $sk = self::OPYS_PENDING;
                } elseif (in_array($type, ['refund'])) {
                    $sk = self::OPYS_REFUNDED;
                } else {
                    $sk = self::OPYS_SUCCESSFUL;
                }
                break;
            case 'DECLINED':
                $sk = self::OPYS_REFUSED;
                break;
            case 'VOIDED':
                $sk = self::OPYS_CANCELLED;
                break;
            case 'REFUNDED':
                if (in_array($type, ['authorize', 'autorization'])) { //always pending
                    $sk = self::OPYS_PENDING;
                } elseif (in_array($type, ['refund'])) {
                    $sk = self::OPYS_REFUNDED;
                } else {
                    $sk = self::OPYS_SUCCESSFUL;
                }
                break;
            case 'PENDING':
            default:
                $sk = self::OPYS_PENDING;
                break;
        }
        return $sk;
    }

  public function paidOrderStatus() {
    $detectedStatus = null;
    if (!empty($this->paid_status)) {
      $detectedStatus = $this->paid_status;
    } elseif (defined('ORDER_STATUS_FULL_AMOUNT') && (int)ORDER_STATUS_FULL_AMOUNT >0){
      $detectedStatus = constant('ORDER_STATUS_FULL_AMOUNT');
    }
/*    if (!empty($detectedStatus) && !\common\helpers\Order::isStatusExist($detectedStatus) ){
      $detectedStatus = null;
    }*/
    if (empty($detectedStatus) ) {
        if (defined('ORDER_STATUS_FULL_AMOUNT') && (int)ORDER_STATUS_FULL_AMOUNT >0) {
          $detectedStatusBE = constant('ORDER_STATUS_FULL_AMOUNT');
/*          if ($detectedStatusBE && !\common\helpers\Order::isStatusExist($detectedStatusBE) ){
            $detectedStatusBE = null;
          }*/
        }
        if (defined('DEFAULT_ONLINE_PAYMENT_SUCCESS_ORDERS_STATUS_ID') && (int)DEFAULT_ONLINE_PAYMENT_SUCCESS_ORDERS_STATUS_ID >0) {
          $detectedStatusFE = constant('DEFAULT_ONLINE_PAYMENT_SUCCESS_ORDERS_STATUS_ID');
          /*if ($detectedStatusFE && !\common\helpers\Order::isStatusExist($detectedStatusFE) ){
            $detectedStatusFE = null;
          }*/
        }
        if (!empty($detectedStatusBE)) {
            $detectedStatus = $detectedStatusBE;
        } elseif (!empty($detectedStatusFE)) {
            $detectedStatus = $detectedStatusFE;
        }
    }

    return $detectedStatus;
  }
  
  public function getDefaultOrderStatusId(){
      return (!empty($this->order_status)?$this->order_status:DEFAULT_ORDERS_STATUS_ID);
  }


    public function generateClientToken() {
        $cid = $this->getCustomerAssigned();
        //in session for 2 min
        global $ppp_token_cid, $ppp_token_key, $ppp_token_until;
        foreach(['ppp_token_cid', 'ppp_token_key', 'ppp_token_until'] as $k) {
            if (!tep_session_is_registered($k)) {
                tep_session_register($k);
            }
        }

        if (!empty($ppp_token_until) && $ppp_token_until<time()) {
            $ppp_token_until = false;
        }
        //active and the same customer
        if (!empty($ppp_token_key) && !empty($ppp_token_until) &&
            (empty($cid) || $cid==$ppp_token_cid)) {
            return $ppp_token_key;
        }

        if (empty($cid)) {
            $cid = substr(sha1(uniqid()), 0, 20);
        }
        $request = new \PayPalCheckoutSdk\Core\ClientTokenRequest($cid);
        try {
            $res = $this->getHttpClient()->execute($request);
            if ($res->statusCode == 200 && !empty($res->result->client_token)) {
                $ppp_token_cid = $cid;
                $ppp_token_key = $res->result->client_token;
                $ppp_token_until = time()+120;//$res->result->expires_in
                return $ppp_token_key;
            } else {
                self::error(print_r($res, true), 'paypal_rest');
            }
        } catch (\Exception $ex) {
            self::error($ex->getMessage(), 'paypal_rest');
        }
        return false;
    }


    public static function setStatus($orderId = 0, $orderStatus = 0, $historyArray = []) {
        $__orders_id = 0;
        $return = false;
        $orderRecord = tep_db_fetch_array(tep_db_query("select * from " . TABLE_ORDERS . " where orders_id='" . (int)$orderId . "'"));
        if (!empty($orderRecord)) {
            $return = (int)$orderRecord['orders_status'];
            $orderRecord = (object) $orderRecord;
            $__orders_id = $orderRecord->orders_id;
            $isHistory = false;
            $orderStatus = (int)$orderStatus;
            $historyArray = (is_array($historyArray) ? $historyArray : []);

            $orderStatusRecord = tep_db_fetch_array(tep_db_query("select * from " . TABLE_ORDERS_STATUS . " where orders_status_id = '" . (int)$orderStatus . "'"));
            if (!empty($orderStatusRecord)) {
                if ($return != $orderStatus) {
                    //update order with new status
                    try {
                        $data = array(
                          'orders_status' => $orderStatus,
                          'last_modified' => date('Y-m-d H:i:s'));
                        tep_db_perform(TABLE_ORDERS, $data, 'update', "orders_id = '" . (int)$__orders_id . "'");
                        $isHistory = true;
                    } catch (\Exception $exc) {
                        self::warning($exc->getMessage() . ' ' . $exc->getTraceAsString());
                    }
                    $return = (int)$orderRecord->orders_status;
                }
            }
            $comments = trim(isset($historyArray['comments']) ? $historyArray['comments'] : '');
            $dateAdded = isset($historyArray['date_added'])?$historyArray['date_added']:'now()';
            $isNotified = (isset($historyArray['customer_notified']) ? (((int)$historyArray['customer_notified'] > 0) ? 1 : 0) : 0);

            if (($isHistory == true) OR ($comments != '') OR ($isNotified > 0)) {

                global $login_id;
                $data = array(
                  'orders_id' => (int)$__orders_id,
                  'orders_status_id' => $return,
                  'date_added' => $dateAdded,
                  //'admin_id' => (int)(isset($login_id) ? $login_id : 0),
                  'comments' => trim($comments),
                  'customer_notified' => ((int)$isNotified > 0 ? 1 : 0));

                tep_db_perform(TABLE_ORDERS_STATUS_HISTORY, $data);

            }
        }
        return $return;
    }


    /**
     * [draft] - calculate paid and refund total values in order payment table
     * @param int $orderId
     * @param float $orderTotal
     * @return array|false array:     [status] => 1 -  has due 0 eq -1 overpay
    [total] => 17.45 - order total (NOT calculated, passed in params)
    [debit] => 17.45 - successful payment transactions (OPYS_SUCCESSFUL)
    [credit] => 1.15  - refunded transactions (OPYS_REFUNDED)
    [discount] => 0 ??
    [paid] => 16.3,      [due] => 1.15   [over] => 0 calculated based on values above
     */
    public static function getTotalStatusArray($orderId = 0, $orderTotal = 0)
    {
        $return = false;
        $orderId = (int)$orderId;
        $orderTotal = (float)$orderTotal;
        if ($orderTotal < 0) {
            $orderTotal = 0;
        }
        $debit = 0;
        $credit = 0;
        $discount = 0;
        $q = tep_db_query("select * from orders_payment where orders_payment_order_id='" . (int)$orderId . "'");
        
        while($orderPaymentParentRecord = tep_db_fetch_array($q)) {
            $orderPaymentParentRecord['orders_payment_amount'] = (float)(((float)$orderPaymentParentRecord['orders_payment_amount'] <= 0)
                ? 0 : $orderPaymentParentRecord['orders_payment_amount']
            );
            $orderPaymentParentRecord['orders_payment_currency_rate'] = (float)(((float)$orderPaymentParentRecord['orders_payment_currency_rate'] <= 0)
                ? 1 : $orderPaymentParentRecord['orders_payment_currency_rate']
            );
            if (in_array((int)$orderPaymentParentRecord['orders_payment_status'], [self::OPYS_SUCCESSFUL, self::OPYS_REFUNDED, self::OPYS_DISCOUNTED])) {
                if ((int)$orderPaymentParentRecord['orders_payment_status'] == self::OPYS_SUCCESSFUL) {
                    $debit += ($orderPaymentParentRecord['orders_payment_amount'] / $orderPaymentParentRecord['orders_payment_currency_rate']);
                } elseif ((int)$orderPaymentParentRecord['orders_payment_status'] == self::OPYS_REFUNDED) {
                    $credit += ($orderPaymentParentRecord['orders_payment_amount'] / $orderPaymentParentRecord['orders_payment_currency_rate']);
                } elseif ((int)$orderPaymentParentRecord['orders_payment_status'] == self::OPYS_DISCOUNTED) {
                    $discount += ($orderPaymentParentRecord['orders_payment_amount'] / $orderPaymentParentRecord['orders_payment_currency_rate']);
                }
            } elseif (in_array((int)$orderPaymentParentRecord['orders_payment_status'], [])) {

            }
        }
        unset($orderPaymentParentRecord);
        $discount = self::toAmount($discount);
        $discount = ($discount <= 0 ? 0 : $discount);
        $return = [
            'status' => 0,
            'total' => $orderTotal,
            'debit' => self::toAmount($debit),
            'credit' => self::toAmount($credit),
            'discount' => $discount,
            'paid' => 0,
            'due' => 0,
            'over' => 0
        ];
        $return['paid'] = (($return['debit'] + $return['discount']) - $return['credit']);
        $return['due'] = ($return['total'] - $return['paid']);
        if ($return['due'] > 0) {
            $return['status'] = 1;
        } elseif ($return['due'] < 0) {
            $return['status'] = -1;
            $return['over'] = abs($return['due']);
            $return['due'] = 0;
        }
        unset($credit);
        unset($debit);
        return $return;
    }
    
    private static function toAmount($amount = 0) {
        return round((float)$amount, 2);
    }

  public function updatePaidTotalsAndNotify($commentary = '') {
    $updateOrder = true;
    if ($this->isInstance() ) {
      $order = $this->getOrderInstance();
    }
    self::debug(print_r($order, true), 'order');

    if ($updateOrder && $order && $order->info['orders_id']) {
      $updated = self::getTotalStatusArray($order->info['orders_id']);
      if ($updated) { //update order status and notify customer if required
        $status = '';
        if (isset($updated['paid']) ) {
          if ($updated['details']['status']>0) {
            $status = $this->paidOrderStatus();
          } else {
         //   $status = $this->partlyPaidOrderStatus();
            $status = $this->order_status;
          }
        } elseif (isset($updated['refund']) && ($updated['details']['credit']>0 || $updated['details']['due']>0)) {
            $tmp = (($updated['details']['credit']??0)>0 ? $updated['details']['credit'] : $updated['details']['due']);
            /*
          if (abs(
              round($updated['details']['total'], 2) -
              round($tmp, 2)
              //round($updated['details']['credit'], 2)
              ) < 0.01) {
            $status = $this->refundOrderStatus();
          } else {
            $status = $this->partialRefundOrderStatus();
          }*/
        }
      self::debug(print_r($updated, true), 'getTotalStatusArray $status' . $status);


        if (1 && !empty($status) /*&& $status != $order->info['order_status']*/) {
            $check_status_query = tep_db_query("select customers_name, customers_email_address, orders_status, date_purchased from " . TABLE_ORDERS . " where orders_id = '" . (int)$order->info['orders_id'] . "'");
            $check_status = tep_db_fetch_array($check_status_query);
            self::debug(print_r($commentary, true), '$commentary');
            if ( ($check_status['orders_status'] != $status) || tep_not_null($commentary)) {
                self::setStatus($order->info['orders_id'], $status, ['comments' => $commentary]);
                //notify
                global $language, $languages_id;
                try {
                    if (function_exists('tep_catalog_href_link')) {
                        $link = tep_catalog_href_link(FILENAME_CATALOG_ACCOUNT_HISTORY_INFO, 'order_id=' . $order->info['orders_id'], 'SSL');
                    } else {
                        $link = tep_href_link(FILENAME_ACCOUNT_HISTORY_INFO, 'order_id=' . $order->info['orders_id'], 'SSL');
                    }
                    $orders_status__name = tep_db_fetch_array(tep_db_query("select orders_status_id, orders_status_name from " . TABLE_ORDERS_STATUS . " where language_id = '" . (int)$languages_id . "' and orders_status_id='" . $status . "'"));

                    @include_once('admin/' . DIR_WS_LANGUAGES . $language . '/orders.php');

                    $email = STORE_NAME . "\n" . EMAIL_SEPARATOR . "\n" . EMAIL_TEXT_ORDER_NUMBER . ' ' . $order->info['orders_id'] . "\n" . EMAIL_TEXT_INVOICE_URL . ' ' . $link . "\n" . EMAIL_TEXT_DATE_ORDERED . ' ' . tep_date_long($check_status['date_purchased']) . "\n\n" . $commentary . sprintf(EMAIL_TEXT_STATUS_UPDATE, $orders_status__name);

                    tep_mail($check_status['customers_name'], $check_status['customers_email_address'], EMAIL_TEXT_SUBJECT, $email, STORE_OWNER, STORE_OWNER_EMAIL_ADDRESS);
                } catch (\Exception $ex) {
                    self::debug(print_r($ex->getMessage(), true));
                }
            }
        }
      }
    }
  }

}